#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/*
	Create a volumetric dataset containing the so called "blob surface"
	described by the equation 
	f(x,y,z) = x^2 + y^2 + z^2 + sin(4x) + sin(4y) + sin(4z)
	The blob surface is normally defined where f(x,y,z) - 1 = 0
	Define over the range -3 to 3 on each axis.
*/

#define TRUE  1
#define FALSE 0

int debug = FALSE;
int nvol = 256;
unsigned short ***vol = NULL;
double xmin = -3,ymin = -3,zmin = -3;
double xmax =  3,ymax =  3,zmax =  3;

int main(int argc,char **argv)
{
	int i,j,k;
	double x,y,z,d;
	FILE *fptr;

   // Create space for the volumentric representation
   vol = malloc(nvol*sizeof(unsigned short **));
   for (i=0;i<nvol;i++)
      vol[i] = malloc(nvol*sizeof(unsigned short *));
   for (i=0;i<nvol;i++)
      for (j=0;j<nvol;j++)
         vol[i][j] = malloc(nvol*sizeof(unsigned short));
   for (i=0;i<nvol;i++)
      for (j=0;j<nvol;j++)
			for (k=0;k<nvol;k++)
				vol[i][j][k] = 0;

	// Compute the volume
	for (k=0;k<nvol;k++) {
		z = zmin + k * (zmax - zmin) / (nvol - 1);
      for (j=0;j<nvol;j++) {
         y = ymin + j * (ymax - ymin) / (nvol - 1);
			for (i=0;i<nvol;i++) {
				x = xmin + i * (xmax - xmin) / (nvol - 1);

				// Evaluate the function
				d = x*x + y*y + z*z + sin(4 * x) + sin(4 * y) + sin(4 * z) - 1;

				// Abbitrarily shift it to positive coordinates
				d = d * 2000 + 32000;

				// Clip to short
				if (d < 0) d = 0;
				if (d > 64000) d = 64000;

				// Write to volume
				vol[i][j][k] = rint(d);
			}
		}
	}

   // Write "really" raw file
   if ((fptr = fopen("blob.raw","w")) == NULL) {
		fprintf(stderr,"Raw data write failed\n");
		exit(-1);
	}
   for (i=0;i<nvol;i++) {
      for (j=0;j<nvol;j++) {
         for (k=0;k<nvol;k++) {
            fwrite(&vol[i][j][k],sizeof(unsigned short),1,fptr);
         }
      }
   }
   fclose(fptr);

	exit(0);
}

