#include "offviewer.h"

/*
	Fast OFF file viewer
	Sept 2006   Added Geomview style off support
               Added background colour option
               Added cropping box, options.cropmin and options.cropmax
*/

char *interfacestring = "Interface\n\
   Left mouse button            Rotate camera around model\n\
   Left mouse button + shift    Pan camera\n\
   Center mouse button          Roll camera\n\
   Center mouse button + shift  Move camera forward or backward\n\
   Right mouse button           Popup menus\n\
   Arrow keys                   Rotate camera\n\
   Arrow keys + shift           Pan camera\n\
   < >                          Move camera forward or backward\n\
   + -                          Zoom in/out (Change aperture)\n\
   [ ]                          Roll camera\n\
   a A                          Automatic rotation\n\
   9 0                          Change zero parallax distance\n\
   i h c                        Information, help, construction lines\n\
   f1 ... f6                    Special camera views\n\
   1 2 3                        Render modes\n\
   w                            Window dump to TGA file\n\
   r                            Toggle continuous window dump (Use with care)\n\
   q esc                        Quit";

OPTIONS options;
CAMERA  camera;
INTERFACESTATE interfacestate;

int nvertex = 0;
fXYZ *vertices = NULL;
fXYZ *normals = NULL;
fCOLOUR *colours = NULL;
int npolygon = 0;
ITRIANGLE *polygons = NULL;
XYZ themin = {1e32,1e32,1e32},themax = {-1e32,-1e32,-1e32},themid;
double framerate = 0;

int main(int argc,char **argv)
{
	int i;
	COLOUR grey = {0.2,0.2,0.2};
	XYZ verysmall = {-1e32,-1e32,-1e32},verylarge = {1e32,1e32,1e32};

   // The default camera attributes 
   CameraHome(0);
   camera.stereo       = NOSTEREO;
   camera.screenwidth  = 800;
   camera.screenheight = 600;

	// Default options 
	options.debug        = FALSE;
	options.fullscreen   = FALSE;
	options.windowdump   = FALSE;
	options.record       = FALSE;
	options.targetfps    = 30;
	options.rendermode   = WIREFRAME;
	options.showinfo     = FALSE;
	options.showhelp     = FALSE;
	options.showconstr   = FALSE;
	options.autorotate   = 0;
	options.offtype      = 0;
	options.bgcolour     = grey;
	options.cropmin      = verysmall;
	options.cropmax      = verylarge;

	/* State of the input device, mouse in this case */
	interfacestate.button = -1;
	interfacestate.shift  = FALSE;
	interfacestate.mouseh = 0;
	interfacestate.mousev = 0;

	/* Parse the command line arguments */
	if (argc < 2) 
		GiveUsage(argv[0]);
   for (i=1;i<argc-1;i++) {
      if (strcmp(argv[i],"-h") == 0) 
         GiveUsage(argv[0]);
      if (strcmp(argv[i],"-f") == 0)
         options.fullscreen = !options.fullscreen;
      if (strcmp(argv[i],"-s") == 0)
         camera.stereo = ACTIVESTEREO;
		if (strcmp(argv[i],"-o") == 0) 
			options.offtype = atoi(argv[i+1]) % 2; // Just 0 or 1
		if (strcmp(argv[i],"-ss") == 0) {
         camera.stereo = DUALSTEREO;
			camera.screenwidth *= 2;
		}
		if (strcmp(argv[i],"-rw") == 0)
         options.rendermode = WIREFRAME;
      if (strcmp(argv[i],"-rs") == 0)
         options.rendermode = SHADED;
      if (strcmp(argv[i],"-rp") == 0)
         options.rendermode = SPECULAR;
      if (strcmp(argv[i],"-a") == 0)
         options.autorotate = 1;
		if (strcmp(argv[i],"-bg") == 0) {
			options.bgcolour.r = atof(argv[i+1]);
         options.bgcolour.g = atof(argv[i+2]);
         options.bgcolour.b = atof(argv[i+3]);
		}
		if (strcmp(argv[i],"-fps") == 0)
			options.targetfps = atof(argv[i+1]);
		if (strcmp(argv[i],"-d") == 0)
			options.debug = TRUE;
		if (strcmp(argv[i],"-dx") == 0) {
			options.cropmin.y = atof(argv[i+1]);
         options.cropmax.y = atof(argv[i+2]);
		}
      if (strcmp(argv[i],"-dy") == 0) {
         options.cropmin.y = atof(argv[i+1]);
         options.cropmax.y = atof(argv[i+2]);
      }
      if (strcmp(argv[i],"-dz") == 0) {
         options.cropmin.z = atof(argv[i+1]);
         options.cropmax.z = atof(argv[i+2]);
      }
   }
	
	// Attempt to read data
	if (!ReadOFF(argv[argc-1],options.offtype)) {
		fprintf(stderr,"File read failed - fatal\n");
		exit(-1);
	}

   // Set things up and go 
   glutInit(&argc,argv);
   if (camera.stereo == ACTIVESTEREO)
		glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_STEREO);
	else
      glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

	// Create the window and handlers 
   glutCreateWindow("OFF Viewer");
	glutReshapeWindow(camera.screenwidth,camera.screenheight);
   if (options.fullscreen)
      glutFullScreen();
   glutDisplayFunc(HandleDisplay);
	glutReshapeFunc(HandleReshape);
	CreateMenus();
   glutKeyboardFunc(HandleKeyboard);
   glutSpecialFunc(HandleSpecialKeyboard);
   glutMouseFunc(HandleMouse);
	glutMotionFunc(HandleMouseMotion);
	glutPassiveMotionFunc(HandlePassiveMotion);
	glutVisibilityFunc(HandleVisibility);      
	if (camera.stereo == DUALSTEREO)
		glutSetCursor(GLUT_CURSOR_NONE);
	glutIdleFunc(HandleIdle);
   CreateEnvironment();

	// Ready to go!
	CameraHome(0);
   glutMainLoop();
   return(0);
}

/*
   This is where global OpenGL settings are made.
	Since this is not designed to demonstrate how to render
	objects in OpenGL, it's pretty boring. 
*/
void CreateEnvironment(void)
{
	/* Miscellaneous settings */
	glDisable(GL_CULL_FACE);
   glEnable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
   glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glPixelStorei(GL_UNPACK_ALIGNMENT,1);

	/* These are hardware dependent */
	glDisable(GL_DITHER);
   glDisable(GL_LINE_SMOOTH);
   glDisable(GL_POINT_SMOOTH);
   glDisable(GL_POLYGON_SMOOTH);
   glHint(GL_POINT_SMOOTH_HINT,GL_FASTEST);
   glHint(GL_LINE_SMOOTH_HINT,GL_FASTEST);
   glHint(GL_POLYGON_SMOOTH_HINT,GL_FASTEST);
	glHint(GL_FOG_HINT,GL_FASTEST);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_FASTEST);

	/* Default geometry rendering */
   glLineWidth(1.0);
   glPointSize(1.0);
   glFrontFace(GL_CCW);
   glShadeModel(GL_SMOOTH);
   glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);

   /* Default meterials */
	glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

   /* Turn off all the lights */
   glDisable(GL_LIGHT0);
   glDisable(GL_LIGHT1);
   glDisable(GL_LIGHT2);
   glDisable(GL_LIGHT3);
   glDisable(GL_LIGHT4);
   glDisable(GL_LIGHT5);
   glDisable(GL_LIGHT6);
   glDisable(GL_LIGHT7);
   glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_FALSE);
   glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
	glClearColor(options.bgcolour.r,options.bgcolour.g,options.bgcolour.b,0.0);
}

/*
   This is the basic display callback routine
	This is where the real projection work for stereo
	is done, note the viewer is assumed to be located
	centrally to the screen and looking along a vector
	perpendicular to the screen.
*/
void HandleDisplay(void)
{
	XYZ r;
   double ratio,radians,wd2,ndfl;
   double left,right,top,bottom;
	double dt;
	static int framecounter = 0;
	static double tstart = -1;

	if (tstart < 0)
		tstart = GetRunTime();

   /* Misc stuff needed for the frustum */
   ratio   = camera.screenwidth / (double)camera.screenheight;
	if (camera.stereo == DUALSTEREO)
		ratio /= 2;
   radians = DTOR * camera.aperture / 2;
   wd2     = camera.near * tan(radians);
   ndfl    = camera.near / camera.focallength;
   top     =   wd2;
   bottom  = - wd2;

	/* Determine the right eye vector */
   CROSSPROD(camera.vd,camera.vu,r);
   Normalise(&r);
   r.x *= camera.eyesep / 2.0;
   r.y *= camera.eyesep / 2.0;
   r.z *= camera.eyesep / 2.0;

	if (camera.stereo == ACTIVESTEREO || camera.stereo == DUALSTEREO) {

      if (camera.stereo == DUALSTEREO) {
         glDrawBuffer(GL_BACK);
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      }

   	glMatrixMode(GL_PROJECTION);
   	glLoadIdentity();
      left  = - ratio * wd2 - 0.5 * camera.eyesep * ndfl;
      right =   ratio * wd2 - 0.5 * camera.eyesep * ndfl;
      glFrustum(left,right,bottom,top,camera.near,camera.far);
		if (camera.stereo == DUALSTEREO)
			glViewport(camera.screenwidth/2,0,camera.screenwidth/2,camera.screenheight);
		else
			glViewport(0,0,camera.screenwidth,camera.screenheight);
      glMatrixMode(GL_MODELVIEW);
      glDrawBuffer(GL_BACK_RIGHT);
      if (camera.stereo == ACTIVESTEREO)
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      glLoadIdentity();
      gluLookAt(camera.vp.x + r.x,camera.vp.y + r.y,camera.vp.z + r.z,
                camera.vp.x + r.x + camera.vd.x,
                camera.vp.y + r.y + camera.vd.y,
                camera.vp.z + r.z + camera.vd.z,
                camera.vu.x,camera.vu.y,camera.vu.z);
      CreateGeometry(FALSE);

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      left  = - ratio * wd2 + 0.5 * camera.eyesep * ndfl;
      right =   ratio * wd2 + 0.5 * camera.eyesep * ndfl;
      glFrustum(left,right,bottom,top,camera.near,camera.far);
      if (camera.stereo == DUALSTEREO)
         glViewport(0,0,camera.screenwidth/2,camera.screenheight);
      else
         glViewport(0,0,camera.screenwidth,camera.screenheight);
      glMatrixMode(GL_MODELVIEW);
      glDrawBuffer(GL_BACK_LEFT);
      if (camera.stereo == ACTIVESTEREO)
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      glLoadIdentity();
      gluLookAt(camera.vp.x - r.x,camera.vp.y - r.y,camera.vp.z - r.z,
                camera.vp.x - r.x + camera.vd.x,
                camera.vp.y - r.y + camera.vd.y,
                camera.vp.z - r.z + camera.vd.z,
                camera.vu.x,camera.vu.y,camera.vu.z);
      CreateGeometry(FALSE);

	} else {

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
		glViewport(0,0,camera.screenwidth,camera.screenheight); 
      left  = - ratio * wd2;
      right =   ratio * wd2;
      glFrustum(left,right,bottom,top,camera.near,camera.far);
      glMatrixMode(GL_MODELVIEW);
      glDrawBuffer(GL_BACK);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      glLoadIdentity();
      gluLookAt(camera.vp.x,camera.vp.y,camera.vp.z,
                camera.vp.x + camera.vd.x,
                camera.vp.y + camera.vd.y,
                camera.vp.z + camera.vd.z,
                camera.vu.x,camera.vu.y,camera.vu.z);
   	CreateGeometry(FALSE);
	}

	/* Dump the window in compressed TGA format */
	if (options.record || options.windowdump) {
		if (camera.stereo == ACTIVESTEREO)
			WindowDump("",camera.screenwidth,camera.screenheight,TRUE,12);
		else
			WindowDump("",camera.screenwidth,camera.screenheight,FALSE,12);
		options.windowdump = FALSE;
	}

   /* Swap buffers */
   glutSwapBuffers();

   /* Compute the instantaneous frame rate */
   dt = GetRunTime() - tstart;
   if (dt > 0.5) {
      framerate = framecounter / dt;
      tstart = GetRunTime();
      framecounter = 0;
   }
   framecounter++;

	if (options.autorotate != 0)
		RotateCamera(1.0,0.0,0.0,options.autorotate/5.0);
}

/*
	Sort out rendering and lighting modes 
*/
void MakeLighting(void)
{
   GLfloat p[4];
   GLfloat white[4] = {1.0,1.0,1.0,1.0};
	GLfloat midgrey[4] = {0.2,0.2,0.2,1.0};
	GLfloat black[4] = {0.0,0.0,0.0,1.0};
	GLfloat shiny[1] = {100.0};

	/* Create lights */
	p[0] = themid.x + (themax.x - themin.x);
	p[1] = themid.y + (themax.y - themin.y);
	p[2] = themid.z + (themax.z - themin.z);
	p[3] = 1.0;
	glLightfv(GL_LIGHT0,GL_POSITION,p);
   glLightfv(GL_LIGHT0,GL_DIFFUSE,white);
	switch (options.rendermode) {
	case WIREFRAME:
	case SHADED:
		glLightfv(GL_LIGHT0,GL_SPECULAR,black);
		break;
	case SPECULAR:
		glLightfv(GL_LIGHT0,GL_SPECULAR,white);
		break;
	}
	glLightfv(GL_LIGHT0,GL_AMBIENT,black);
	glEnable(GL_LIGHT0);

   p[0] = camera.vp.x + camera.vu.x;
   p[1] = camera.vp.y + camera.vu.y;
   p[2] = camera.vp.z + camera.vu.z;
   p[3] = 1.0;
   glLightfv(GL_LIGHT0,GL_POSITION,p);
   glLightfv(GL_LIGHT0,GL_DIFFUSE,white);
   switch (options.rendermode) {
   case WIREFRAME:
   case SHADED:
      glLightfv(GL_LIGHT1,GL_SPECULAR,black);
      break;
   case SPECULAR:
      glLightfv(GL_LIGHT1,GL_SPECULAR,white);
      break;
   }
   glLightfv(GL_LIGHT0,GL_AMBIENT,black);
   //glEnable(GL_LIGHT1);

   /* Rendering mode specific settings */
   switch (options.rendermode) {
   case WIREFRAME:
      glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
      glLightModelfv(GL_LIGHT_MODEL_AMBIENT,white);
      glDisable(GL_LIGHTING);
      break;
   case SHADED:
	case SPECULAR:
      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      glLightModelfv(GL_LIGHT_MODEL_AMBIENT,midgrey);
      glEnable(GL_LIGHTING);
      break;
   }

	/* Materials */
   switch (options.rendermode) {
   case WIREFRAME:
      glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,white);
      glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,shiny);
		glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,black);
      break;
   case SHADED:
	case SPECULAR:
      glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,white);
      glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,shiny);
		glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,black);
		break;
	}
}

/*
	Create the geometry
	Cropping is done here, retain whole model in memery
*/
void CreateGeometry(int build)
{
	int i,index1,index2,index3;
	static int first = TRUE;
	fCOLOUR lastcolour = {-1.0,-1.0,-1.0};
	fXYZ p1,p2,p3;

	MakeLighting();

	if (build || first) {
		//glEnableClientState(GL_COLOR_ARRAY);
		//glEnableClientState(GL_VERTEX_ARRAY);
		//glEnableClientState(GL_NORMAL_ARRAY);
		//glColorPointer(3,GL_FLOAT,0,colours);
		//glNormalPointer(GL_FLOAT,0,normals);
		//glVertexPointer(3,GL_FLOAT,0,vertices);
		if (options.debug)
			fprintf(stderr,"Rebuilding display list\n");
		glNewList(1000,GL_COMPILE);
		glBegin(GL_TRIANGLES);
		for (i=0;i<npolygon;i++) {
			//glArrayElement(polygons[i].p1);
         //glArrayElement(polygons[i].p2);
         //glArrayElement(polygons[i].p3);
         index1 = polygons[i].p1;
         index2 = polygons[i].p2;
         index3 = polygons[i].p3;
			p1 = vertices[index1];
			p2 = vertices[index2];
			p3 = vertices[index3];

			if (p1.x < options.cropmin.x && p2.x < options.cropmin.x && p3.x < options.cropmin.x)
				continue;
         if (p1.x > options.cropmax.x && p2.x > options.cropmax.x && p3.x > options.cropmax.x)
            continue;
         if (p1.y < options.cropmin.y && p2.y < options.cropmin.y && p3.y < options.cropmin.y)
            continue;
         if (p1.y > options.cropmax.y && p2.y > options.cropmax.y && p3.y > options.cropmax.y)
            continue;
         if (p1.z < options.cropmin.z && p2.z < options.cropmin.z && p3.z < options.cropmin.z)
            continue;
         if (p1.z > options.cropmax.z && p2.z > options.cropmax.z && p3.z > options.cropmax.z)
            continue;

			if (!SameColour(lastcolour,colours[index1])) {
				glColor3f(colours[index1].r,colours[index1].g,colours[index1].b);
				lastcolour = colours[index1];
			}
     	   glNormal3f(normals[index1].x,normals[index1].y,normals[index1].z);
     	   glVertex3f(p1.x,p1.y,p1.z);
			if (!SameColour(lastcolour,colours[index2])) {
     	   	glColor3f(colours[index2].r,colours[index2].g,colours[index2].b);
				lastcolour = colours[index2];
			}
     	   glNormal3f(normals[index2].x,normals[index2].y,normals[index2].z); 
     	   glVertex3f(p2.x,p2.y,p2.z);
			if (!SameColour(lastcolour,colours[index3])) {
     	   	glColor3f(colours[index3].r,colours[index3].g,colours[index3].b);
				lastcolour = colours[index3];
			}
     	   glNormal3f(normals[index3].x,normals[index3].y,normals[index3].z); 
     	   glVertex3f(p3.x,p3.y,p3.z);
		}
		glEnd();
		glEndList();
	
		first = FALSE;
		return;
	}

	glCallList(1000);
	DrawInfoHelp();
}

/*
   Deal with plain key strokes
*/
void HandleKeyboard(unsigned char key,int x, int y)
{
   if (glutGetModifiers() == GLUT_ACTIVE_SHIFT)
      interfacestate.shift = TRUE;
   else
      interfacestate.shift = FALSE;

	/* Handle keyboard commands */
   switch (key) {
   case ESC: 		
   case 'Q':
   case 'q': 
		exit(0);
		break;
	case '1':
   	options.rendermode = WIREFRAME;
   	CreateGeometry(TRUE);
		break;
   case '2':
      options.rendermode = SHADED;
      CreateGeometry(TRUE);
      break;
   case '3':
      options.rendermode = SPECULAR;
      CreateGeometry(TRUE);
      break;
	case 'a':
		options.autorotate--;
		break;
	case 'A':
		options.autorotate++;
		break;
   case 'i':
   case 'I':
      options.showinfo = !options.showinfo;
      break;
	case 'h':
	case 'H':
		options.showhelp = !options.showhelp;
      break;
   case 'c':
   case 'C':
      options.showconstr = !options.showconstr;
      break;
   case '[':
      RotateCamera(0.0,0.0,-1.0,1.0);
      break;
   case ']':
      RotateCamera(0.0,0.0,1.0,1.0);
      break;
   case 'w':
	case 'W':
		options.windowdump = TRUE;
      break;
   case 'r':
   case 'R':
      options.record = !options.record;
      break;
   case '<':
   case ',':
		TranslateCamera(0.0,0.0,-camera.focallength/50);
		break;
   case '>':
   case '.':
		TranslateCamera(0.0,0.0,+camera.focallength/50);
      break;
	case '-':
	case '_':
		if (camera.aperture < 90)
			camera.aperture *= 1.05;
		break;
	case '+':
   case '=':
		if (camera.aperture > 5)
			camera.aperture /= 1.05;
      break;
	case '(':
	case '9':
		camera.focallength *= 1.05;
		break;
	case ')':
	case '0':
		camera.focallength /= 1.05;
		break;
	}
}

/*
   Deal with special key strokes
*/
void HandleSpecialKeyboard(int key,int x, int y)
{
   if (glutGetModifiers() == GLUT_ACTIVE_SHIFT)
      interfacestate.shift = TRUE;
   else
      interfacestate.shift = FALSE;

   switch (key) {
   case GLUT_KEY_LEFT:  
		if (interfacestate.shift)
			TranslateCamera(-camera.focallength/200,0.0,0.0);
		else
			RotateCamera(-1.0,0.0,0.0,1.0); 
		break;
   case GLUT_KEY_RIGHT: 
      if (interfacestate.shift)
         TranslateCamera(camera.focallength/200,0.0,0.0);
      else
			RotateCamera(1.0,0.0,0.0,1.0);  
		break;
   case GLUT_KEY_UP:    
      if (interfacestate.shift)
         TranslateCamera(0.0,camera.focallength/200,0.0);
      else
			RotateCamera(0.0,1.0,0.0,1.0);  
		break;
   case GLUT_KEY_DOWN:  
      if (interfacestate.shift)
         TranslateCamera(0.0,-camera.focallength/200,0.0);
      else
			RotateCamera(0.0,-1.0,0.0,1.0); 
		break;
   case GLUT_KEY_F1:
      CameraHome(1);
      break;
   case GLUT_KEY_F2:
      CameraHome(2);
      break;
   case GLUT_KEY_F3:
      CameraHome(3);
      break;
   case GLUT_KEY_F4:
      CameraHome(4);
      break;
   case GLUT_KEY_F5:
      CameraHome(5);
      break;
   case GLUT_KEY_F6:
      CameraHome(6);
      break;
   }
}

/*
   Rotate (ix,iy) or roll (iz) the camera about the focal point
   ix,iy,iz are flags, 0 do nothing, +- 1 rotates in opposite directions
   Correctly updating all camera attributes
*/
void RotateCamera(double ix,double iy,double iz,double dtheta)
{
   XYZ vp,vu,vd;
   XYZ right;
   XYZ newvp,newr;
   double radius,dd,radians;
   double dx,dy,dz;

   vu = camera.vu;
   Normalise(&vu);
   vp = camera.vp;
   vd = camera.vd;
   Normalise(&vd);
   CROSSPROD(vd,vu,right);
   Normalise(&right);
   radians = dtheta * PI / 180.0;

   /* Handle the roll */
   if (ABS(iz) > 0.0001) {
      camera.vu.x += iz * right.x * radians;
      camera.vu.y += iz * right.y * radians;
      camera.vu.z += iz * right.z * radians;
      Normalise(&camera.vu);
      return;
   }
 
   /* Distance from the rotate point */
   dx = camera.vp.x - camera.pr.x;
   dy = camera.vp.y - camera.pr.y;
   dz = camera.vp.z - camera.pr.z;
   radius = sqrt(dx*dx + dy*dy + dz*dz);
 
   /* Determine the new view point */
   dd = radius * radians;
   newvp.x = vp.x + dd * ix * right.x + dd * iy * vu.x - camera.pr.x;
   newvp.y = vp.y + dd * ix * right.y + dd * iy * vu.y - camera.pr.y;
   newvp.z = vp.z + dd * ix * right.z + dd * iy * vu.z - camera.pr.z;
   Normalise(&newvp);
   camera.vp.x = camera.pr.x + radius * newvp.x;
   camera.vp.y = camera.pr.y + radius * newvp.y;
   camera.vp.z = camera.pr.z + radius * newvp.z;
 
   /* Determine the new right vector */
   newr.x = camera.vp.x + right.x - camera.pr.x;
   newr.y = camera.vp.y + right.y - camera.pr.y;
   newr.z = camera.vp.z + right.z - camera.pr.z;
   Normalise(&newr);
   newr.x = camera.pr.x + radius * newr.x - camera.vp.x;
   newr.y = camera.pr.y + radius * newr.y - camera.vp.y;
   newr.z = camera.pr.z + radius * newr.z - camera.vp.z;
   Normalise(&newr);
 
   camera.vd.x = camera.pr.x - camera.vp.x;
   camera.vd.y = camera.pr.y - camera.vp.y;
   camera.vd.z = camera.pr.z - camera.vp.z;
   Normalise(&camera.vd);
 
   /* Determine the new up vector */
   CROSSPROD(newr,camera.vd,camera.vu);
   Normalise(&camera.vu);
}

/*
   Translate (pan) the camera view point
*/
void TranslateCamera(double ix,double iy,double iz)
{
   XYZ right;

   CROSSPROD(camera.vd,camera.vu,right);
   Normalise(&right);

   camera.vp.x += iy*camera.vu.x + ix*right.x + iz*camera.vd.x;
   camera.vp.y += iy*camera.vu.y + ix*right.y + iz*camera.vd.y;
   camera.vp.z += iy*camera.vu.z + ix*right.z + iz*camera.vd.z;

   camera.pr.x += iy*camera.vu.x + ix*right.x + iz*camera.vd.x;
   camera.pr.y += iy*camera.vu.y + ix*right.y + iz*camera.vd.y;
   camera.pr.z += iy*camera.vu.z + ix*right.z + iz*camera.vd.z;
}

/*
   Handle mouse events.
	Do nothing special, just remember the state.
*/
void HandleMouse(int button,int state,int x,int y)
{
   if (glutGetModifiers() == GLUT_ACTIVE_SHIFT)
      interfacestate.shift = TRUE;
   else
      interfacestate.shift = FALSE;

   if (state == GLUT_DOWN) {
      if (button == GLUT_LEFT_BUTTON) {
			interfacestate.button = GLUT_LEFT_BUTTON;
      } else if (button == GLUT_MIDDLE_BUTTON) {
			interfacestate.button = GLUT_MIDDLE_BUTTON;
      } else {
			/* Assigned to menus */
		}
   }
   interfacestate.mouseh = x;
   interfacestate.mousev = y;
}

/*
   Handle mouse motion
*/
void HandleMouseMotion(int x,int y)
{
   double dx,dy;

   dx = x - interfacestate.mouseh;
   dy = y - interfacestate.mousev;

   if (interfacestate.button == GLUT_LEFT_BUTTON) {
		if (interfacestate.shift)
			TranslateCamera(-camera.focallength*dx/100,camera.focallength*dy/100,0.0);
		else
      	RotateCamera(-dx,dy,0.0,0.5);
   } else if (interfacestate.button == GLUT_MIDDLE_BUTTON) {
      if (interfacestate.shift)
         TranslateCamera(0.0,0.0,camera.focallength*dx/100);
      else
      	RotateCamera(0.0,0.0,dx,0.5);
	}

   interfacestate.mouseh = x;
   interfacestate.mousev = y;
}

/*
	What to do on moviing nouse, the button isn't down.
*/
void HandlePassiveMotion(int x,int y)
{
   interfacestate.mouseh = x;
   interfacestate.mousev = y;
}

/*
   How to handle visibility
	If the window is not visible then don't draw it,
	allows layered windows without a processor load blowout.
*/
void HandleVisibility(int visible)
{
   if (visible == GLUT_VISIBLE)
      glutIdleFunc(HandleIdle);
	else
		glutIdleFunc(NULL);
}

/*
   What to do on an idle event.
	Maintain a constant frame rate.
*/
void HandleIdle(void)
{
   static double tstart = -1;
   double tstop;

   if (tstart < 0)
      tstart = GetRunTime();
   tstop = GetRunTime();
   if (tstop - tstart > 1.0/options.targetfps) {
      // Nasty hack to solve Mac OS-X cursor problem
      if (camera.stereo == DUALSTEREO) {
         glutSetCursor(GLUT_CURSOR_CROSSHAIR);
         glutSetCursor(GLUT_CURSOR_NONE);
      }
      glutPostRedisplay();
      tstart = tstop;
   }
}

/*
   Handle a window reshape/resize
*/
void HandleReshape(int w,int h)
{
   camera.screenwidth = w;
   camera.screenheight = h;
}

/*
	Move the camera to the home position 
*/
void CameraHome(int position)
{
	XYZ up = {0.0,0.0,1.0},x = {1.0,0.0,0.0};

	camera.aperture = 60;
	camera.focallength = VectorLength(themin,themax);
   camera.eyesep = camera.focallength / 25;
   camera.near = camera.focallength / 10;
   camera.far = camera.focallength * 10;

	//options.autorotate = 0;

	switch (position) {
	case 0:
		camera.vp.x = themid.x - camera.focallength;
		camera.vp.y = themid.y; 
		camera.vp.z = themid.z;
   	camera.vd.x = 1;
      camera.vd.y = 0;
      camera.vd.z = 0;
   	camera.vu = up;
		break;
	case 1: /* Front */
      camera.vp.x = themid.x - camera.focallength;
      camera.vp.y = themid.y;
      camera.vp.z = themid.z;
      camera.vd.x = 1;
      camera.vd.y = 0;
      camera.vd.z = 0;
		camera.vu = up;
		break;
	case 2: /* Back */
      camera.vp.x = themid.x + camera.focallength;
      camera.vp.y = themid.y;
      camera.vp.z = themid.z;
      camera.vd.x = -1;
      camera.vd.y = 0;
      camera.vd.z = 0;
		camera.vu = up;
		break;
	case 3: /* Left */
      camera.vp.x = themid.x;
      camera.vp.y = themid.y + camera.focallength;
      camera.vp.z = themid.z;
      camera.vd.x = 0;
      camera.vd.y = -1;
      camera.vd.z = 0;
		camera.vu = up;
		break;
	case 4: /* Right */
      camera.vp.x = themid.x;
      camera.vp.y = themid.y - camera.focallength;
      camera.vp.z = themid.z;
      camera.vd.x = 0;
      camera.vd.y = 1;
      camera.vd.z = 0;
		camera.vu = up;
		break;
	case 5: /* Top */
      camera.vp.x = themid.x;
      camera.vp.y = themid.y;
      camera.vp.z = themid.z + camera.focallength;
      camera.vd.x = 0;
      camera.vd.y = 0;
      camera.vd.z = -1;
		camera.vu = x;
		break;
	case 6: /* Bottom */
      camera.vp.x = themid.x;
      camera.vp.y = themid.y; 
      camera.vp.z = themid.z - camera.focallength;
      camera.vd.x = 0;
      camera.vd.y = 0;
      camera.vd.z = 1;
		camera.vu = x;
		break;
	}

	camera.pr = themid;

	Normalise(&camera.vd);
	Normalise(&camera.vu);
}

/*
   Draw various bits of information 
*/
void DrawInfoHelp(void)
{
   char s[256];
   int y = 6;
	double len;
	XYZ p;

   if (options.showinfo) {
		glDisable(GL_LIGHTING);
		glColor3f(1.0,1.0,1.0);

   	sprintf(s,"Vertices: %d, Polygons: %d",nvertex,npolygon);
   	DrawGLText(10,y+=14,s);

   	sprintf(s,"Frame rate: %.1f",framerate);
   	DrawGLText(10,y+=14,s);

		sprintf(s,"        %g <= z <= %g",themin.z,themax.z);
		DrawGLText(10,y+=14,s);
      sprintf(s,"        %g <= y <= %g",themin.y,themax.y);
      DrawGLText(10,y+=14,s);
      sprintf(s,"Bounds: %g <= x <= %g",themin.x,themax.x);
      DrawGLText(10,y+=14,s);
	}

	if (options.showconstr) {

		/* Draw the rotation point */
		len = camera.focallength / 20;
		glColor3f(0.5,1.0,0.5);
   	glBegin(GL_LINES);
   	glVertex3f(camera.pr.x-len,camera.pr.y,camera.pr.z); /* x */
   	glVertex3f(camera.pr.x+len,camera.pr.y,camera.pr.z);
   	glVertex3f(camera.pr.x,camera.pr.y-len,camera.pr.z); /* y */
   	glVertex3f(camera.pr.x,camera.pr.y+len,camera.pr.z);
   	glVertex3f(camera.pr.x,camera.pr.y,camera.pr.z-len); /* z */
   	glVertex3f(camera.pr.x,camera.pr.y,camera.pr.z+len);
   	glEnd();
	
      /* Draw the focal point */
      len = camera.focallength / 25;
		p.x = camera.vp.x + camera.focallength * camera.vd.x;
      p.y = camera.vp.y + camera.focallength * camera.vd.y; 
      p.z = camera.vp.z + camera.focallength * camera.vd.z; 
      glColor3f(1.0,0.0,0.0);
      glBegin(GL_LINES);
      glVertex3f(p.x-len,p.y,p.z); /* x */
      glVertex3f(p.x+len,p.y,p.z);
      glVertex3f(p.x,p.y-len,p.z); /* y */
      glVertex3f(p.x,p.y+len,p.z);
      glVertex3f(p.x,p.y,p.z-len); /* z */
      glVertex3f(p.x,p.y,p.z+len);
      glEnd();

		/* Draw the bounding box */
		glColor3f(1.0,0.5,0.5);
		CreateBox(themin,themax,0);

		/* Draw the axes */
		glBegin(GL_LINES);
		glColor3f(1.0,0.0,0.0);
		glVertex3f(0.0,0.0,0.0);
		glVertex3f(len,0.0,0.0);
   	glColor3f(0.0,1.0,0.0);
   	glVertex3f(0.0,0.0,0.0);
   	glVertex3f(0.0,len,0.0);
   	glColor3f(0.0,0.0,1.0);
   	glVertex3f(0.0,0.0,0.0);
   	glVertex3f(0.0,0.0,len);
		glEnd();
	}

	if (options.showhelp) {
		glColor3f(1.0,1.0,1.0);
		DrawGLText(10,camera.screenheight-20,interfacestring);
	}
}

/*
	Read an off file, both variants
	variation 0 consists of
		nvertices
		list of vertices normals and colours all on one line
		npolygons
		list of polygons
	variation 1 consists of
		OFF
		nvertices npolygons nedges
		list of vertices
		list of polygons
*/
int ReadOFF(char *fname,int variation)
{
	int i,nedge,j;
	int len,index1,index2,index3;
	char ident[256];
	XYZ pa,pb,normal;
	double length;
	FILE *fptr;

   // Attempt to open the file 
   if ((fptr = fopen(fname,"r")) == NULL) {
      fprintf(stderr,"Failed to read file \"%s\"\n",fname);
      return(FALSE);
   }

	// Read header
	if (variation == 0) {
   	if (fscanf(fptr,"%d",&nvertex) != 1) {
   	   fprintf(stderr,"Failed to read the number of vertices\n");
   	   return(FALSE);
   	}
	} else {
		if (fscanf(fptr,"%s",ident) != 1) {
         fprintf(stderr,"Failed to read header line\n");
         return(FALSE);
		}
		if (strcmp(ident,"OFF") != 0) {
			fprintf(stderr,"Expected \"OFF\" on first line\n");
			return(FALSE);
		}
		if (fscanf(fptr,"%d %d %d",&nvertex,&npolygon,&nedge) != 3) {
			fprintf(stderr,"Failed to read number of vertices and polygons\n");
			return(FALSE);
		}
	}
	if (options.debug)
		fprintf(stderr,"Expect %d vertices\n",nvertex);

	// Malloc space
   if ((vertices = malloc(nvertex*sizeof(XYZ))) == NULL) {
      fprintf(stderr,"Vertex malloc failed\n");
      return(FALSE);
   }
   if ((normals = malloc(nvertex*sizeof(XYZ))) == NULL) {
      fprintf(stderr,"Normals malloc failed\n");
      return(FALSE);
   }
   if ((colours = malloc(nvertex*sizeof(COLOUR))) == NULL) {
      fprintf(stderr,"Colour malloc failed\n");
      return(FALSE);
   }

	// Read vertices
   for (i=0;i<nvertex;i++) {
      if (fscanf(fptr,"%f %f %f",&(vertices[i].x),&(vertices[i].y),&(vertices[i].z)) != 3) {
         fprintf(stderr,"Failed to read vertices (%d)\n",i);
         return(FALSE);
      }
      themin.x = MIN(themin.x,vertices[i].x);
      themin.y = MIN(themin.y,vertices[i].y);
      themin.z = MIN(themin.z,vertices[i].z);
      themax.x = MAX(themax.x,vertices[i].x);
      themax.y = MAX(themax.y,vertices[i].y);
      themax.z = MAX(themax.z,vertices[i].z);
		if (variation == 0) {
      	if (fscanf(fptr,"%f %f %f",&(normals[i].x),&(normals[i].y),&(normals[i].z)) != 3) {
      	   fprintf(stderr,"Failed to read normal (%d)\n",i);
      	   return(FALSE);
      	}
      	if (fscanf(fptr,"%f %f %f",&(colours[i].r),&(colours[i].g),&(colours[i].b)) != 3) {
      	   fprintf(stderr,"Failed to read colour (%d)\n",i);
      	   return(FALSE);
      	}
		} else {
			colours[i].r = 0.5, colours[i].g = 0.5; colours[i].b = 0.5;
			normals[i].x = 0; normals[i].y = 0; normals[i].z = 0;
		}
   }
   themid.x = (themin.x + themax.x) / 2;
   themid.y = (themin.y + themax.y) / 2;
   themid.z = (themin.z + themax.z) / 2;
	if (options.debug) {
   	fprintf(stderr,"Bounds: %g <= x <= %g\n",themin.x,themax.x);
   	fprintf(stderr,"        %g <= y <= %g\n",themin.y,themax.y);
   	fprintf(stderr,"        %g <= z <= %g\n",themin.z,themax.z);
	}

	if (variation == 0) {
   	if (fscanf(fptr,"%d",&npolygon) != 1) {
   	   fprintf(stderr,"Failed to read the number of polygons\n");
   	   return(FALSE);
   	}
	}
	if (options.debug)
   	fprintf(stderr,"Expect %d polygons\n",npolygon);
   if ((polygons = malloc(npolygon*sizeof(ITRIANGLE))) == NULL) {
      fprintf(stderr,"Polygon malloc failed\n");
      return(FALSE);
   }

	// Read polygons
   for (i=0;i<npolygon;i++) {
      if (fscanf(fptr,"%d %d %d %d",&len,&index1,&index2,&index3) != 4) {
         fprintf(stderr,"Failed to read polygon (%d)\n",i);
         return(FALSE);
      }
      polygons[i].p1 = index1;
      polygons[i].p2 = index2;
      polygons[i].p3 = index3;
      if (len != 3) {
         fprintf(stderr,"Can only handle triangles (%d), ignoring\n",i);
         i--;
         npolygon--;
         continue;
      }
   }
   fclose(fptr);

	// Calculate normals for variation 1
	if (variation == 1) {
		if (options.debug)
			fprintf(stderr,"Computing normals\n");
		for (j=0;j<npolygon;j++) {
			pa.x = vertices[polygons[j].p2].x - vertices[polygons[j].p1].x;
         pa.y = vertices[polygons[j].p2].y - vertices[polygons[j].p1].y;
         pa.z = vertices[polygons[j].p2].z - vertices[polygons[j].p1].z;
         pb.x = vertices[polygons[j].p3].x - vertices[polygons[j].p1].x;
         pb.y = vertices[polygons[j].p3].y - vertices[polygons[j].p1].y;
         pb.z = vertices[polygons[j].p3].z - vertices[polygons[j].p1].z;
			normal.x = (pa.y * pb.z - pa.z * pb.y);
			normal.y = (pa.z * pb.x - pa.x * pb.z);
			normal.z = (pa.x * pb.y - pa.y * pb.x);
			normals[polygons[j].p1].x += normal.x; 
			normals[polygons[j].p1].y += normal.y; 
			normals[polygons[j].p1].z += normal.z;
         normals[polygons[j].p2].x += normal.x;
         normals[polygons[j].p2].y += normal.y;
         normals[polygons[j].p2].z += normal.z;
         normals[polygons[j].p3].x += normal.x;
         normals[polygons[j].p3].y += normal.y;
         normals[polygons[j].p3].z += normal.z;
		}
		for (i=0;i<nvertex;i++) {
   		length = normals[i].x * normals[i].x + normals[i].y * normals[i].y + normals[i].z * normals[i].z;
   		if (length > 0) {
   		   length = sqrt(length);
   		   normals[i].x /= length;
   		   normals[i].y /= length;
   		   normals[i].z /= length;
   		} else {
				fprintf(stderr,"Zero length normal, should nothappen\n");
   		}
		}
	}

	return(TRUE);
}

