#include <stdio.h>
#include <string.h>
#include "polyr.h"

/* 
	Parses the command line options 
*/
int parse_options(int argc, char **argv)
{
  int i;
  int found;
  int founddowns=0,foundfilter=0;
  double atof();
  
  i = 1;
  while (i < (argc - 2)) {
    found = 0;
    
    if (!strcmp(argv[i], "-r")) {
      if (i >= (argc - 3)) {
			fprintf(stderr, "not enough arguments for -r\n",argv[0]);
			return -1;
      }
      i++;
      found = 1;
      REQPOLS=atof(argv[i++]);
      if ((REQPOLS<=0 || REQPOLS>=10000000)) {
			fprintf(stderr, "%s: number of pols out of bounds\n",argv[0]);
			return -1;
      }
    }
    if (!strcmp(argv[i], "-d")) {
      if (i >= (argc - 5)) {
			fprintf(stderr, "%s: not enough arguments for -d\n",argv[0]);
			return -1;
      }
      i++;
      found = 1;
      founddowns=1;
      DOWNSAMPL[0]=(float)atof(argv[i++]);
      DOWNSAMPL[1]=(float)atof(argv[i++]);
      DOWNSAMPL[2]=(float)atof(argv[i++]);
      if ((DOWNSAMPL[0]<1.0) || (DOWNSAMPL[1]<1.0) || (DOWNSAMPL[2]<1.0)) {
			fprintf(stderr, "%s: downsampling ratio out of bounds\n",argv[0]);
			return -1;
      }
    }
    if (!strcmp(argv[i], "-D")) {
      if (founddowns) {
			fprintf(stderr, "%s: both -d and -D specified\n",argv[0]);
			return -1;
      }
      if (i >= (argc - 3)) {
			fprintf(stderr, "%s: not enough arguments for -D\n",argv[0]);
			return -1;
      }
      i++;
      found = 1;
    
      DOWNSAMPL[0]=DOWNSAMPL[1]=DOWNSAMPL[2]=(float)atof(argv[i++]);
      if ((DOWNSAMPL[0]<1.0)) {
			fprintf(stderr, "%s: downsampling ratio out of bounds\n",argv[0]);
			return -1;
      }
    }
    if (!strcmp(argv[i], "-f")) {
      if (i >= (argc - 5)) {
			fprintf(stderr, "%s: not enough arguments for -d\n",argv[0]);
			return -1;
      }
      i++;
      found = 1;
      foundfilter=1;
      
      FILTER[0]=(float)atof(argv[i++]);
      FILTER[1]=(float)atof(argv[i++]);
      FILTER[2]=(float)atof(argv[i++]);
      if ((FILTER[0]<1.0) || (FILTER[1]<1.0) || (FILTER[2]<1.0)) {
			fprintf(stderr, "%s: filter ratio out of bounds\n",argv[0]);
			return -1;
      }
    }
    if (!strcmp(argv[i], "-F")) {
      if (foundfilter) {
			fprintf(stderr, "%s: both -f and -F specified\n",argv[0]);
			return -1;
      }
      if (i >= (argc - 3)) {
			fprintf(stderr, "%s: not enough arguments for -d\n",argv[0]);
			return -1;
      }
      i++;
      found = 1;
      
      FILTER[0]=FILTER[1]=FILTER[2]=(float)atof(argv[i++]);
      if ((FILTER[0]<3.0)) {
			fprintf(stderr, "%s: filter ratio out of bounds\n",argv[0]);
			return -1;
      }
    }
    if (!strcmp(argv[i], "-m")) {
      if (i >= (argc - 3)) {
			fprintf(stderr, "%s: not enough arguments for -d\n",argv[0]);
			return -1;
      }
      i++;
      found = 1;
      
      MAXPOLS=atof(argv[i++]);
      if (MAXPOLS<0) {
			fprintf(stderr, "%s: maxpols must be positive\n",argv[0]);
			return -1;
      }
    }
	 if (!strcmp(argv[i],"-z")) {
		 ZSCALE = atof(argv[i+1]);
	 }
    if (!strcmp(argv[i], "-o")) {
      if (i >= (argc - 3)) {
			fprintf(stderr, "%s: not enough arguments for -o\n",argv[0]);
			return -1;
      }
      if (STORE_POLYGONS) {
			fprintf(stderr, "%s: only one output file allowed\n",argv[0]);
			return -1;
      }
      i++;
      found = 1;
      strcpy(OUTNAME, argv[i++]);
      DUMP_OFF=1;
      STORE_POLYGONS=1;
    }
    if (!strcmp(argv[i], "-n")) {
      if (i >= (argc - 3)) {
			fprintf(stderr, "%s: not enough arguments for -n\n",argv[0]);
			return -1;
      }
      if (STORE_POLYGONS) {
			fprintf(stderr, "%s: only one output file allowed\n",argv[0]);
			return -1;
      }
      i++;
      found = 1;
      strcpy(OUTNAME, argv[i++]);
      DUMP_NFF=1;
      STORE_POLYGONS=1;
    }
    if (!strcmp(argv[i], "-p")) {
      if (i >= (argc - 3)) {
         fprintf(stderr, "%s: not enough arguments for -n\n",argv[0]);
         return -1;
      }
      if (STORE_POLYGONS) {
         fprintf(stderr, "%s: only one output file allowed\n",argv[0]);
         return -1;
      }
      i++;
      found = 1;
      strcpy(OUTNAME, argv[i++]);
      DUMP_POV=1;
      STORE_POLYGONS=1;
    }
    if (!strcmp(argv[i], "-l")) {
      if (i >= (argc - 3)) {
         fprintf(stderr, "%s: not enough arguments for -n\n",argv[0]);
         return -1;
      }
      if (STORE_POLYGONS) {
         fprintf(stderr, "%s: only one output file allowed\n",argv[0]);
         return -1;
      }
      i++;
      found = 1;
      strcpy(OUTNAME, argv[i++]);
      DUMP_OBJ=1;
      STORE_POLYGONS=1;
    }
    if (!strcmp(argv[i], "-i")) {
      if (i >= (argc - 3)) {
			fprintf(stderr, "%s: not enough arguments for -i\n",argv[0]);
			return -1;
      }
      if (STORE_POLYGONS) {
			fprintf(stderr, "%s: only one output file allowed\n",argv[0]);
			return -1;
      }
      i++;
      found = 1;
      strcpy(OUTNAME, argv[i++]);
      DUMP_IV=1;
      STORE_POLYGONS=1;
    }
    if (!strcmp(argv[i], "-v")) {
      i++;
		found = 1;
      VERBOSE = 1;
    }
    if (!strcmp(argv[i], "-raw")) {
      if (i >= (argc - 5)) {
			fprintf(stderr, "%s: not enough arguments for -raw\n",argv[0]);
			return -1;
      }
      XDIM = atoi(argv[++i]);
      YDIM = atoi(argv[++i]);
      ZDIM = atoi(argv[++i]);
      RAW_INPUT = 1;
      i++;
      found = 1;
      
    }
    if (!strcmp(argv[i], "-ana")) {
      if (i >= (argc - 2)) {
			fprintf(stderr, "%s: not enough arguments for -ana\n",argv[0]);
			return -1;
      }
      ANA_INPUT = 1;
      i++;
      found = 1;
    }
    if (!strcmp(argv[i], "-ms")) {
      if (i >= (argc - 2)) {
			fprintf(stderr, "%s: not enough arguments for -ms\n",argv[0]);
			return -1;
      }
      MS_INPUT = 1;
      i++;
      found = 1;
    }
    if (!strcmp(argv[i], "-hdf")) {
      if (i >= (argc - 2)) {
			fprintf(stderr, "%s: not enough arguments for -hdf\n",argv[0]);
			return -1;
      }
      HDF_INPUT = 1;
      i++;
      found = 1;
    }
    if (!found) {
      printf("%s: error, illegal option %s\n",argv[0],argv[i]);
      return -1;
    }
  }
  return 0;
}

/*
 	Usage
*/
void GiveUsage(char *s)
{
   printf("\nusage: %s [options] filetype infile threshold\n",s);
   printf("options:  -r n         Request number of polygons\n");
   printf("          -f x y z     Mean filter using a boxfilter [x,y,z]\n");
   printf("          -F n         Mean filter using a boxfilter [n,n,n]\n");
   printf("          -d x y z     Downsample array by x, y and z\n");
   printf("          -D n         Downsample array by n in all directions\n");
   printf("          -m           Min. number of connected tris (default 150)\n");
   printf("          -o name      Geomview '.off' output\n");
   printf("          -n name      WTK '.nff' output\n");
   printf("          -i name      Inventor '.iv' output\n");
	printf("          -p name      PovRay mesh2 output\n");
	printf("          -l name      Lightwave OBJ file\n");
   printf("          -z n         Z axis scale factor\n");
   printf("          -v           Verbose operation\n");
   printf("filetype: -ana         Analyze file\n");
   printf("          -ms          Multistruct file\n");
   printf("          -hdf         NCSA HDF file\n");
   printf("          -raw x y z   Raw input file with dims\n");
   printf("The operations on are preformed in the following order:\n");
   printf("  Filtering, Downsampling, Isosurfacing, Mesh decimation.\n");
   printf("\n");
}
