#include "glvol.h"

extern OPTIONS options;
extern VOLUME volume;
extern CAMERA camera;
extern RGBAI *colourmap;
extern int *histogram;
extern TEXTURE *textures;
extern int *histogram;
extern MARKER *markers;
extern int nmarkers;
extern int currentbutton;
extern XYZ cursor3d;

#define HISTLENGTH 512
#define HISTHEIGHT 128

/*
   Create all the geometry
*/
void DrawAll(void)
{
   CreateEnvironment();             // Default OpenGL settings

   DrawDecoration();                // Draw both bounding boxes

   if (options.show3dcursor)
      Draw3dCursor();

   if (options.showmarkers) {
      DrawMarkers();
   }

   DrawPlanes();

   if (options.showhistogram) {
      glDisable(GL_DEPTH_TEST);
      DrawHistogram();
      glEnable(GL_DEPTH_TEST);
   }
   if (options.showinfo) {
      glDisable(GL_DEPTH_TEST);
      DrawInfo();
      glEnable(GL_DEPTH_TEST);
   }
}

/*
   Draw the histogram
	The mappings are MAPPINGSIZE long
*/
void DrawHistogram(void)
{
   int i,j;
   int offset = 10,border = 6;
   int themax = 0;
	double v,sum;

   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
	if (camera.stereo == DUALSTEREO)
   	glOrtho(0,camera.screenwidth/2,0,camera.screenheight,-1,1);
	else
		glOrtho(0,camera.screenwidth,0,camera.screenheight,-1,1);
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glLoadIdentity();

   glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);

   /* Transparent patch */
   glColor4f(1.0,1.0,1.0,0.2);
   glBegin(GL_QUADS);
   glVertex3i(offset,offset,0);
   glVertex3i(offset,offset + HISTHEIGHT + 5*border,0);
   glVertex3i(offset + HISTLENGTH + 2*border,offset + HISTHEIGHT + 5*border,0);
   glVertex3i(offset + HISTLENGTH + 2*border,offset,0);
   glEnd();

   /* Graph - log scale */
   for (i=1;i<MAPPINGSIZE;i++)
   	themax = MAX(themax,histogram[i]);
   glColor3f(1,0.75,0.75);
   glBegin(GL_LINES);
   for (i=0;i<HISTLENGTH;i++) {

		/* Determine the average over the drawn histogram line */
		sum = 0;
		for (j=i*(MAPPINGSIZE/HISTLENGTH);j<(i+1)*(MAPPINGSIZE/HISTLENGTH);j++)
			sum += histogram[j];
		sum /= (MAPPINGSIZE/HISTLENGTH);

		/* Don't do anything for 0 values */
		if (sum < 1)
			continue;

		/* Draw the line */
      v = log(sum) / log((double)themax);
		if ((j = rint(HISTHEIGHT * v)) > HISTHEIGHT)
			j = HISTHEIGHT;
      glVertex3i(offset+border+i,offset+4*border,0);
      glVertex3i(offset+border+i,offset+4*border+j,0);
   }
	glEnd();

   /* Colour ramp */
	glBegin(GL_LINES);
   for (i=0;i<HISTLENGTH;i++) {
		j = i * (MAPPINGSIZE/HISTLENGTH);
      glColor3f(colourmap[j].r,colourmap[j].g,colourmap[j].b);
      glVertex3i(offset+border+i,offset+2*border,0);
      glVertex3i(offset+border+i,offset+4*border-1,0);
   }
   glEnd();

   /* Draw the alpha channel */
   glColor3f(1.0,1.0,1.0);
   glBegin(GL_LINE_STRIP);
   for (i=0;i<HISTLENGTH;i++) {
      j = HISTHEIGHT * colourmap[i*(MAPPINGSIZE/HISTLENGTH)].a;
      glVertex3i(offset + border + i,offset + 4*border + j,0);
   }  
   glEnd();

   /* Axes */
   glColor3f(0.0,0.0,0.0);
   glBegin(GL_LINES);
   glVertex3i(offset + border - 1,offset + HISTHEIGHT + 4*border,0);
   glVertex3i(offset + border - 1,offset + 2*border,0);
   glVertex3i(offset + border - 1,offset + 4*border - 1,0);
   glVertex3i(offset + border + HISTLENGTH,offset + 4*border - 1,0);
   glEnd();
	
	/* Ticks */
	glColor3f(0.0,0.0,0.0);
	glBegin(GL_LINES);
	for (i=0;i<=50;i++) {
		if (i % 5 == 0) {
			glVertex3i(offset + border + i * HISTLENGTH / 50.0,offset + border,0);
			glVertex3i(offset + border + i * HISTLENGTH / 50.0,offset + 2*border,0);
		} else {
         glVertex3i(offset + border + i * HISTLENGTH / 50.0,offset + 3*border/2,0);
         glVertex3i(offset + border + i * HISTLENGTH / 50.0,offset + 2*border,0);
		}
	}
	glEnd();

   glPopMatrix();
   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
}

/*
   Show various decorations
   Note, the corner of the volume is always at the origin
*/
void DrawDecoration(void)
{
   double d = 5;
   XYZ p1,p2;

   /* Entire volume */
   p1.x = volume.x;
   p1.y = volume.y;
   p1.z = volume.z;
   p2.x = p1.x + volume.nx * volume.dx;
   p2.y = p1.y + volume.ny * volume.dy;
   p2.z = p1.z + volume.nz * volume.dz;
   glColor3f(0.5,0.5,0.5);
   CreateBox(p1,p2,0);

   if (!options.showdecoration)
      return;

   /* Inspection box */
   p1.x = volume.x + camera.x0 * volume.dx;
   p1.y = volume.y + camera.y0 * volume.dy;
   p1.z = volume.z + camera.z0 * volume.dz;
   p2.x = volume.x + (camera.x0 + camera.xdim) * volume.dx;
   p2.y = volume.y + (camera.y0 + camera.ydim) * volume.dy;
   p2.z = volume.z + (camera.z0 + camera.zdim) * volume.dz;
   glColor3f(1.0,1.0,0.0);
   CreateBox(p1,p2,0);

   /* Inspection Axes */
   glBegin(GL_LINES);
   glColor3f(1.0,0.0,0.0);
   glVertex3f(p1.x-d*volume.dx,p1.y-d*volume.dy,p1.z-d*volume.dz);
   glVertex3f(p2.x            ,p1.y-d*volume.dy,p1.z-d*volume.dz);
   glColor3f(0.0,1.0,0.0);
   glVertex3f(p1.x-d*volume.dx,p1.y-d*volume.dy,p1.z-d*volume.dz);
   glVertex3f(p1.x-d*volume.dx,p2.y            ,p1.z-d*volume.dz);
   glColor3f(0.0,0.0,1.0);
   glVertex3f(p1.x-d*volume.dx,p1.y-d*volume.dy,p1.z-d*volume.dz);
   glVertex3f(p1.x-d*volume.dx,p1.y-d*volume.dy,p2.z            );
   glEnd(); 

   /* Volume axes 
   glBegin(GL_LINES);
   glColor3f(1.0,0.0,0.0);
   glVertex3f(-d*volume.dx,-d*volume.dy,-d*volume.dz);
   glVertex3f(volume.nx*volume.dx,-d*volume.dy,-d*volume.dz);
   glColor3f(0.0,1.0,0.0);
   glVertex3f(-d*volume.dx,-d*volume.dy,-d*volume.dz);
   glVertex3f(-d*volume.dx,volume.ny*volume.dy,-d*volume.dz);
   glColor3f(0.0,0.0,1.0);
   glVertex3f(-d*volume.dx,-d*volume.dy,-d*volume.dz);
   glVertex3f(-d*volume.dx,-d*volume.dy,volume.nz*volume.dz);
   glEnd(); */
}

/*          
   Draw the planes
   The order and orientation is determined by the camera position
   Drawing is done back to front
   It is assumed that the textures are set up for each slice
*/       
void DrawPlanes(void)
{        
   int i=0,j=0,k=0,ii=0,jj=0,kk=0;
   int axis = 0;
   int dx,dy,dz;
   XYZ cp,p[4]; 
         
   if (!options.showtextures && !options.showdecoration)
      return;

   /* Vector from center of inspection box to camera - determines quadrant */
   cp.x = camera.vp.x - (volume.x + (camera.x0 + camera.xdim/2) * volume.dx);
   cp.y = camera.vp.y - (volume.y + (camera.y0 + camera.ydim/2) * volume.dy);
   cp.z = camera.vp.z - (volume.z + (camera.z0 + camera.zdim/2) * volume.dz);
         
   /* Which axis? */
   if (ABS(cp.x) >= ABS(cp.y) && ABS(cp.x) >= ABS(cp.z)) 
      axis = 1;
   if (ABS(cp.y) >= ABS(cp.x) && ABS(cp.y) >= ABS(cp.z))
      axis = 2;
   if (ABS(cp.z) >= ABS(cp.x) && ABS(cp.z) >= ABS(cp.y))
      axis = 3;
   if (axis == 0) 
      fprintf(stderr,"Camera axis = 0, shouldn't happen\n");

   if (options.showtextures) {
      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
		glColor3f(1.0,1.0,1.0);
   } else {
      glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
		glColor3f(0.5,0.5,0.5);
   }        
         
   /* If the mouse is down use a lower resolution version */
   dx = options.xskip;
   dy = options.yskip;
   dz = options.xskip;
   if (currentbutton == GLUT_LEFT_BUTTON || currentbutton == GLUT_RIGHT_BUTTON) {
      dx *= 2;
      dy *= 2;
      dz *= 2;
   }     

  	/* Draw x slices */
  	if (axis == 1) {
  	   for (i=0;i<volume.dim2;i+=dx) {
  	      ii = i;
  	      if (SIGN(cp.x) < 0)
  	         ii = volume.dim2 - dx - i;
  	      p[0].x = camera.x0 + ii; 
  	      if (p[0].x < 0 || p[0].x >= volume.nx)
  	         continue;      
			if (textures[ii].blank && options.showtextures)
				continue;
  	      p[0].y = camera.y0;
  	      p[0].z = camera.z0;
  	      p[1].x = p[0].x;
  	      p[1].y = camera.y0;
  	      p[1].z = camera.z0 + volume.dim2;
  	      p[2].x = p[0].x;
  	      p[2].y = camera.y0 + volume.dim2;
  	      p[2].z = camera.z0 + volume.dim2;
  	      p[3].x = p[0].x;
  	      p[3].y = camera.y0 + volume.dim2;
  	      p[3].z = camera.z0;
  	      if (options.showtextures) {
  	         glBindTexture(GL_TEXTURE_2D,textures[ii].id);
  	         glEnable(GL_TEXTURE_2D);
  	      } else {
				if (textures[ii].blank)
					glColor3f(0.5,1.0,0.5);
				else
					glColor3f(0.5,0.5,0.5);
			}
  	      glBegin(GL_QUADS);
  	      glTexCoord2f(0.0,0.0);
  	      glVertex3f(volume.x+p[0].x*volume.dx,volume.y+p[0].y*volume.dy,volume.z+p[0].z*volume.dz);
  	      glTexCoord2f(0.0,1.0);
  	      glVertex3f(volume.x+p[1].x*volume.dx,volume.y+p[1].y*volume.dy,volume.z+p[1].z*volume.dz);
  	      glTexCoord2f(1.0,1.0);
  	      glVertex3f(volume.x+p[2].x*volume.dx,volume.y+p[2].y*volume.dy,volume.z+p[2].z*volume.dz);
  	      glTexCoord2f(1.0,0.0);
  	      glVertex3f(volume.x+p[3].x*volume.dx,volume.y+p[3].y*volume.dy,volume.z+p[3].z*volume.dz);
  	      glEnd(); 
  	      if (options.showtextures)
  	         glDisable(GL_TEXTURE_2D);
  	   }
  	}

   /* Draw y slices */
  	if (axis == 2) {
  	   for (j=0;j<volume.dim2;j+=dy) {
  	      jj = j;
  	      if (SIGN(cp.y) < 0)
  	         jj = volume.dim2 - dy - j;
  	      p[0].y = camera.y0 + jj;
  	      if (p[0].y < 0 || p[0].y >= volume.ny)
  	         continue;
  	      if (textures[jj].blank && options.showtextures)
  	         continue;
  	      p[0].x = camera.x0;
  	      p[0].z = camera.z0;
  	      p[1].x = camera.x0;
  	      p[1].y = p[0].y;
  	      p[1].z = camera.z0 + volume.dim2;
  	      p[2].x = camera.x0 + volume.dim2;
  	      p[2].y = p[0].y;
  	      p[2].z = camera.z0 + volume.dim2;
  	      p[3].x = camera.x0 + volume.dim2;
  	      p[3].y = p[0].y;
  	      p[3].z = camera.z0;
  	      if (options.showtextures) {
  	         glBindTexture(GL_TEXTURE_2D,textures[jj].id);
  	         glEnable(GL_TEXTURE_2D);
  	      } else {
  	      	if (textures[jj].blank)
  	      	   glColor3f(0.5,1.0,0.5);
  	      	else
  	      	   glColor3f(0.5,0.5,0.5);
			}
  	      glBegin(GL_QUADS);
  	      glTexCoord2f(0.0,0.0);
  	      glVertex3f(volume.x+p[0].x*volume.dx,volume.y+p[0].y*volume.dy,volume.z+p[0].z*volume.dz);
  	      glTexCoord2f(0.0,1.0);
  	      glVertex3f(volume.x+p[1].x*volume.dx,volume.y+p[1].y*volume.dy,volume.z+p[1].z*volume.dz);
  	      glTexCoord2f(1.0,1.0);
  	      glVertex3f(volume.x+p[2].x*volume.dx,volume.y+p[2].y*volume.dy,volume.z+p[2].z*volume.dz);
  	      glTexCoord2f(1.0,0.0);
  	      glVertex3f(volume.x+p[3].x*volume.dx,volume.y+p[3].y*volume.dy,volume.z+p[3].z*volume.dz);
  	      glEnd(); 
  	      if (options.showtextures)
  	         glDisable(GL_TEXTURE_2D);
  	   }
  	}

  	/* Draw z slices */
  	if (axis == 3) {
  	   for (k=0;k<volume.dim2;k+=dy) {
  	      kk = k;
  	      if (SIGN(cp.z) < 0)
  	         kk = volume.dim2 - dy - kk;
  	      p[0].z = camera.z0 + kk;
  	      if (p[0].z < 0 || p[0].z >= volume.nz)
  	         continue;
  	      if (textures[kk].blank && options.showtextures)
  	         continue;
  	      p[0].x = camera.x0;
  	      p[0].y = camera.y0;
  	      p[1].x = camera.x0 + volume.dim2;
  	      p[1].y = camera.y0;
  	      p[1].z = p[0].z;
  	      p[2].x = camera.x0 + volume.dim2;
  	      p[2].y = camera.y0 + volume.dim2;
  	      p[2].z = p[0].z;
  	      p[3].x = camera.x0;
  	      p[3].y = camera.y0 + volume.dim2;
  	      p[3].z = p[0].z;
  	      if (options.showtextures) {
  	         glBindTexture(GL_TEXTURE_2D,textures[kk].id);
  	         glEnable(GL_TEXTURE_2D);
  	      } else {
  	      	if (textures[kk].blank)
  	      	   glColor3f(0.5,1.0,0.5);
  	      	else
  	      	   glColor3f(0.5,0.5,0.5);
			}
  	      glBegin(GL_QUADS);
  	      glTexCoord2f(0.0,0.0);
  	      glVertex3f(volume.x+p[0].x*volume.dx,volume.y+p[0].y*volume.dy,volume.z+p[0].z*volume.dz);
  	      glTexCoord2f(0.0,1.0);
  	      glVertex3f(volume.x+p[1].x*volume.dx,volume.y+p[1].y*volume.dy,volume.z+p[1].z*volume.dz);
  	      glTexCoord2f(1.0,1.0);
  	      glVertex3f(volume.x+p[2].x*volume.dx,volume.y+p[2].y*volume.dy,volume.z+p[2].z*volume.dz);
  	      glTexCoord2f(1.0,0.0);
  	      glVertex3f(volume.x+p[3].x*volume.dx,volume.y+p[3].y*volume.dy,volume.z+p[3].z*volume.dz);
  	      glEnd(); 
  	      if (options.showtextures)
  	         glDisable(GL_TEXTURE_2D);
  	   }
  	}
}

void DrawInfo(void)
{
	int y = 20;
	char s[64];

	glColor3f(1.0,1.0,1.0);

   sprintf(s,"Volume file name: %s\n",volume.name);
   DrawGLText(10,camera.screenheight - y,s);
   y += 13;
	sprintf(s,"Volume resolution: %d x %d x %d\n",volume.nx,volume.ny,volume.nz);
	DrawGLText(10,camera.screenheight - y,s);
	y += 13;
   sprintf(s,"Voxel dimensions: %g x %g x %g\n",volume.dx,volume.dy,volume.dz);
   DrawGLText(10,camera.screenheight - y,s);
   y += 13; 
   sprintf(s,"Voxel data: %d %s\n",volume.datasize/8,volume.datasize/8>1?"bytes":"byte");
   DrawGLText(10,camera.screenheight - y,s);
   y += 13;
   sprintf(s,"Volume position: (%g, %g, %g)\n",volume.x,volume.y,volume.z);
   DrawGLText(10,camera.screenheight - y,s);
   y += 13; 
   sprintf(s,"Inspection resolution: %d x %d x %d\n",camera.xdim,camera.ydim,camera.zdim);
   DrawGLText(10,camera.screenheight - y,s);
   y += 13;
   sprintf(s,"Inspection position: (%d, %d, %d)\n",camera.x0,camera.y0,camera.z0);
   DrawGLText(10,camera.screenheight - y,s);
   y += 13;
   sprintf(s,"Volume subsampling: %d x %d x %d\n",options.xskip,options.yskip,options.zskip);
   DrawGLText(10,camera.screenheight - y,s);
   y += 13;

	if (options.show3dcursor) {
   	sprintf(s,"Crosshair position: %d x %d x %d\n",(int)cursor3d.x,(int)cursor3d.y,(int)cursor3d.z);
   	DrawGLText(10,camera.screenheight - y,s);
   	y += 13;
	}
}

void Draw3dCursor(void)
{
	glBegin(GL_LINES);
	glColor3f(1.0,1.0,1.0);
	glVertex3f(cursor3d.x-volume.nx/16.0,cursor3d.y,cursor3d.z);
   glVertex3f(cursor3d.x+volume.nx/16.0,cursor3d.y,cursor3d.z);
   glVertex3f(cursor3d.x,cursor3d.y-volume.ny/16.0,cursor3d.z);
   glVertex3f(cursor3d.x,cursor3d.y+volume.ny/16.0,cursor3d.z);
   glVertex3f(cursor3d.x,cursor3d.y,cursor3d.z-volume.nx/16.0);
   glVertex3f(cursor3d.x,cursor3d.y,cursor3d.z+volume.nx/16.0);
	glEnd();
}

void DrawMarkers(void)
{
	int i;

	for (i=0;i<nmarkers;i++) {
		if (markers[i].n == 1) {
			glBegin(GL_POINTS);
			glColor3f(markers[i].c.r,markers[i].c.g,markers[i].c.b);
			glVertex3f(markers[i].p[0].x,markers[i].p[0].y,markers[i].p[0].z);
			glEnd();
		}
	}

   glBegin(GL_LINES);
   for (i=0;i<nmarkers;i++) {
      if (markers[i].n == 2) {
			glBegin(GL_POINTS);
         glColor3f(markers[i].c.r,markers[i].c.g,markers[i].c.b);
         glVertex3f(markers[i].p[0].x,markers[i].p[0].y,markers[i].p[0].z);
			glVertex3f(markers[i].p[1].x,markers[i].p[1].y,markers[i].p[1].z);
			glEnd();
      }
   }
   glEnd();
}

