#include "glvol.h"

#include "colourform.h"
#include "scaleform.h"
#include "opacityform.h"

extern FD_scaleform *scaleform;
extern FD_colourform *colourform;
extern FD_opacityform *opacityform;

extern RGBAI *colourmap;
extern VOLUME volume;
extern CAMERA camera;
extern OPTIONS options;
extern COLOURRAMP colourramp[11];

/*
	Update all the fields on all the forms 
*/
void UpdateForms(void)
{
	int i;
	static int first = TRUE;
	char s[64];

	if (first) {

		/* Scale form */
		sprintf(s,"%g",volume.dx);
      fl_set_input(scaleform->Dimx,s);
      sprintf(s,"%g",volume.dy);
      fl_set_input(scaleform->Dimy,s);
      sprintf(s,"%g",volume.dz);
      fl_set_input(scaleform->Dimz,s);
      sprintf(s,"%d",camera.xdim);
      fl_set_input(scaleform->Inspx,s);
      sprintf(s,"%d",camera.ydim);
      fl_set_input(scaleform->Inspy,s);
      sprintf(s,"%d",camera.zdim);
      fl_set_input(scaleform->Inspz,s);

		/* Opacity form */
		fl_set_button(opacityform->Constant,FALSE);
      fl_set_button(opacityform->Linear,TRUE);
      fl_set_button(opacityform->Power,FALSE);
      fl_set_button(opacityform->Gaussian,FALSE);
   	fl_set_input(opacityform->Scalefactor,"1.0");
      sprintf(s,"0");
      fl_set_input(opacityform->Linearmin,s);
      sprintf(s,"1");
      fl_set_input(opacityform->Linearmax,s);
      sprintf(s,"2");
      fl_set_input(opacityform->Powerexp,s);
      sprintf(s,"50");
      fl_set_input(opacityform->Gaussianmean,s);
      sprintf(s,"5");
      fl_set_input(opacityform->Gaussianstd,s);

		/* Colour form */
		fl_set_slider_bounds(colourform->Red,0.0,1.0);
      fl_set_slider_bounds(colourform->Green,0.0,1.0);
      fl_set_slider_bounds(colourform->Blue,0.0,1.0);
		fl_set_button(colourform->But0,FALSE);
      fl_set_button(colourform->But10,FALSE);
      fl_set_button(colourform->But20,FALSE);
      fl_set_button(colourform->But30,FALSE);
      fl_set_button(colourform->But40,FALSE);
      fl_set_button(colourform->But50,FALSE);
      fl_set_button(colourform->But60,FALSE);
      fl_set_button(colourform->But70,FALSE);
      fl_set_button(colourform->But80,FALSE);
      fl_set_button(colourform->But90,FALSE);
      fl_set_button(colourform->But100,FALSE);
		fl_set_input(colourform->Pos0,"0");
		fl_deactivate_object(colourform->Pos0);
      fl_set_input(colourform->Pos10,"10");
      fl_set_input(colourform->Pos20,"20");
      fl_set_input(colourform->Pos30,"30");
      fl_set_input(colourform->Pos40,"40");
      fl_set_input(colourform->Pos50,"50");
      fl_set_input(colourform->Pos60,"60");
      fl_set_input(colourform->Pos70,"70");
      fl_set_input(colourform->Pos80,"80");
      fl_set_input(colourform->Pos90,"90");
      fl_set_input(colourform->Pos100,"100");
		fl_deactivate_object(colourform->Pos100);
      fl_set_button(colourform->Act0,TRUE);
		fl_deactivate_object(colourform->Act0);
      fl_set_button(colourform->Act10,FALSE);
      fl_set_button(colourform->Act20,FALSE);
      fl_set_button(colourform->Act30,FALSE);
      fl_set_button(colourform->Act40,FALSE);
      fl_set_button(colourform->Act50,FALSE);
      fl_set_button(colourform->Act60,FALSE);
      fl_set_button(colourform->Act70,FALSE);
      fl_set_button(colourform->Act80,FALSE);
      fl_set_button(colourform->Act90,FALSE);
      fl_set_button(colourform->Act100,TRUE);
		fl_deactivate_object(colourform->Act100);
		
		/* Colour map 20 to 30 are used */
		for (i=0;i<=10;i++) 
			fl_mapcolor(20+i,(int)(colourramp[i].r*255),(int)(colourramp[i].g*255),(int)(colourramp[i].b*255));
		fl_set_object_color(colourform->Col0,20,0);
      fl_set_object_color(colourform->Col10,21,0);
      fl_set_object_color(colourform->Col20,22,0);
      fl_set_object_color(colourform->Col30,23,0);
      fl_set_object_color(colourform->Col40,24,0);
      fl_set_object_color(colourform->Col50,25,0);
      fl_set_object_color(colourform->Col60,26,0);
      fl_set_object_color(colourform->Col70,27,0);
      fl_set_object_color(colourform->Col80,28,0);
      fl_set_object_color(colourform->Col90,29,0);
      fl_set_object_color(colourform->Col100,30,0);
		fl_redraw_object(colourform->Col0);
		fl_redraw_object(colourform->Col100);
		fl_set_button(colourform->Opacityon,FALSE);

		first = FALSE;
	}

}

/*
	Create an opacity function from the opacity form
*/
void HandleOpacityForm(void)
{
	int i;
	double v,v1,v2;
	double scale;

	if (fl_get_button(opacityform->Constant)) {
		for (i=0;i<MAPPINGSIZE;i++) 
         colourmap[i].a = 1;
	} else if (fl_get_button(opacityform->Linear)) {
      v1 = atof(fl_get_input(opacityform->Linearmin));
      v2 = atof(fl_get_input(opacityform->Linearmax));
      for (i=0;i<MAPPINGSIZE;i++) 
         colourmap[i].a = v1 + (v2 - v1) * i / (double)MAPPINGSIZE;
   } else if (fl_get_button(opacityform->Power)) {
      v1 = atof(fl_get_input(opacityform->Powerexp));
      for (i=0;i<MAPPINGSIZE;i++) 
         colourmap[i].a = pow(i/(double)MAPPINGSIZE,v1);
   } else if (fl_get_button(opacityform->Gaussian)) {
      v1 = atof(fl_get_input(opacityform->Gaussianmean)); // mean
      v2 = atof(fl_get_input(opacityform->Gaussianstd)); // std
		v2 /= 100;
      for (i=0;i<MAPPINGSIZE;i++) {
			v = i / (double)MAPPINGSIZE - v1 / 100;
         colourmap[i].a = exp(-0.5 * v * v / (v2*v2));
      }
	}

	/* Scale factor - applies to all */
	scale = atof(fl_get_input(opacityform->Scalefactor));
	for (i=0;i<MAPPINGSIZE;i++)
		colourmap[i].a *= scale;

	UpdateColourMap();
}

/*
   Create an colour map from the colour ramp
*/
void MapFromRamp(int usealpha)
{
	int i,j;
	double pos[11],ptmp;
	double mu;
	RGBAI col[11],ctmp,cstart,cstop,black = {0,0,0};
	int istart,istop;
	int nnodes = 0;

	for (i=0;i<=10;i++) {
		if (colourramp[i].active) {
			col[nnodes].r = colourramp[i].r;
      	col[nnodes].g = colourramp[i].g;
      	col[nnodes].b = colourramp[i].b;
			col[nnodes].a = colourramp[i].a;
			pos[nnodes]   = colourramp[i].pos;
			nnodes++;
		}
	}
	if (options.debug) {
		for (i=0;i<nnodes;i++) {
			fprintf(stderr,"%d %g %g %g %g %g\n",
				i,pos[i],col[i].r,col[i].g,col[i].b,col[i].a);
		}
	}

	/* Remove any duplicates */
	for (i=0;i<nnodes-1;i++) {
		for (j=i+1;j<nnodes;j++) {
			if (ABS(pos[i]-pos[j]) < 0.001) {
				pos[j] = pos[nnodes-1];
				col[j] = col[nnodes-1];
				nnodes--;	
				j--;
			}
		}
	}
	if (nnodes < 2) {
		fprintf(stderr,"Insufficient colour nodes, need at least 2\n");
		return;
	}

	/* Sort the list increasing */
	for (i=0;i<nnodes-1;i++) {
		for (j=i+1;j<nnodes;j++) {
			if (pos[i] > pos[j]) {
				ptmp = pos[i];
				pos[i] = pos[j];
				pos[j] = ptmp;
				ctmp = col[i];
				col[i] = col[j];
				col[j] = ctmp;
			}
		}
	}

	/* Finally create the colour ramp */
	for (i=0;i<MAPPINGSIZE;i++) {
		colourmap[i].r = black.r;
      colourmap[i].g = black.g;
      colourmap[i].b = black.b;
		if (usealpha)
			colourmap[i].a = 0;
	}
	for (j=0;j<nnodes;j++) {
		cstart = col[j];
		cstop  = col[j+1];
		istart = MAPPINGSIZE * pos[j] / 100.0;
		istop  = MAPPINGSIZE * pos[j+1] / 100.0;
		if (istart == istop)
			continue;
		for (i=istart;i<istop;i++) {
			mu = (i - istart) / (double)(istop - istart);
			colourmap[i].r = cstart.r + mu * (cstop.r - cstart.r);
         colourmap[i].g = cstart.g + mu * (cstop.g - cstart.g);
         colourmap[i].b = cstart.b + mu * (cstop.b - cstart.b);
			if (usealpha)
				colourmap[i].a = cstart.a + mu * (cstop.a - cstart.a);
		}
	}

	UpdateColourMap();
}

/*
	Read the colour form data 
*/
void ReadColourRamp(char *fname)
{
	int i;
   FILE *fptr;
   const char *fn;
	int onoff;
	double pos,r,g,b,a;
	char s[64];

   if (strlen(fname) < 1) {
      fn = fl_show_file_selector("Camera file","","","");
      if (fn == NULL || strlen(fn) <= 0)
         return;
      if ((fptr = fopen(fn,"r")) == NULL) {
         fprintf(stderr,"Unable to open camera file\n");
         return;
      }
   } else {
      if ((fptr = fopen(fname,"r")) == NULL) {
         fprintf(stderr,"Unable to open camera file\n");
         return;
      }
   }

	for (i=0;i<=10;i++) {
		if (fscanf(fptr,"%d %lf %lf %lf %lf %lf",&onoff,&pos,&r,&g,&b,&a) != 6) {
			fprintf(stderr,"Failed to correctly read ramp data\n");
			return;
		}
		colourramp[i].active = onoff;
		colourramp[i].pos = pos;
		colourramp[i].r = r;
		colourramp[i].g = g;
		colourramp[i].b = b;
		fl_mapcolor(20+i,(int)(r*255.0),(int)(g*255.0),(int)(b*255.0));

		switch (i) {
		case 0:
			fl_set_button(colourform->Act0,onoff);
			sprintf(s,"%g",pos);
			fl_set_input(colourform->Pos0,s);
			fl_set_object_color(colourform->Col0,20+i,0);
			fl_redraw_object(colourform->Col0);
			break;
      case 1:
         fl_set_button(colourform->Act10,onoff);
         sprintf(s,"%g",pos);
         fl_set_input(colourform->Pos10,s);
         fl_set_object_color(colourform->Col10,20+i,0);
         fl_redraw_object(colourform->Col10);
         break;
      case 2:
         fl_set_button(colourform->Act20,onoff);
         sprintf(s,"%g",pos);
         fl_set_input(colourform->Pos20,s);
         fl_set_object_color(colourform->Col20,20+i,0);
         fl_redraw_object(colourform->Col20);
         break;
      case 3:
         fl_set_button(colourform->Act30,onoff);
         sprintf(s,"%g",pos);
         fl_set_input(colourform->Pos30,s);
         fl_set_object_color(colourform->Col30,20+i,0);
         fl_redraw_object(colourform->Col30);
         break;
      case 4:
         fl_set_button(colourform->Act40,onoff);
         sprintf(s,"%g",pos);
         fl_set_input(colourform->Pos40,s);
         fl_set_object_color(colourform->Col40,20+i,0);
         fl_redraw_object(colourform->Col40);
         break;
      case 5:
         fl_set_button(colourform->Act50,onoff);
         sprintf(s,"%g",pos);
         fl_set_input(colourform->Pos50,s);
         fl_set_object_color(colourform->Col50,20+i,0);
         fl_redraw_object(colourform->Col50);
         break;
      case 6:
         fl_set_button(colourform->Act60,onoff);
         sprintf(s,"%g",pos);
         fl_set_input(colourform->Pos60,s);
         fl_set_object_color(colourform->Col60,20+i,0);
         fl_redraw_object(colourform->Col60);
         break;
      case 7:
         fl_set_button(colourform->Act70,onoff);
         sprintf(s,"%g",pos);
         fl_set_input(colourform->Pos70,s);
         fl_set_object_color(colourform->Col70,20+i,0);
         fl_redraw_object(colourform->Col70);
         break;
      case 8:
         fl_set_button(colourform->Act80,onoff);
         sprintf(s,"%g",pos);
         fl_set_input(colourform->Pos80,s);
         fl_set_object_color(colourform->Col80,20+i,0);
         fl_redraw_object(colourform->Col80);
         break;
      case 9:
         fl_set_button(colourform->Act90,onoff);
         sprintf(s,"%g",pos);
         fl_set_input(colourform->Pos90,s);
         fl_set_object_color(colourform->Col90,20+i,0);
         fl_redraw_object(colourform->Col90);
         break;
      case 10:
         fl_set_button(colourform->Act100,onoff);
         sprintf(s,"%g",pos);
         fl_set_input(colourform->Pos100,s);
         fl_set_object_color(colourform->Col100,20+i,0);
         fl_redraw_object(colourform->Col100);
         break;
		}
	}

	fclose(fptr);
}

/*
   Save the data for the colour form
*/
void SaveColourRamp(void)
{
	int i;
   FILE *fptr;
   const char *fn;
   char fname[128];

   fn = fl_show_file_selector("Colour ramp file","","","Untitled.ramp");
   if (fn == NULL || strlen(fn) <= 0)
      return;
   strcpy(fname,fn);
   if ((fptr = fopen(fname,"w")) == NULL) {
      fprintf(stderr,"Failed to open map file\n");
      return;
   }

	for (i=0;i<=10;i++) {
		fprintf(fptr,"%d %g %g %g %g %g\n",
			colourramp[i].active,colourramp[i].pos,
			colourramp[i].r,colourramp[i].b,colourramp[i].g,colourramp[i].a);
	}

   fclose(fptr);
}



