#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#if defined(__linux__)
   #include <GL/glut.h>
#elif defined(__APPLE__)
   #include <GLUT/glut.h>
#else
   #warning "Not sure where to find glut.h"
#endif
#include <sys/resource.h>
#include "paulslib.h"
#include "bitmaplib.h"
#include "opengllib.h"
#include "forms.h"

#define MAXTEXTURE 512
#define MAPPINGSIZE 65536

/* 1 == INTEL, 0 = Motorola/IBM/Mac */
#define WHATENDIAN 1

typedef struct {
	PIXELA *pixels;
	GLuint id;  // OpenGL texture ID
	int blank; // True if all pixels are transparent
} TEXTURE;

typedef struct {
	int active;      // true or false
	double pos;      // percentage
	double r,g,b,a;  // double versions for precision
} COLOURRAMP;

typedef struct {
	double r,g,b,a;
	PIXELA c;
} RGBAI;

typedef struct {
   XYZ vp;                /* View position           */
   XYZ vd;                /* View direction vector   */
   XYZ vu;                /* View up direction       */
   XYZ pr;                /* Point to rotate about   */
   double focallength;    /* Focal Length along vd   */
   double aperture;       /* Camera aperture         */
   double eyesep;         /* Eye separation          */
	double near,far;       /* Cutting plane distances */
	int stereo;				  /* Are we in stereo mode   */
   int screenwidth;       /* Screen dimensions       */
   int screenheight;      /*                         */
	int tracking;          /* Tracking inspection volume */
   int autorotate;
   XYZ autoscan;
   int x0,y0,z0;          /* Lower corner of inspection box, voxel indixes */
   int xdim,ydim,zdim;    /* Size of inspection box */
} CAMERA;

typedef struct {
	int debug;
	int record,windowdump; /* Image recording modes */
	int fullscreen;        /* Game mode or not */
	int showdecoration;
	int showtextures;
	int showhistogram;
	int showinfo;
	int show3dcursor;
	int showmarkers;
	int xskip,yskip,zskip; /* Subsampling */
	int animate;
	int tweenframes;
} OPTIONS;

typedef struct {
	char name[256];
	char comment[256];
	int nx,ny,nz;          /* Dimensions, voxel */
	double dx,dy,dz;       /* Cell size, real world coordinates */
	double x,y,z;          /* Corner of volume, real world coordinates */
	int datasize;          /* in Bytes */
	int dim2;              /* Largest power of 2 texture */
	unsigned char ***data1;
	short ***data2;
	int ***data3;
	int vmin,vmax;         /* Data bounds == histogram bounds */
} VOLUME;

typedef struct {
	int n;
	XYZ p[2];
	COLOUR c;
} MARKER;

#define NOSTEREO 0
#define ACTIVESTEREO 1
#define DUALSTEREO 2

void HandleDisplay(void);
void HandleKeyboard(unsigned char,int,int);
void HandleSpecialKeyboard(int,int,int);
void HandleMouse(int,int,int,int);
void HandleReshape(int,int);
void HandleMouseMotion(int,int);
void HandlePassiveMotion(int,int);
void HandlePassiveMotion(int,int);
void HandleIdle(void);
void RotateCamera(double,double,double);
void TranslateCamera(int,int);
void CameraHome(void);
void FocusCamera(void);
void CleanExit(void);
int  ReadVol(char *);
int  MallocTextures(void);
void LoadTextures(int);
void ReadMapFile(char *);
void ReadMarkers(char *);

/* Menus */
void MakeMenus(void);
void HandleMainMenu(int);
void HandleSubMenu(int);
void HandleDimMenu(int);
void HandleColourMapMenu(int);
void HandleOpacityMapMenu(int);
void HandleWindowMenu(int);

/* Misc */
void GiveUsage(char *);
void CreateEnvironment(void);
void HandleVisibility(int);
void VolMemory(int,int,int,int);
int  ReadColourMap(char *,int,RGBAI *);
void SaveColourMap(void);
int  ReadCameraFile(char *,CAMERA *);
void SaveCameraFile(void);
void SaveHistogram(void);
void UpdateColourMap(void);
int  CalcPower2(int,int,int);
void AnimateNextFrame(int);

/* Display */
void DrawAll(void);
void DrawHistogram(void);
void DrawDecoration(void);
void DrawPlanes(void);
void DrawInfo(void);
void Draw3dCursor(void);
void DrawMarkers(void);

/* Forms */
void UpdateForms();
void HandleOpacityForm(void);
void MapFromRamp(int);
void SaveColourRamp(void);
void ReadColourRamp(char *);

