#include "glvol.h"
#include "colourform.h"
#include "scaleform.h"
#include "opacityform.h"

extern RGBAI *colourmap;
extern CAMERA camera;
extern OPTIONS options;
extern FD_scaleform *scaleform;
extern int scaleformvisible;
extern FD_colourform *colourform;
extern int colourformvisible;
extern FD_opacityform *opacityform;
extern int opacityformvisible;

void MakeMenus(void)
{
	int mainmenu,submenu,colourmapmenu,opacitymapmenu;
	int windowmenu;

   windowmenu = glutCreateMenu(HandleWindowMenu);
   glutAddMenuEntry(" 160 x 120 (NTSC 1/16)",1);
   glutAddMenuEntry(" 192 x 144 (PAL 1/16)",2);
   glutAddMenuEntry(" 320 x 240 (NTSC 1/4)",3);
   glutAddMenuEntry(" 384 x 256 (Photo CD)",4);
   glutAddMenuEntry(" 384 x 288 (PAL 1/4)",5);
   glutAddMenuEntry(" 400 x 300",6);
   glutAddMenuEntry(" 400 x 400",7);
   glutAddMenuEntry(" 640 x 480 (Full NTSC)",8);
   glutAddMenuEntry(" 600 x 600",9);
   glutAddMenuEntry(" 720 x 576 (DVD PAL)",10);
   glutAddMenuEntry(" 768 x 576 (Full PAL)",11);
   glutAddMenuEntry(" 800 x 600",12);
   glutAddMenuEntry("1024 x 768",13);
   glutAddMenuEntry("1000 x 1000",14);

   colourmapmenu = glutCreateMenu(HandleColourMapMenu);
   glutAddMenuEntry("Colour designer",1);
	glutAddMenuEntry("Read from file",100);
	glutAddMenuEntry("Invert",101);
	glutAddMenuEntry("Flip",102);

   opacitymapmenu = glutCreateMenu(HandleOpacityMapMenu);
	glutAddMenuEntry("Opacity designer",1);
   glutAddMenuEntry("Read from File",100);
	glutAddMenuEntry("Invert",101);
	glutAddMenuEntry("Flip",102);

   submenu = glutCreateMenu(HandleSubMenu);
   glutAddMenuEntry("1 x 1 x 1 (1)",1);
   glutAddMenuEntry("2 x 2 x 2 (2)",2);
   glutAddMenuEntry("4 x 4 x 4 (3)",4);
   glutAddMenuEntry("8 x 8 x 8 (4)",8);

   mainmenu = glutCreateMenu(HandleMainMenu);
   glutAddMenuEntry("Toggle decorations (f1)",1);
	glutAddMenuEntry("Toggle textures (f2)",2);
	glutAddMenuEntry("Toggle histogram (f3)",3);
	glutAddMenuEntry("Toggle information (f4)",4);
	glutAddMenuEntry("Toggle tracking (f5)",5);
   glutAddMenuEntry("Toggle autorotate (f6)",6);
   glutAddMenuEntry("Toggle 3dcursor (f7)",7);
   glutAddMenuEntry("Toggle markers (f8)",8);
	glutAddSubMenu("Subsampling",submenu);
	glutAddSubMenu("Colour map",colourmapmenu);
   glutAddSubMenu("Opacity map",opacitymapmenu);
	glutAddMenuEntry("Voxel Options",9);
	glutAddMenuEntry("Read colour and opacity map",10);
	glutAddMenuEntry("Save colour and opacity map",11);
   glutAddMenuEntry("Save camera file",12);
   glutAddMenuEntry("Read camera file",13);
	glutAddMenuEntry("Save histogram",14);
	glutAddSubMenu("Window size",windowmenu);
   glutAddMenuEntry("Quit (esc)",20);

   glutAttachMenu(GLUT_RIGHT_BUTTON);
}

/*
   Handle the window menu
*/
void HandleWindowMenu(int whichone)
{
   switch (whichone) {
   case  1: camera.screenwidth =  160; camera.screenheight = 120; break;
   case  2: camera.screenwidth =  192; camera.screenheight = 144; break;
   case  3: camera.screenwidth =  320; camera.screenheight = 240; break;
   case  4: camera.screenwidth =  384; camera.screenheight = 256; break;
   case  5: camera.screenwidth =  384; camera.screenheight = 288; break;
   case  6: camera.screenwidth =  400; camera.screenheight = 300; break;
   case  7: camera.screenwidth =  400; camera.screenheight = 400; break;
   case  8: camera.screenwidth =  640; camera.screenheight = 480; break;
   case  9: camera.screenwidth =  600; camera.screenheight = 600; break;
   case 10: camera.screenwidth =  720; camera.screenheight = 576; break;
   case 11: camera.screenwidth =  768; camera.screenheight = 576; break;
   case 12: camera.screenwidth =  800; camera.screenheight = 600; break;
   case 13: camera.screenwidth = 1024; camera.screenheight = 768; break;
   case 14: camera.screenwidth = 1000; camera.screenheight = 1000; break;
   }
   glutReshapeWindow(camera.screenwidth,camera.screenheight);
}

/*
   Handle the main menu
*/
void HandleMainMenu(int whichone)
{
   switch (whichone) {
   case 1:
      options.showdecoration = !options.showdecoration;
      LoadTextures(TRUE);
      break;
   case 2:
      options.showtextures = !options.showtextures;
		LoadTextures(TRUE);
      break;
   case 3:
      options.showhistogram = !options.showhistogram;
      break;
	case 4:
		options.showinfo = !options.showinfo;
		break;
	case 5:
		camera.tracking = !camera.tracking;
		break;
   case 6:
      camera.autorotate = !camera.autorotate;
      break;
   case 7:
      options.show3dcursor = !options.show3dcursor;
      break;
   case 8:
      options.showmarkers = !options.showmarkers;
      break;
	case 9:
	   UpdateForms();
      fl_show_form(scaleform->scaleform,FL_PLACE_CENTER,TRUE,NULL);
      scaleformvisible = TRUE;
		break;
	case 10: // Read colour and opacity map file
		ReadColourMap("",0,colourmap);
		UpdateColourMap();
		LoadTextures(TRUE);
		break;
	case 11: // Save colour and opacity map file
		SaveColourMap();
		break;
   case 12: // Save camera file
		SaveCameraFile();
      break;
   case 13: // Read camera file
      ReadCameraFile("",&camera);
      MallocTextures();
      LoadTextures(TRUE);
      break;
   case 14: // Save histogram
      SaveHistogram();
      break;
   case 20:
      CleanExit();
      break;
   }
}

/*
   Handle the sub sampling menu
*/
void HandleSubMenu(int whichone)
{
	options.xskip = whichone;
   options.yskip = whichone;
   options.zskip = whichone;
	LoadTextures(TRUE);
}

/*
   Handle the colour map menu
	Leave the alpha channel alone
*/
void HandleColourMapMenu(int whichone)
{
	int i;
	RGBAI tmp;
	COLOUR c;

	if (whichone == 1) {
      UpdateForms();
      fl_show_form(colourform->colourform,FL_PLACE_CENTER,TRUE,NULL);
      colourformvisible = TRUE;
	} else if (whichone < 99) {
		for (i=0;i<MAPPINGSIZE;i++) {
			c = GetColour((double)i,0.0,(double)MAPPINGSIZE,whichone);
			colourmap[i].r = c.r;
         colourmap[i].g = c.g;
         colourmap[i].b = c.b;
		}
	} else if (whichone == 100) { // Read from file, ignore alpha
		ReadColourMap("",1,colourmap);
		UpdateColourMap();
   } else if (whichone == 101) { // invert
      for (i=0;i<MAPPINGSIZE;i++) {
         colourmap[i].r = 1 - colourmap[i].r;
         colourmap[i].g = 1 - colourmap[i].g;
         colourmap[i].b = 1 - colourmap[i].b;
      }
	} else if (whichone == 102) { // flip
      for (i=0;i<MAPPINGSIZE/2;i++) {
         tmp = colourmap[i];
         colourmap[i].r = colourmap[MAPPINGSIZE-1-i].r;
         colourmap[MAPPINGSIZE-1-i].r = tmp.r;
         colourmap[i].g = colourmap[MAPPINGSIZE-1-i].g;
         colourmap[MAPPINGSIZE-1-i].g = tmp.g;
         colourmap[i].b = colourmap[MAPPINGSIZE-1-i].b;
         colourmap[MAPPINGSIZE-1-i].b = tmp.b;
      }
	}

	UpdateColourMap();
   LoadTextures(TRUE);
}

/*
   Handle the opacity menu
*/
void HandleOpacityMapMenu(int whichone)
{
	int i;
   RGBAI tmp;

	switch (whichone) {
	case 1:
		UpdateForms();
		fl_show_form(opacityform->opacityform,FL_PLACE_CENTER,TRUE,NULL);
      opacityformvisible = TRUE;
		break;
	case 100: // Read from file
		ReadColourMap("",2,colourmap);
		UpdateColourMap();
		break;
   case 101: // invert
      for (i=0;i<MAPPINGSIZE;i++) {
         colourmap[i].a = 1 - colourmap[i].a;
      }
      break;
	case 102: // flip
		for (i=0;i<MAPPINGSIZE/2;i++) {
			tmp = colourmap[i];
         colourmap[i].a = colourmap[MAPPINGSIZE-1-i].a;
         colourmap[MAPPINGSIZE-1-i].a = tmp.a;
		}
		break;
	}

	UpdateColourMap();
   LoadTextures(TRUE);
}

