#include "geomviewer.h"

extern int ndometop,ndomebottom;
extern int ndomeleft,ndomeright;
extern DOMEFACE *dometop,*domebottom;
extern DOMEFACE *domeleft,*domeright;

extern BITMAP4 *texturetop,*texturebottom;
extern BITMAP4 *textureleft,*textureright;
extern GLuint textureid[4];

extern MESHNODE **mesh;
extern int meshtype,meshnx,meshny;

extern OPTIONS options;

/*
	Reset the contents of the texture image buffers
	in preparation for them being copied into texture objects
*/
void DomeTextures(void)
{
	BITMAP4 red = {255,0,0,255},blue = {0,0,255,255};
	BITMAP4 green = {0,255,0,255}, magenta = {255,0,255,255}; 

	if (texturetop != NULL)
		free(texturetop);
	texturetop = malloc(TEXTURESIZE*TEXTURESIZE*sizeof(BITMAP4));
	if (texturebottom != NULL)
		free(texturebottom);
   texturebottom = malloc(TEXTURESIZE*TEXTURESIZE*sizeof(BITMAP4));
	if (textureleft != NULL)
		free(textureleft);
   textureleft = malloc(TEXTURESIZE*TEXTURESIZE*sizeof(BITMAP4));
	if (textureright != NULL)
		free(textureright);
   textureright = malloc(TEXTURESIZE*TEXTURESIZE*sizeof(BITMAP4));
	if (texturetop == NULL || texturebottom == NULL || textureleft == NULL || textureright == NULL) { 
		fprintf(stderr,"Texture read failed\n");
		exit(-1);
	}

	/* Plain colours */
	Erase_Bitmap(texturetop,   TEXTURESIZE,TEXTURESIZE,blue);
   Erase_Bitmap(texturebottom,TEXTURESIZE,TEXTURESIZE,green);
   Erase_Bitmap(textureleft,  TEXTURESIZE,TEXTURESIZE,red);
   Erase_Bitmap(textureright, TEXTURESIZE,TEXTURESIZE,magenta);

}

/*
   Create a facet representation of the dome disk to a particular resolution
   Sets all the correct texture coordinates
   Deal with offset projection
   This is all done algorithmically rather than a final mathematical equation
   Dome disk is in the x-z plane at y=0
   The camera is lacated at (0,-1,0), see HandleDIsplay()
              | z
              |
              |
              |
              +--------- y
             /
            /
           / x
*/
int DomeCreate(void)
{
   int i,j;
   double xmin = 1e32,zmin = 1e32,xmax = -1e32,zmax = -1e32;
	
	/* Top face - just a triangle */
	dometop = (DOMEFACE *)realloc(dometop,sizeof(DOMEFACE));
	dometop[0].p[0].x = -sqrt(2.0) / 2.0;
   dometop[0].p[0].y = 0;
	dometop[0].p[0].z = 0.5;
	dometop[0].u[0] = 0;
	dometop[0].v[0] = 1;
   dometop[0].p[1].x = 0;
   dometop[0].p[1].y = sqrt(2.0) / 2.0;
   dometop[0].p[1].z = 0.5;
   dometop[0].u[1] = 0;
   dometop[0].v[1] = 0;
   dometop[0].p[2].x = sqrt(2.0) / 2.0;
   dometop[0].p[2].y = 0;
   dometop[0].p[2].z = 0.5;
   dometop[0].u[2] = 1;
   dometop[0].v[2] = 0;
	ndometop = 1;
 
	/* Bottom face - just a triangle */
   domebottom = (DOMEFACE *)realloc(domebottom,sizeof(DOMEFACE));
   domebottom[0].p[0].x = -sqrt(2.0) / 2.0;
   domebottom[0].p[0].y = 0;
   domebottom[0].p[0].z = -0.5;
   domebottom[0].u[0] = 1;
   domebottom[0].v[0] = 0;
   domebottom[0].p[1].x = sqrt(2.0) / 2.0;
   domebottom[0].p[1].y = 0;
   domebottom[0].p[1].z = -0.5;
   domebottom[0].u[1] = 0;
   domebottom[0].v[1] = 1;
   domebottom[0].p[2].x = 0;
   domebottom[0].p[2].y = sqrt(2.0) / 2.0;
   domebottom[0].p[2].z = -0.5;
   domebottom[0].u[2] = 0;
   domebottom[0].v[2] = 0;
	ndomebottom = 1;

	/* Left face - two triangles */
   domeleft = (DOMEFACE *)realloc(domeleft,2*sizeof(DOMEFACE));
   domeleft[0].p[0].x = -sqrt(2.0) / 2.0;
   domeleft[0].p[0].y = 0;
   domeleft[0].p[0].z = -0.5;
   domeleft[0].u[0] = 0;
   domeleft[0].v[0] = 0;
   domeleft[0].p[1].x = 0;
   domeleft[0].p[1].y = sqrt(2.0) / 2.0;
   domeleft[0].p[1].z = -0.5;
   domeleft[0].u[1] = 1;
   domeleft[0].v[1] = 0;
   domeleft[0].p[2].x = -sqrt(2.0) / 2.0;
   domeleft[0].p[2].y = 0;
   domeleft[0].p[2].z = 0.5;
   domeleft[0].u[2] = 0;
   domeleft[0].v[2] = 1;
   domeleft[1].p[0].x = -sqrt(2.0) / 2.0;
   domeleft[1].p[0].y = 0;
   domeleft[1].p[0].z = 0.5;
   domeleft[1].u[0] = 0;
   domeleft[1].v[0] = 1;
   domeleft[1].p[1].x = 0;
   domeleft[1].p[1].y = sqrt(2.0) / 2.0;
   domeleft[1].p[1].z = -0.5;
   domeleft[1].u[1] = 1;
   domeleft[1].v[1] = 0;
   domeleft[1].p[2].x = 0;
   domeleft[1].p[2].y = sqrt(2.0) / 2.0;
   domeleft[1].p[2].z = 0.5;
   domeleft[1].u[2] = 1;
   domeleft[1].v[2] = 1;
   ndomeleft = 2;

	/* Right face - two triangles */
   domeright  = (DOMEFACE *)realloc(domeright,2*sizeof(DOMEFACE));
   domeright[0].p[0].x = 0;
   domeright[0].p[0].y = sqrt(2.0) / 2.0;
   domeright[0].p[0].z = -0.5;
   domeright[0].u[0] = 0;
   domeright[0].v[0] = 0;
   domeright[0].p[1].x = sqrt(2.0) / 2.0;
   domeright[0].p[1].y = 0;
   domeright[0].p[1].z = -0.5;
   domeright[0].u[1] = 1;
   domeright[0].v[1] = 0;
   domeright[0].p[2].x = sqrt(2.0) / 2.0;
   domeright[0].p[2].y = 0;
   domeright[0].p[2].z = 0.5;
   domeright[0].u[2] = 1;
   domeright[0].v[2] = 1;
   domeright[1].p[0].x = 0;
   domeright[1].p[0].y = sqrt(2.0) / 2.0;
   domeright[1].p[0].z = -0.5;
   domeright[1].u[0] = 0;
   domeright[1].v[0] = 0;
   domeright[1].p[1].x = sqrt(2.0) / 2.0;
   domeright[1].p[1].y = 0;
   domeright[1].p[1].z = 0.5;
   domeright[1].u[1] = 1;
   domeright[1].v[1] = 1;
   domeright[1].p[2].x = 0;
   domeright[1].p[2].y = sqrt(2.0) / 2.0;
   domeright[1].p[2].z = 0.5;
   domeright[1].u[2] = 0;
   domeright[1].v[2] = 1;
	ndomeright = 2;

	/* 
		Refine the faces to the desired resolution 
		Achieved by splitting each triangles into 4
		Preserve texture coordinates (of course)
	*/
	for (i=0;i<5;i++) {
		dometop = (DOMEFACE *)realloc(dometop,(4*ndometop)*sizeof(DOMEFACE));
		DomeSplitFace(dometop,&ndometop);
		domebottom = (DOMEFACE *)realloc(domebottom,(4*ndomebottom)*sizeof(DOMEFACE));
      DomeSplitFace(domebottom,&ndomebottom);
		domeleft = (DOMEFACE *)realloc(domeleft,(4*ndomeleft)*sizeof(DOMEFACE));
      DomeSplitFace(domeleft,&ndomeleft);
		domeright = (DOMEFACE *)realloc(domeright,(4*ndomeright)*sizeof(DOMEFACE));
      DomeSplitFace(domeright,&ndomeright);
	}

	/* Turn into a hemisphere */
	for (j=0;j<3;j++) {
		for (i=0;i<ndometop;i++)
			Normalise(&(dometop[i].p[j]));
   	for (i=0;i<ndomebottom;i++)
         Normalise(&(domebottom[i].p[j]));
   	for (i=0;i<ndomeleft;i++)
         Normalise(&(domeleft[i].p[j]));
   	for (i=0;i<ndomeright;i++)
         Normalise(&(domeright[i].p[j]));
	}

	/* Apply offaxis shift 
	for (j=0;j<3;j++) {
      for (i=0;i<ndometop;i++) {
         dometop[i].p[j].x += camera.offset.x;
         dometop[i].p[j].y += camera.offset.y;
			dometop[i].p[j].z += camera.offset.z;
			Normalise(&(dometop[i].p[j]));
		}
      for (i=0;i<ndomebottom;i++) {
         domebottom[i].p[j].x += camera.offset.x;
         domebottom[i].p[j].y += camera.offset.y;
         domebottom[i].p[j].z += camera.offset.z;         
			Normalise(&(domebottom[i].p[j]));
		}
      for (i=0;i<ndomeleft;i++) {
         domeleft[i].p[j].x += camera.offset.x;
         domeleft[i].p[j].y += camera.offset.y;
         domeleft[i].p[j].z += camera.offset.z;
			Normalise(&(domeleft[i].p[j]));
		}
      for (i=0;i<ndomeright;i++) {
         domeright[i].p[j].x += camera.offset.x;
         domeright[i].p[j].y += camera.offset.y;
         domeright[i].p[j].z += camera.offset.z;         
			Normalise(&(domeright[i].p[j]));
		}
	} */

	/* Flatten onto xz plane */
	for (j=0;j<3;j++) {
      for (i=0;i<ndometop;i++) {
         DomeFlatten(&(dometop[i].p[j]),180.0);
         xmin = MIN(xmin,dometop[i].p[j].x);
         zmin = MIN(zmin,dometop[i].p[j].z);
         xmax = MAX(xmax,dometop[i].p[j].x);
         zmax = MAX(zmax,dometop[i].p[j].z);
      }
      for (i=0;i<ndomebottom;i++) {
         DomeFlatten(&(domebottom[i].p[j]),180.0);
         xmin = MIN(xmin,domebottom[i].p[j].x);
         zmin = MIN(zmin,domebottom[i].p[j].z);
         xmax = MAX(xmax,domebottom[i].p[j].x);
         zmax = MAX(zmax,domebottom[i].p[j].z);
      }
      for (i=0;i<ndomeleft;i++) {
         DomeFlatten(&(domeleft[i].p[j]),180.0);
         xmin = MIN(xmin,domeleft[i].p[j].x);
         zmin = MIN(zmin,domeleft[i].p[j].z);
         xmax = MAX(xmax,domeleft[i].p[j].x);
         zmax = MAX(zmax,domeleft[i].p[j].z);
      }  
      for (i=0;i<ndomeright;i++) {
         DomeFlatten(&(domeright[i].p[j]),180.0);
         xmin = MIN(xmin,domeright[i].p[j].x);
         zmin = MIN(zmin,domeright[i].p[j].z);
         xmax = MAX(xmax,domeright[i].p[j].x);
         zmax = MAX(zmax,domeright[i].p[j].z);
      }
	}

   // Fill in the warp map values
   CalcWarpData();

	return(TRUE);
}

/*
	Flatten a dome to a planar object
	It is the planar object that is finally rendered in orthographic mode
*/
void DomeFlatten(XYZ *f,double aperture)
{
	double phi,r;

	r   = atan2(sqrt(f->x*f->x + f->z*f->z),f->y) / (DTOR * aperture / 2);
	phi = atan2(f->z,f->x);
	
	f->x = r * cos(phi);
	f->y = 0;
	f->z = r * sin(phi);
}

/*
	Refine a triangular mesh by bisecting each edge
	Forms 3 new triangles for each existing triangle on each iteration
	Could be made more efficient for drawing if the triangles were
	ordered in a fan or strip!
*/
void DomeSplitFace(DOMEFACE *f,int *n)
{
	int i;
	int n1,n2;

	n1 = *n;
	n2 = *n;

	for (i=0;i<n1;i++) {

		f[n2].p[0] = MidPoint(f[i].p[0],f[i].p[1]);
      f[n2].p[1] = f[i].p[1];
      f[n2].p[2] = MidPoint(f[i].p[1],f[i].p[2]);
      f[n2].u[0] = LinearInterpolate(f[i].u[0],f[i].u[1],0.5);
      f[n2].u[1] = f[i].u[1];
      f[n2].u[2] = LinearInterpolate(f[i].u[1],f[i].u[2],0.5);
      f[n2].v[0] = LinearInterpolate(f[i].v[0],f[i].v[1],0.5);
      f[n2].v[1] = f[i].v[1];
      f[n2].v[2] = LinearInterpolate(f[i].v[1],f[i].v[2],0.5);

      f[n2+1].p[0] = MidPoint(f[i].p[1],f[i].p[2]);
      f[n2+1].p[1] = f[i].p[2];
      f[n2+1].p[2] = MidPoint(f[i].p[2],f[i].p[0]);
      f[n2+1].u[0] = LinearInterpolate(f[i].u[1],f[i].u[2],0.5);
      f[n2+1].u[1] = f[i].u[2];
      f[n2+1].u[2] = LinearInterpolate(f[i].u[2],f[i].u[0],0.5);
      f[n2+1].v[0] = LinearInterpolate(f[i].v[1],f[i].v[2],0.5);
      f[n2+1].v[1] = f[i].v[2];
      f[n2+1].v[2] = LinearInterpolate(f[i].v[2],f[i].v[0],0.5);

      f[n2+2].p[0] = MidPoint(f[i].p[0],f[i].p[1]);
      f[n2+2].p[1] = MidPoint(f[i].p[1],f[i].p[2]);
      f[n2+2].p[2] = MidPoint(f[i].p[2],f[i].p[0]);
      f[n2+2].u[0] = LinearInterpolate(f[i].u[0],f[i].u[1],0.5);
      f[n2+2].u[1] = LinearInterpolate(f[i].u[1],f[i].u[2],0.5);
      f[n2+2].u[2] = LinearInterpolate(f[i].u[2],f[i].u[0],0.5);
      f[n2+2].v[0] = LinearInterpolate(f[i].v[0],f[i].v[1],0.5);
      f[n2+2].v[1] = LinearInterpolate(f[i].v[1],f[i].v[2],0.5);
      f[n2+2].v[2] = LinearInterpolate(f[i].v[2],f[i].v[0],0.5);

      //f[i].p[0] = f[i].p[0];
      f[i].p[1] = MidPoint(f[i].p[0],f[i].p[1]);
      f[i].p[2] = MidPoint(f[i].p[0],f[i].p[2]);
      //f[i].u[0] = f[i].u[0];
      f[i].u[1] = LinearInterpolate(f[i].u[0],f[i].u[1],0.5);
      f[i].u[2] = LinearInterpolate(f[i].u[0],f[i].u[2],0.5);
      //f[i].v[0] = f[i].v[0];
      f[i].v[1] = LinearInterpolate(f[i].v[0],f[i].v[1],0.5);
      f[i].v[2] = LinearInterpolate(f[i].v[0],f[i].v[2],0.5);

		n2 += 3;
	}

	*n = n2;
}

/*
	Draw the dome
	Create display lists the first time through
*/
void DrawDome(int usetexture,int makelists)
{
	int i;
	static int first = TRUE;
	static int displaylist[4] = {-1,-1,-1,-1};

	if (makelists)
		first = TRUE;

	glDisable(GL_DEPTH);

	/* Top face */
	if (usetexture) {
  		glEnable(GL_TEXTURE_2D);
  		glBindTexture(GL_TEXTURE_2D,textureid[0]);
	}
	if (first) {
   	if (displaylist[0] > 0)
      	glDeleteLists(displaylist[0],1);
		displaylist[0] = glGenLists(1);
		glNewList(displaylist[0],GL_COMPILE);
		glBegin(GL_TRIANGLES);
		glNormal3f(0.0,1.0,0.0);
		for (i=0;i<ndometop;i++) 
			SetVertex(dometop[i]);
		glEnd();
		glEndList();
	} else {
		glCallList(displaylist[0]);
	}

	/* Bottom face */
	if (usetexture) 
  		glBindTexture(GL_TEXTURE_2D,textureid[1]);
  	if (first) {
      if (displaylist[1] > 0)
         glDeleteLists(displaylist[1],1);
  	   displaylist[1] = glGenLists(1);
  	   glNewList(displaylist[1],GL_COMPILE);     
  		glBegin(GL_TRIANGLES);
		glNormal3f(0.0,1.0,0.0);
  		for (i=0;i<ndomebottom;i++) 
			SetVertex(domebottom[i]);
		glEnd();
  	   glEndList();
	} else {
		glCallList(displaylist[1]);
	}

	/* Left face */
	if (usetexture) 
   	glBindTexture(GL_TEXTURE_2D,textureid[2]);
   if (first) {
      if (displaylist[2] > 0)
         glDeleteLists(displaylist[2],1);
      displaylist[2] = glGenLists(1);
      glNewList(displaylist[2],GL_COMPILE);
   	glBegin(GL_TRIANGLES);
		glNormal3f(0.0,1.0,0.0);
   	for (i=0;i<ndomeleft;i++) 
			SetVertex(domeleft[i]);
		glEnd();
      glEndList();
	} else {
		glCallList(displaylist[2]);
	}

	/* Right face */
	if (usetexture) 
  		glBindTexture(GL_TEXTURE_2D,textureid[3]);
  	if (first) {
      if (displaylist[3] > 0)
         glDeleteLists(displaylist[3],1);
  	   displaylist[3] = glGenLists(1);
  	   glNewList(displaylist[3],GL_COMPILE);
  		glBegin(GL_TRIANGLES);
		glNormal3f(0.0,1.0,0.0);
  		for (i=0;i<ndomeright;i++) 
			SetVertex(domeright[i]);
		glEnd();
  	   glEndList();
	} else {
		glCallList(displaylist[3]);
	}

   if (usetexture)
      glDisable(GL_TEXTURE_2D);
	first = FALSE;
	if (makelists)
		return;
}

void SetVertex(DOMEFACE face)
{
   int j;
   XYZ p;

	// If all intensities are negative, dont draw the triangle
	if (options.dometype == WARPMAP || options.dometype == MIRROR1) {
   	if (face.wi[0] < 0 || face.wi[1] < 0 || face.wi[2] < 0)
   	   return;
	}

   for (j=0;j<3;j++) {
      glTexCoord2f(face.u[j],face.v[j]);
      if (options.dometype == WARPMAP || options.dometype == MIRROR1) {
         if (options.showdomeintensity)
            glColor3f(face.wi[j],face.wi[j],face.wi[j]);
         p = face.wp[j];
         glVertex3f(p.x,p.y,p.z);
      } else {
         p = face.p[j];
         glVertex3f(p.x,p.y,p.z);
      }
   }
}

/*
   Create a dummy warping mesh, one that should have no effect
*/
void MakeDefaultMesh(void)
{
   int i,j;
   
   // Type and size
   meshtype = RECTANGULAR;
   meshnx = 1000;
   meshny = 1000;
   
   // Memory
   mesh = malloc(meshnx*sizeof(MESHNODE *));
   for (i=0;i<meshnx;i++)
      mesh[i] = malloc(meshny*sizeof(MESHNODE));
   
   // Contents
   for (j=0;j<meshny;j++) {
      for (i=0;i<meshnx;i++) {
         mesh[i][j].u = i / (double)(meshnx-1);
         mesh[i][j].v = j / (double)(meshny-1);
         mesh[i][j].x = 2 * i / (double)(meshnx-1) - 1;
         mesh[i][j].y = 2 * j / (double)(meshny-1) - 1;
         mesh[i][j].i = 1;
      }
   }
}

/*
   Create the additional warp mapping
   Fill in warp data in dometop, domebottom, domeleft, domeright
   Plane lies in the x-z plane
*/
void CalcWarpData(void)
{      
   int i,j;
   double x,z,b;
   
   for (i=0;i<ndometop;i++) {
      for (j=0;j<3;j++) {
         EstimateWarp(dometop[i].p[j].x,dometop[i].p[j].z,&x,&z,&b);
         dometop[i].wp[j].x = x;
         dometop[i].wp[j].y = 0;
         dometop[i].wp[j].z = z;
         dometop[i].wi[j] = b;
      }
   }
   
   for (i=0;i<ndomebottom;i++) {
      for (j=0;j<3;j++) {
         EstimateWarp(domebottom[i].p[j].x,domebottom[i].p[j].z,&x,&z,&b);
         domebottom[i].wp[j].x = x;
         domebottom[i].wp[j].y = 0;
         domebottom[i].wp[j].z = z;
         domebottom[i].wi[j] = b;
      }
   }

   for (i=0;i<ndomeleft;i++) {
      for (j=0;j<3;j++) {
         EstimateWarp(domeleft[i].p[j].x,domeleft[i].p[j].z,&x,&z,&b);
         domeleft[i].wp[j].x = x;
         domeleft[i].wp[j].y = 0;
         domeleft[i].wp[j].z = z;
         domeleft[i].wi[j] = b;
      }
   }
   
   for (i=0;i<ndomeright;i++) {
      for (j=0;j<3;j++) {
         EstimateWarp(domeright[i].p[j].x,domeright[i].p[j].z,&x,&z,&b);
         domeright[i].wp[j].x = x;
         domeright[i].wp[j].y = 0;
         domeright[i].wp[j].z = z;
         domeright[i].wi[j] = b;
      }
   }
}

/*
   Lookup inverse warp map, goes from fisheye mesh coordinates (x,y),
   treat them as (u,v), lookup the warp map (x,y).
   Original (ux,vy) ranges from -1 to 1
   (tu,tv) ranges from 0 to 1
   (ix iy) ranges from 0 to meshnx and meshny
*/
void EstimateWarp(double u,double v,double *x,double *z,double *br)
{
   int i,n,ix,iy;
   double tu,tv;
   double mu,longitude,latitude;
   XYZ p1,p,p0;
  
   double domeradius = 2.5;
   double mirrorradius = 0.3;
   XYZ mirrorpos = {2.4,0,0};
   XYZ projectorpos = {1.85,0,0};
   double aspectratio = 4/3.0;
   double throwdist = 1.5;
   XYZ frustum[4];
 
   if (options.dometype == MIRROR1) {

      // Location of projection plane, 1 unit in front of projector
      for (i=0;i<4;i++)
         frustum[i].x = projectorpos.x + 1;
      frustum[0].y = -0.5/throwdist;
      frustum[1].y = -0.5/throwdist;
      frustum[2].y = 0.5/throwdist;
      frustum[3].y = 0.5/throwdist;
      frustum[0].z = 0;
      frustum[1].z = 1.0/(throwdist*aspectratio);
      frustum[2].z = 1.0/(throwdist*aspectratio);
      frustum[3].z = 0;

      // longitude and latitude, latitude clampd at pi/2
      // latitude = 0 at north pole of dome, pi/2 at horizon of dome
      longitude = atan2(v,u); // -pi to pi
      latitude  = PID2 * sqrt(u*u + v*v);
      if (latitude > PID2)
         latitude = PID2;

      // p1 is the point on the dome, dome centered at the origin
      p1.x = domeradius * sin(latitude) * sin(longitude);
      p1.y = domeradius * sin(latitude) * cos(longitude);
      p1.z = domeradius * cos(latitude);
      
      // Calculate the reflection position on the dome
      if ((n = MirrorPosition(mirrorpos,mirrorradius,p1,projectorpos,&p0)) != 0) {
         fprintf(stderr,"Error %d, at (%g,%g) = (%g,%g)\n",n,u,v,latitude,longitude);
      }
 
      // Calculate the point "p" on the projection plane
      mu = (frustum[0].x - projectorpos.x) / (p0.x - projectorpos.x);
      p.x = frustum[0].x;
      p.y = projectorpos.y + mu * (p0.y - projectorpos.y);
      p.z = projectorpos.z + mu * (p0.z - projectorpos.z);

      // Set brigthness
      *br = 1;
      if (p1.x > 0)
         *br = 1 - p1.x / domeradius;
      if (*br < 0)
         *br = 0;
      if (p1.x >= mirrorpos.x)
         *br = -1;

      p.y *= throwdist;
      p.y *= 2;
      p.y *= aspectratio;
      p.z *= throwdist;
      p.z *= aspectratio;
      p.z *= 2;
      p.z -= 1;
      *x = p.y;
      *z = p.z;

	} else { // WARPMAP

   	tu = (u + 1) / 2.0;
   	if (tu < 0) tu = 0;
   	if (tu > 1) tu = 1;
   
   	tv = (v + 1) / 2.0;
   	if (tv < 0) tv = 0;
   	if (tv > 1) tv = 1;
   
   	ix = (int)(tu * (meshnx - 1));
   	iy = (int)(tv * (meshny - 1));
   
   	*x  = mesh[ix][iy].x;
   	*z  = mesh[ix][iy].y;
   	*br = mesh[ix][iy].i;
	}
}

/*
   Read a mesh file, this is an inverse mapping file
   See mirrorsim.c, and similar
   First line is the dimensions
   Subsequent lines are the mesh nodes: u,v,x,y,i
*/
int ReadMesh(char *fname)
{  
   int i,j;
   double x,y,u,v,br;
   double xmin = 1e32,ymin = 1e32,xmax = -1e32,ymax = -1e32;
   FILE *fptr;
   
   if (strlen(fname) < 1)
      return(FALSE);
   
   // Attempt to open the file
   if ((fptr = fopen(fname,"r")) == NULL) {
      fprintf(stderr,"Failed to open inverse map file \"%s\"\n",fname);
      return(FALSE);
   }
   // Get the mesh type
   if (fscanf(fptr,"%d",&meshtype) != 1) {
      fprintf(stderr,"Failed to read the mesh type\n");
      fclose(fptr);
      return(FALSE);
   }
   if (meshtype != POLAR && meshtype != RECTANGULAR) {
      fprintf(stderr,"Failed to get a recognised map type (%d)\n",meshtype);
      fclose(fptr);
      return(FALSE);
   }
   if (meshtype != RECTANGULAR) {
      fprintf(stderr,"Currently only support a rectangular\n");
      fclose(fptr);
      return(FALSE);
   }

   // Get the dimensions
   if (fscanf(fptr,"%d %d",&meshnx,&meshny) != 2) {
      fprintf(stderr,"Failed to read the mesh dimensions\n");
      fclose(fptr);
      return(FALSE);
   }
   if (meshnx < 4 || meshny < 4 || meshnx > 100000 || meshny > 100000) {
      fprintf(stderr,"Didn't read acceptable mesh resolution (%d,%d)\n",meshnx,meshny);
      fclose(fptr);
      return(FALSE);
   }

   // Create new mesh
   mesh = (MESHNODE **)malloc(meshnx*sizeof(MESHNODE *));
   for (i=0;i<meshnx;i++)
      mesh[i] = (MESHNODE *)malloc(meshny*sizeof(MESHNODE));

   // Read the values
   for (j=0;j<meshny;j++) {
      for (i=0;i<meshnx;i++) {
         if (fscanf(fptr,"%lf %lf %lf %lf %lf",&u,&v,&x,&y,&br) != 5) {
            fprintf(stderr,"Unexpected end of mesh file encountered, fatal\n");
            fclose(fptr);
            exit(-1);
         }
         xmin = MIN(xmin,x);
         ymin = MIN(ymin,y);
         xmax = MAX(xmax,x);
         ymax = MAX(ymax,y);
         mesh[i][j].x = x;
         mesh[i][j].y = y;
         mesh[i][j].u = u;
         mesh[i][j].v = v;
         mesh[i][j].i = br;
			// Now use -1 to indicate not to draw the triangle
         //if (br < 0)
         //   mesh[i][j].i = 0;
         if (br > 1)
            mesh[i][j].i = 1;
      }
   }

   fclose(fptr);
   return(TRUE);
}

/*
   c  is the mirror position
   r is the mirror radius
   p1 is the position on the dome
   p2 is the projector position
*/ 
int MirrorPosition(XYZ c,double r,XYZ p1,XYZ p2,XYZ *p0)
{
   int i;
   int nres = 1000;
   double theta,thetamax,thetamin,phi;
   double ctheta,stheta;
   double l1,l2,val,valmin;
   XYZ p; 
   
   p0->x = 0;
   p0->y = 0;
   p0->z = 0;
   
   if (p1.z < 0)
      return(1);
      
   // Translate to origin
   p1.x -= c.x; 
   p2.x -= c.x;

   // Rotate about z axis so p1 lies in x-z plane
   phi = atan2(p1.y,p1.z);
   p1 = RotateX(p1,-phi);
   if (fabs(p1.y) > 0.00001)
      return(2);

   // Calculate the maximum possible value of theta
   // Mirror is at origin
   if (fabs(p2.x) <= r)
      return(3);
   thetamax = acos(r/fabs(p2.x));

   // Iterate to find the minimum light path length
   thetamin = thetamax;
   valmin = 1e32;
   for (i=0;i<=nres;i++) {
      theta = i * PID2 / (double)nres;
      if (theta >= thetamax)
         break;
      ctheta = cos(theta);
      stheta = sin(theta);
      l1 = (p2.x + r*ctheta)*(p2.x + r*ctheta) + (r*stheta)*(r*stheta);
      l2 = (p1.x + r*ctheta)*(p1.x + r*ctheta) + (p1.z - r*stheta)*(p1.z - r*stheta);
      val = sqrt(l1) + sqrt(l2);
      if (val < valmin) {
         valmin = val;
         thetamin = theta;
      }
   }
   //fprintf(stderr,"%d %g %g (%g,%g,%g) %g\n",i,thetamin,thetamax,p1.x,p1.y,p1.z,p2.x);
   if (thetamin < 0)
      return(4);

   // Calculate p0
   p.x = -r * cos(thetamin);
   p.y = 0;
   p.z = r * sin(thetamin);

   // Rotate back
   p = RotateX(p,phi);

   // Translate back
   p.x += c.x;

   *p0 = p;

   return(0);
}

