#include "geomviewer.h"
#include "keyboardstring.h"

/*
   Third upgrade to "glstereo", now geomviewer

	Mar 2005  Split with stereo2, added dome support
	Dec 2005  Added support for non square textures but removed it because gl_repeat isn't supported.
	Jan 2006  Added GLUT_MULTISAMPLE
             Removed grey background for warped meshes.
				 Home honours view file
*/

// Forms library 
#include "cameraform.c"
#include "materialform.c"
#include "lightform.c"
#include "appearform.c"
FD_cameraform *cameraform;
FD_materialform *materialform;
FD_lightform *lightform;
FD_appearform *appearform;
int cameraformvisible   = FALSE;
int materialformvisible = FALSE;
int lightformvisible    = FALSE;
int appearformvisible   = FALSE;

// Options
OPTIONS options;

// Interface state
INTERFACESTATE interfacestate;

/* Geometry Structure */
int nball    = 0; BALL    *ball    = NULL;
int nballt   = 0; BALLT   *ballt   = NULL;
int ndisk    = 0; DISK    *disk    = NULL;
int ncone    = 0; CONE    *cone    = NULL;
int ndot     = 0; DOT     *dot     = NULL;
int nline    = 0; LINE    *line    = NULL;
int nface3   = 0; FACE3   *face3   = NULL;
int nface4   = 0; FACE4   *face4   = NULL;
int nface4t  = 0; FACE4T  *face4t  = NULL;
int nlabel   = 0; LABEL   *label   = NULL;
XYZ pmin={-1.0,-1.0,-1.0},pmax={1.0,1.0,1.0},pmid={0.0,0.0,0.0};
double rangemin=1e32,rangemax=-1e32;

// Camera 
CAMERA camera;

// Lights as provided in the scene file 
#define MAXLIGHT 8
int nlight = 0;
ALIGHT lights[MAXLIGHT];

// Global lights, set through the light form 
GLfloat globalambient[4] = {0.3,0.3,0.3,1.0};
int defaultlights = 2;
XYZ deflightpos[8] = {{1,1,1}, {0,0,0}, {0,1,1}, {1,0,0}, 
                      {0,0,1}, {1,1,0}, {1,0,1}, {0,1,0}};

// Global Materials, set through the materials form 
GLfloat specularcolour[4] = {1.0,1.0,1.0,1.0};
GLfloat shininess[1]      = {40.0};
GLfloat emission[4]       = {0.0,0.0,0.0,1.0};
GLfloat transparency      = 1.0;

// Coordinate and texture coordinates for the 4 sides
int ndometop = 0, ndomebottom = 0;
int ndomeleft = 0, ndomeright = 0;
DOMEFACE *dometop = NULL,*domebottom = NULL;
DOMEFACE *domeleft = NULL,*domeright = NULL;

// Texture maps for the 4 walls
BITMAP4 *texturetop = NULL,*texturebottom = NULL;
BITMAP4 *textureleft = NULL,*textureright = NULL;
GLuint textureid[4];

// Optional mapper
MESHNODE **mesh = NULL;
int meshtype=0,meshnx=0,meshny=0;

int main(int argc,char **argv)
{
   int i,gotdata = FALSE;
	char offname[64],geomname[64],viewname[64],autopilotname[64];
	char meshfile[128] = "";
	COLOUR black = {0,0,0};
	XYZ unity = {1,1,1},zero = {0,0,0};

	// Defalt options
	options.debug             = FALSE;
	options.stereo            = NOSTEREO;   
	options.fullscreen        = FALSE;
	options.dometype          = NODOME;
	options.showdomeintensity = TRUE;
	options.rendermode        = WIREFRAME;  
	options.showtextures      = TRUE;
	options.showboundingbox   = FALSE;
	options.recordimages      = FALSE;
	options.windowdump        = 0;
	options.makeqtvr          = FALSE;
	options.showhelpinfo      = FALSE;
	options.projectiontype    = PERSPECTIVE;
	options.selectmode        = FALSE;
	options.targetfps         = 30;
	options.interaction       = INSPECT;
	options.autopilot         = FALSE;
	options.autospin          = zero;  // 0, 1, or -1 for still, left, right 
	options.interactspeed     = 1;     // Scale factor on deltamove
	options.deltamove         = -1;    // Movement/rotation increment 
	options.background        = black;
	options.sphereresolution  = 16;
	options.doantialias       = FALSE;
	options.pointscale        = 1;
	options.linescale         = 1;
	options.scalefactor       = unity;
	options.localviewer       = TRUE;   
	options.twosided          = FALSE;  
	options.faceorder         = GL_CCW;
	options.texturetype       = GL_TEXTURE_2D;
	camera.fishrotate         = 0;

	// Interface default
	interfacestate.currentbutton  = -1;
	interfacestate.mousexlast     = -1;
	interfacestate.mouseylast     = -1;
	interfacestate.framerate      = 0;

	offname[0] = '\0';
	geomname[0] = '\0';
	viewname[0] = '\0';
   for (i=1;i<argc;i++) {
      if (strcmp(argv[i],"-h") == 0) 
         GiveUsage(argv[0]);
		if (strcmp(argv[i],"-s") == 0)
			options.stereo = ACTIVESTEREO;
		if (strcmp(argv[i],"-ss") == 0) 
         options.stereo = DUALSTEREO;
      if (strcmp(argv[i],"-f") == 0)
         options.fullscreen = TRUE;
		if (strcmp(argv[i],"-fly") == 0)
			options.interaction = FLY;
		if (strcmp(argv[i],"-walk") == 0)
			options.interaction = WALK;
      if (strcmp(argv[i],"-d") == 0)
         options.debug = TRUE;
		if (strcmp(argv[i],"-as") == 0) {
			options.autospin.x = 1; 
			options.autospin.y = 0; 
			options.autospin.z = 0;
		}
      if (strcmp(argv[i],"-cw") == 0)
         options.faceorder = GL_CW;

		// Fisheye/dome projection
		if (strcmp(argv[i],"-fish") == 0) {
			i++;
         options.dometype = atof(argv[i]);
			options.projectiontype = FISHEYE;
			if (options.dometype < TRUNCBOTTOM || options.dometype > MIRROR1) {
				fprintf(stderr,"Unrecognised dome type\n");
				options.dometype = NODOME;
				options.projectiontype = PERSPECTIVE;
			}
		}
		if (strcmp(argv[i],"-fishr") == 0) {
         i++;
         camera.fishrotate = DTOR*atof(argv[i]);
		}

      // Read warp file
      if (strcmp(argv[i],"-m") == 0) {
         i++;
         strcpy(meshfile,argv[i]);
			options.dometype = 5;
			options.projectiontype = FISHEYE;
      }

      if (strcmp(argv[i],"-fps") == 0) {
         i++;
         options.targetfps = atof(argv[i]);
      }
		if (strcmp(argv[i],"-dm") == 0) {
         i++;
         options.deltamove = atof(argv[i]);
      }
		if (strcmp(argv[i],"-sh") == 0) {
         i++;
			shininess[0] = atof(argv[i]);	
		}
		if (strcmp(argv[i],"-ap") == 0) {
			options.autopilot = TRUE;
			i++;
			strcpy(autopilotname,argv[i]);
		}
      if (strcmp(argv[i],"-rw") == 0)
         options.rendermode = WIREFRAME;
      if (strcmp(argv[i],"-rf") == 0)
         options.rendermode = SHADE_FLAT;
      if (strcmp(argv[i],"-rd") == 0)
         options.rendermode = SHADE_DIFFUSE;
      if (strcmp(argv[i],"-rs") == 0)
         options.rendermode = SHADE_SPECULAR;
		if (strcmp(argv[i],"-og") == 0) {
			i++;
			strcpy(geomname,argv[i]);
		}
      if (strcmp(argv[i],"-oo") == 0) {
         i++;
         strcpy(offname,argv[i]);
      }
      if (strcmp(argv[i],"-vf") == 0) {
         i++;
         strcpy(viewname,argv[i]);
      }
		if (strcmp(argv[i],"-bb") == 0)
			options.showboundingbox = !options.showboundingbox;
		if (strcmp(argv[i],"-bg") == 0) {
			i++;
			options.background.r = atof(argv[i++]);
			options.background.g = atof(argv[i++]);
			options.background.b = atof(argv[i]);
		}
		if (strcmp(argv[i],"-sr") == 0) {
         i++;
         options.sphereresolution = atoi(argv[i]);
		}
      if (strcmp(argv[i],"-sc") == 0) {
			i++;
			options.scalefactor.x = atof(argv[i]);
			i++;
         options.scalefactor.y = atof(argv[i]);
         i++;
         options.scalefactor.z = atof(argv[i]);
		}
   }

	// Check fisheye vs stereo
	if (options.stereo != NOSTEREO && options.dometype != NODOME) {
		fprintf(stderr,"Stereo and fisheye are incompatable, turning both off\n");
		options.stereo = NOSTEREO;
		options.dometype = NODOME;
	}

   // Setup, initialise OpenGL and GLUT
   if (options.debug)
      fprintf(stderr,"Initialising glut\n");
   glutInit(&argc,argv);
   CreateOpenGL();
   CreateMenus();

	// Can we handle non power of 2 textures?
   #if GL_EXT_texture_rectangle
      options.texturetype = GL_TEXTURE_RECTANGLE_EXT;
   #else
      #if GL_NV_texture_rectangle
         options.texturetype = GL_TEXTURE_RECTANGLE_NV;
      #else
         options.texturetype = GL_TEXTURE_2D;
      #endif
   #endif
   if (options.debug) {
		if (options.texturetype == GL_TEXTURE_2D)
      	fprintf(stderr,"Non power of 2 textures NOT supported\n");
   	else
      	fprintf(stderr,"Non power of 2 textures supported\n");
	}

	// Deal with dome projection
  	// Possibly read the mesh file, for warped dome projection
  	if (!ReadMesh(meshfile)) {
  	   if (options.debug)
  	      fprintf(stderr,"Creating default mesh\n");
  	   MakeDefaultMesh();
  	}
	if (options.debug)
		fprintf(stderr,"Creating the dome\n");
  	if (!DomeCreate()) {
  	   fprintf(stderr,"Failed to create dome geometry\n");
  	   CleanExit();
  	}

	// Create the texture buffers and default images
	if (options.debug)
		fprintf(stderr,"Creating dome textures\n");
	DomeTextures();

  	// Texture IDs for each face
	if (options.debug)
      fprintf(stderr,"Binding textures\n");
  	glGenTextures(4,textureid);
  	if (textureid[0] < 1 || textureid[1] < 1 || textureid[2] < 1 || textureid[3] < 1) {
  	   fprintf(stderr,"Received bad texture id <%d,%d,%d,%d>\n",
  	      (int)textureid[0],(int)textureid[1],(int)textureid[2],(int)textureid[3]);
  	}
  	for (i=0;i<4;i++) {
		if (options.debug)
			fprintf(stderr,"%d\n",i);
  	   glBindTexture(GL_TEXTURE_2D,textureid[i]);
  	   glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
  	   glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
  	   glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  	   glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  	   glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
  	   switch (i) {
  	   case 0:
  	      glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,TEXTURESIZE,TEXTURESIZE,0,GL_RGBA,GL_UNSIGNED_BYTE,texturetop);
  	      break;
  	   case 1:
  	      glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,TEXTURESIZE,TEXTURESIZE,0,GL_RGBA,GL_UNSIGNED_BYTE,texturebottom);
  	      break;
  	   case 2:
  	      glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,TEXTURESIZE,TEXTURESIZE,0,GL_RGBA,GL_UNSIGNED_BYTE,textureleft);
  	      break;
  	   case 3:
  	      glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,TEXTURESIZE,TEXTURESIZE,0,GL_RGBA,GL_UNSIGNED_BYTE,textureright);
  	   }
  	}

   // Initialise and setup the forms 
	if (options.debug)
		fprintf(stderr,"Initialising forms library\n");
   i = 1;
   fl_initialize(&i,argv,"Geom File Viewer",0,0);
   cameraform   = create_form_cameraform();
	materialform = create_form_materialform();
	lightform    = create_form_lightform();
	appearform   = create_form_appearform();

	/* 
		Initial model if there is one 
		If none then create a 3D plus sign
	*/
	if (options.debug)
		fprintf(stderr,"Sorting out geometry\n");
	DeleteGeometry();
	if (strlen(geomname) > 0) 
      gotdata = ReadGeometry(geomname);
	if (strlen(offname) > 0) 
      gotdata = ReadOFF(offname);
	if (!gotdata) {
		CreateAxes();
		gotdata = TRUE;
	}
	CleanGeometry();

	// Initialise things before starting 
	CameraHome(HOME);
	options.camerahome = camera;
	if (strlen(viewname) > 0)
  		ReadViewFile(viewname);
	MakeGeometry(TRUE);
	if (options.autopilot) 
		options.autopilot = AutoPilot(0,autopilotname);
	UpdateForm(); 

	// Lets go, we're ready for event handling 
	if (options.debug)
		fprintf(stderr,"Starting\n");
   glutMainLoop();
   return(0);
}

/*
   This is where OpenGL and GLUT are initialised
   Global/Initial OpenGL settings are made
*/
void CreateOpenGL(void)
{
	GLint num[10];
	GLfloat fsizes[2],fstep;

	// Initialise GLUT 
   if (options.stereo == ACTIVESTEREO)
		glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_STEREO | GLUT_MULTISAMPLE);
	else
      ;glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_MULTISAMPLE);
   glutCreateWindow("Geom File Viewer");
   options.screenwidth  = 800;
   options.screenheight = 600;
	if (options.stereo == DUALSTEREO)
		options.screenwidth *= 2;
   if (options.fullscreen)
      glutFullScreen();
	else
		glutReshapeWindow(options.screenwidth,options.screenheight);
   glutReshapeFunc(HandleReshape);
	glutDisplayFunc(HandleDisplay); 
   glutVisibilityFunc(HandleVisibility);
   glutKeyboardFunc(HandleKeyboard);
	glutIdleFunc(HandleIdle);
   glutSpecialFunc(HandleSpecialKeyboard);
   glutMouseFunc(HandleMouse);
   glutMotionFunc(HandleMouseMotion);
   glutPassiveMotionFunc(HandlePassiveMotion);
	if (options.stereo == DUALSTEREO)
   	glutSetCursor(GLUT_CURSOR_NONE);

	// glEnable(GL_NORMALIZE); Do this ourselves 
   glEnable(GL_DEPTH_TEST);
   glEnable(GL_BLEND);
   glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_DITHER);
   glDisable(GL_LINE_SMOOTH);
   glDisable(GL_POINT_SMOOTH);
   glDisable(GL_POLYGON_SMOOTH);
	glDisable(GL_CULL_FACE);

   glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);
   glClearColor(options.background.r,options.background.g,options.background.b,0.0);
	glPixelStorei(GL_UNPACK_ALIGNMENT,1);
	glFrontFace(options.faceorder);

	// Test miscellaneous aspects of the OpenGL state 
	if (options.debug) {
		fprintf(stderr,"Screen size: %d x %d\n",glutGet(GLUT_SCREEN_WIDTH),glutGet(GLUT_SCREEN_HEIGHT));

		glGetFloatv(GL_POINT_SIZE_RANGE,fsizes);
		glGetFloatv(GL_POINT_SIZE_GRANULARITY,&fstep);
		fprintf(stderr,"Point sizes supported: %g to %g in steps of %g\n",fsizes[0],fsizes[1],fstep);
      glGetFloatv(GL_LINE_WIDTH_RANGE,fsizes);
      glGetFloatv(GL_LINE_WIDTH_GRANULARITY,&fstep);
      fprintf(stderr,"Line widths supported: %g to %g in steps of %g\n",fsizes[0],fsizes[1],fstep);

		glGetIntegerv(GL_MAX_TEXTURE_SIZE,num); 
		fprintf(stderr,"Maximum texture size: %d\n",(int)num[0]);
      glGetIntegerv(GL_MAX_PROJECTION_STACK_DEPTH,num);
		fprintf(stderr,"Projection stack depth: %d\n",(int)num[0]);
		glGetIntegerv(GL_MAX_MODELVIEW_STACK_DEPTH,num);
		fprintf(stderr,"ModelView stack depth: %d\n",(int)num[0]);
		glGetIntegerv(GL_MAX_CLIP_PLANES,num);
      fprintf(stderr,"Number of clipping planes: %d\n",(int)num[0]);
		glGetIntegerv(GL_DEPTH_BITS,num);
		fprintf(stderr,"Depth buffer \"depth\": %d\n",(int)num[0]);
		glGetIntegerv(GL_MAX_VIEWPORT_DIMS,num);
		fprintf(stderr,"Maximum viewport size: %d x %d\n",(int)num[0],(int)num[1]);
		glGetIntegerv(GL_AUX_BUFFERS,num);
		fprintf(stderr,"Number of aucillary buffers: %d\n",(int)num[0]);
	}
	glGetIntegerv(GL_SAMPLE_BUFFERS,&(num[0]));
   glGetIntegerv(GL_SAMPLES,&(num[1]));
	if (options.debug) {
		fprintf(stderr,"Sample buffers: %d, samples: %d\n",(int)num[0],(int)num[1]);
		if (num[0] > 0 && num[1] > 1)
			fprintf(stderr,"Multisampling enabled\n");
		else
			fprintf(stderr,"Multisampling not enabled\n");
	}
	if (num[0] > 0 && num[1] > 1)
		glEnable(GL_MULTISAMPLE);
}

/*
   Handle a window reshape/resize
*/
void HandleReshape(int w,int h)
{
	options.screenwidth = w;
	options.screenheight = h;
	if (options.debug)
		fprintf(stderr,"Window reshape: %d x %d\n",options.screenwidth,options.screenheight);
}

/*
   This is the basic display callback routine 
	Called to redraw the buffers optionally in stereo
*/
void HandleDisplay(void)
{
	int i,j;
	static int framecount = -1;
	static double tstart;
	double tstop;
	double r,near,far;
   XYZ vp,vd,vr,vl;
   XYZ vright,vleft,vup;

	/* Set the time the first time around */
	if (framecount < 0) {
		framecount = 0;
      tstart = GetRunTime();
	}

   /* Are we in autopilot mode */
   if (options.autopilot) 
      options.autopilot = AutoPilot(1,"");

   /* Are we in autospin mode */
   if ((int)options.autospin.x != 0) 
      RotateCamera(options.autospin.x/5,0.0,0.0,KEYBOARDCONTROL);
	if ((int)options.autospin.y != 0)
      RotateCamera(0.0,options.autospin.y/5,0.0,KEYBOARDCONTROL);
   if ((int)options.autospin.z != 0)
      RotateCamera(0.0,0.0,options.autospin.z/5,KEYBOARDCONTROL);

   // Are we in flying mode 
	if (ABS(camera.speed) > EPSILON) 
      FlyCamera(camera.speed);

	if (options.stereo == ACTIVESTEREO || options.stereo == DUALSTEREO) {

      if (options.stereo == DUALSTEREO) {
			glDrawBuffer(GL_BACK);
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		}

      // Derive the two eye positions 
      CROSSPROD(camera.vd,camera.vu,vright);
      Normalise(&vright);
      vright.x *= camera.eyesep / 2;
      vright.y *= camera.eyesep / 2;
      vright.z *= camera.eyesep / 2;

   	glMatrixMode(GL_PROJECTION);
   	glLoadIdentity();
   	CreateProjection('r');
		if (options.stereo == ACTIVESTEREO) {
   		glViewport(0,0,options.screenwidth,options.screenheight);
		} else {
			glViewport(options.screenwidth/2,0,options.screenwidth/2,options.screenheight);
		}

   	glMatrixMode(GL_MODELVIEW);
      glDrawBuffer(GL_BACK_RIGHT);
		if (options.stereo == ACTIVESTEREO)
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glLoadIdentity();
      gluLookAt(camera.vp.x + vright.x,
					 camera.vp.y + vright.y,
					 camera.vp.z + vright.z,
					 camera.vp.x + vright.x + camera.vd.x,
					 camera.vp.y + vright.y + camera.vd.y,
					 camera.vp.z + vright.z + camera.vd.z,
                camera.vu.x,camera.vu.y,camera.vu.z);
      MakeLighting();
		MakeMaterial();
		MakeGeometry(FALSE);
		DrawExtras();
		
		/* ---- */

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      CreateProjection('l');
		if (options.stereo == ACTIVESTEREO) {
      	glViewport(0,0,options.screenwidth,options.screenheight);
		} else {
			glViewport(0,0,options.screenwidth/2,options.screenheight);
		}

      glMatrixMode(GL_MODELVIEW);
      glDrawBuffer(GL_BACK_LEFT);
		if (options.stereo == ACTIVESTEREO)
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glLoadIdentity();
      gluLookAt(camera.vp.x - vright.x,
					 camera.vp.y - vright.y,
					 camera.vp.z - vright.z,
                camera.vp.x - vright.x + camera.vd.x,
                camera.vp.y - vright.y + camera.vd.y,
                camera.vp.z - vright.z + camera.vd.z,
                camera.vu.x,camera.vu.y,camera.vu.z);
      MakeLighting();
		MakeMaterial();
		MakeGeometry(FALSE);
		DrawExtras();

	} else if (options.projectiontype == FISHEYE) {

   	// Calculate various view vectors 
   	vp = camera.vp;
   	vd = camera.vd;
		vr = CrossProduct(vd,camera.vu);
		vl = CrossProduct(camera.vu,vd);
		vd = ArbitraryRotate(vd,camera.fishrotate,vr);
   	vup = CrossProduct(vr,vd);
   	vright = VectorAdd(vr,vd);
	   vleft = VectorAdd(vl,vd);
   	near = VectorLength(pmin,pmax) / 100;
   	far  = MAX(camera.focallength,VectorLength(pmin,pmax)) * 20;

      // Left
      glDrawBuffer(GL_BACK);
      glReadBuffer(GL_BACK);
      glViewport(0,0,TEXTURESIZE,TEXTURESIZE);
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      gluPerspective(90.0,1.0,near,far);
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      gluLookAt(vp.x,vp.y,vp.z,vp.x+vleft.x,vp.y+vleft.y,vp.z+vleft.z,vup.x,vup.y,vup.z);
      MakeLighting();
      MakeMaterial();
      MakeGeometry(FALSE);
      glBindTexture(GL_TEXTURE_2D,textureid[2]);
      glCopyTexSubImage2D(GL_TEXTURE_2D,0,0,0,0,0,TEXTURESIZE,TEXTURESIZE);

		// Right
      glDrawBuffer(GL_BACK);
      glReadBuffer(GL_BACK);
      glViewport(0,0,TEXTURESIZE,TEXTURESIZE);
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      gluPerspective(90.0,1.0,near,far);
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      gluLookAt(vp.x,vp.y,vp.z,vp.x+vright.x,vp.y+vright.y,vp.z+vright.z,vup.x,vup.y,vup.z);
      MakeLighting();
      MakeMaterial();
      MakeGeometry(FALSE);
      glBindTexture(GL_TEXTURE_2D,textureid[3]);
      glCopyTexSubImage2D(GL_TEXTURE_2D,0,0,0,0,0,TEXTURESIZE,TEXTURESIZE);

      // Top 
      glDrawBuffer(GL_BACK);
      glReadBuffer(GL_BACK);
      glViewport(0,0,TEXTURESIZE,TEXTURESIZE);
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      gluPerspective(90.0,1.0,near,far);
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      gluLookAt(vp.x,vp.y,vp.z,vp.x+vup.x,vp.y+vup.y,vp.z+vup.z,-vright.x,-vright.y,-vright.z);
      MakeLighting();
      MakeMaterial();
      MakeGeometry(FALSE);
      glBindTexture(GL_TEXTURE_2D,textureid[0]);
      glCopyTexSubImage2D(GL_TEXTURE_2D,0,0,0,0,0,TEXTURESIZE,TEXTURESIZE);

      // Bottom 
      glDrawBuffer(GL_BACK);
      glReadBuffer(GL_BACK);
      glViewport(0,0,TEXTURESIZE,TEXTURESIZE);
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      gluPerspective(90.0,1.0,near,far);
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      gluLookAt(vp.x,vp.y,vp.z,vp.x-vup.x,vp.y-vup.y,vp.z-vup.z,-vleft.x,-vleft.y,-vleft.z);
      MakeLighting();
      MakeMaterial();
      MakeGeometry(FALSE);
      glBindTexture(GL_TEXTURE_2D,textureid[1]);
      glCopyTexSubImage2D(GL_TEXTURE_2D,0,0,0,0,0,TEXTURESIZE,TEXTURESIZE);

      // Remember the graphics state and return it at the end
      glPushAttrib(GL_ALL_ATTRIB_BITS);
      glDisable(GL_LIGHTING);
      glDisable(GL_ALPHA_TEST);
      glDisable(GL_COLOR_MATERIAL);
      glDisable(GL_DITHER);
      glDisable(GL_FOG);
      glDisable(GL_LINE_SMOOTH);
      glDisable(GL_LINE_STIPPLE);
      glDisable(GL_SCISSOR_TEST);
      glDisable(GL_STENCIL_TEST);

   	// Setup projections for the dome
   	glDrawBuffer(GL_BACK);
		if (options.dometype == WARPMAP)
   		glClearColor(0.0,0.0,0.0,0.0);
		else
			glClearColor(0.05,0.05,0.05,0.0);
   	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   	glMatrixMode(GL_PROJECTION);
   	glLoadIdentity();
   	glViewport(0,0,options.screenwidth,options.screenheight);
   	r = options.screenwidth / (double)options.screenheight;
   	switch (options.dometype) {
   	case TRUNCTOP:
   	case TRUNCBOTTOM:
   	   glOrtho(-r*0.75,r*0.75,-0.75,0.75,0.1,10.0);
   	   break;
   	case HSPHERICAL:
   	   glOrtho(-r*0.75,r*0.75,-0.75,0.75,0.1,10.0);
   	   break;
   	case VSPHERICAL:
   	case WARPMAP:
   	default:
   	   glOrtho(-r,r,-1.0,1.0,0.1,10.0);
   	   break;
   	}

	   // Create camera projection for dome
	   glMatrixMode(GL_MODELVIEW);
	   glLoadIdentity();
	   switch (options.dometype) {
	   case TRUNCBOTTOM:
	      gluLookAt(0.0,-1.0,0.25,0.0,0.0,0.25,0.0,0.0,1.0);
	      break;
	   case TRUNCTOP:
	      gluLookAt(0.0,-1.0,-0.25,0.0,0.0,-0.25,0.0,0.0,1.0);
	      break;
	   case HSPHERICAL:
	   case VSPHERICAL:
	   case WARPMAP:
	   default:
	      gluLookAt(0.0,-1.0,0.0,0.0,0.0,0.0,0.0,0.0,1.0);
	      break;
	   }

   	// Finally draw the dome geometry
   	glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
   	glColor3f(1.0,1.0,1.0);
  	   DrawDome(TRUE,FALSE);
		DrawExtras();
		glPopAttrib();

	} else {

   	glMatrixMode(GL_PROJECTION);
   	glLoadIdentity();
   	CreateProjection('c');
   	glViewport(0,0,options.screenwidth,options.screenheight);

      glMatrixMode(GL_MODELVIEW);
      glDrawBuffer(GL_BACK);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glLoadIdentity();
      gluLookAt(camera.vp.x,
					 camera.vp.y,
					 camera.vp.z,
					 camera.vp.x + camera.vd.x,
					 camera.vp.y + camera.vd.y,
					 camera.vp.z + camera.vd.z,
                camera.vu.x,camera.vu.y,camera.vu.z);
		MakeLighting();
		MakeMaterial();
		MakeGeometry(FALSE);
		DrawExtras();

	}

	// High res dump hack, make a 4x4 grid of images 
	if (options.windowdump > 1) {
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      CreateProjection('c');
		for (i=0;i<options.windowdump;i++) {
			for (j=0;j<options.windowdump;j++) {
				fprintf(stderr,"Frame (%d,%d)\n",i,j);
				glViewport(
					-i*options.screenwidth,-j*options.screenheight,
					options.windowdump*options.screenwidth,options.windowdump*options.screenheight);
   		   glMatrixMode(GL_MODELVIEW);
   	  		glDrawBuffer(GL_BACK_LEFT);
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   	  		glLoadIdentity();
   	   	gluLookAt(camera.vp.x,camera.vp.y,camera.vp.z,
   	   	          camera.focus.x,camera.focus.y,camera.focus.z,
   	   	          camera.vu.x,camera.vu.y,camera.vu.z);
   	   	MakeLighting();
   	   	MakeMaterial();
   	   	MakeGeometry(FALSE);
	      	DrawExtras();
				glFlush();
				if (options.stereo == ACTIVESTEREO)
					WindowDump("",options.screenwidth,options.screenheight,TRUE,12);
				else
					WindowDump("",options.screenwidth,options.screenheight,FALSE,12);
				glutSwapBuffers();
			}
		}
		options.windowdump = 0;
	}

   // Are we recording images? 
   if (options.recordimages || options.windowdump == 1) {
		if (options.stereo == ACTIVESTEREO)
      	WindowDump("",options.screenwidth,options.screenheight,TRUE,12);
		else
			WindowDump("",options.screenwidth,options.screenheight,FALSE,12);
		options.windowdump = 0;
	}

	/*
	glFlush();
	glFinish(); 
	*/
   glutSwapBuffers();

	/* 
		Handle frame rate dependent features 
		Vary the resolution of spheres
	*/
	framecount++;
   tstop = GetRunTime();
	if (tstop - tstart > 5) {
		interfacestate.framerate = framecount / (tstop - tstart);
		if (options.debug) 
     		fprintf(stderr,"Frame rate = %.1f frames/second (%d)\n",
				interfacestate.framerate,framecount);
		framecount = 0;
		tstart = tstop;
	}
}

/*
	Create the projection matrix
	Support perspective and parallel.
	eye is left or right for stereo
*/
void CreateProjection(int eye)
{
	double dist,ratio,radians=0,wd2,ndfl;
	double left,right,top,bottom,near,far;

	// Sort out near and far cutting plane for ultra stereo
	near = VectorLength(pmin,pmax) / 100;
	far  = MAX(camera.focallength,VectorLength(pmin,pmax)) * 20;
	if (options.stereo != NOSTEREO)
		near = camera.focallength / 5;

	// Window width to height ratio 
	ratio = options.screenwidth / (double)(options.screenheight);
	if (options.stereo == DUALSTEREO)
		ratio /= 2;
	radians = DTOR * camera.aperture / 2;

	switch (options.projectiontype) {
	case PERSPECTIVE:
   	wd2     = near * tan(radians);
   	ndfl    = near / camera.focallength;
		if (eye == 'l') {
			left  = - ratio * wd2 + 0.5 * camera.eyesep * ndfl;
			right =   ratio * wd2 + 0.5 * camera.eyesep * ndfl;
		} else if (eye == 'r') {
         left  = - ratio * wd2 - 0.5 * camera.eyesep * ndfl;
         right =   ratio * wd2 - 0.5 * camera.eyesep * ndfl;
		} else {
      	left  = - ratio * wd2;
      	right =   ratio * wd2;
		}
      top    =   wd2;
		bottom = - wd2;
      glFrustum(left,right,bottom,top,near,far);
		break;
	case ORTHOGRAPHIC:
		dist = VectorLength(camera.vp,camera.pr);
      left   = - dist * ratio * tan(radians);
      right  =   dist * ratio * tan(radians);
      bottom = - dist * tan(radians);
      top    =   dist * tan(radians);
		glOrtho(left,right,bottom,top,near,far);
		break;
	}
}

/*
   Create the geometry
	Draw the shaded objects first, then turn off lighting for points and lines
	Draw the possibly transparent texture objects last
	There is no handling to ensure texture objects are resident
*/
void MakeGeometry(int update)
{
   int i,j;
	XYZ linelist[300*64];
	XYZ normal;
	COLOUR white = {1,1,1};
	int nlinelist = 0;
	int objectid = 1;
	int oldsize = 1;
	static int listindex = -1;

	if (update) {             // Create list for main geometry 
		if (listindex > 0)
			glDeleteLists(listindex,1);
		listindex = glGenLists(1);
		if (options.debug)
			fprintf(stderr,"Recreating geometry, list index: %d\n",listindex);
		glNewList(listindex,GL_COMPILE);  // see end for end of list()
	} else {
		glCallList(listindex);
		return;
	}

   glPushMatrix();
   glScalef(options.scalefactor.x,options.scalefactor.y,options.scalefactor.z);
   glLineWidth(options.linescale);
   glPointSize(options.pointscale);

   // Are the objects transparent? 
   if (transparency < 1)
      glDepthMask(GL_FALSE);

	// Balls 
   for (i=0;i<nball;i++) {
		glLoadName(objectid++);
		glColor4f(ball[i].colour.r,ball[i].colour.g,ball[i].colour.b,transparency);
		CreateASphere(ball[i].p,ball[i].r,options.sphereresolution,1,1);
   }

	// Disks 
   for (i=0;i<ndisk;i++) {
		glLoadName(objectid++);
		glColor4f(disk[i].colour.r,disk[i].colour.g,disk[i].colour.b,transparency);
      CreateDisk(disk[i].p,disk[i].n,disk[i].r2,disk[i].r1,32,0.0,TWOPI);
   }

	// Cones 
   for (i=0;i<ncone;i++) {
		glLoadName(objectid++);
		glColor4f(cone[i].colour.r,cone[i].colour.g,cone[i].colour.b,transparency);
      CreateCone(cone[i].p2,cone[i].p1,cone[i].r2,cone[i].r1,32,0.0,TWOPI);
   }

	// 3 vertex faces 
	if (nface3 > 0) {
		glBegin(GL_TRIANGLES);
   	for (i=0;i<nface3;i++) {
			for (j=0;j<3;j++) {
				glNormal3f(face3[i].n[j].x,face3[i].n[j].y,face3[i].n[j].z);
				glColor4f(face3[i].colour[j].r,face3[i].colour[j].g,face3[i].colour[j].b,transparency);
				glVertex3f(face3[i].p[j].x,face3[i].p[j].y,face3[i].p[j].z);
			}
		}
		glEnd();
	}

	// 4 vertex faces 
	if (nface4 > 0) {
		glBegin(GL_QUADS);
   	for (i=0;i<nface4;i++) {
   	   for (j=0;j<4;j++) {
   	   	glNormal3f(face4[i].n[j].x,face4[i].n[j].y,face4[i].n[j].z);
				glColor4f(face4[i].colour[j].r,face4[i].colour[j].g,face4[i].colour[j].b,transparency);
   	   	glVertex3f(face4[i].p[j].x,face4[i].p[j].y,face4[i].p[j].z);
			}
   	}
		glEnd();
	}

	// Turn of lighting for the points and lines 
	glDisable(GL_LIGHTING);

	// Labels 
	for (i=0;i<nlabel;i++) {
		CreateLabelVector(label[i].s,label[i].p,label[i].right,label[i].up,linelist,&nlinelist);
		glColor4f(label[i].colour.r,label[i].colour.g,label[i].colour.b,transparency);
		for (j=0;j<nlinelist;j+=2) {
   		glBegin(GL_LINES);
   		glVertex3f(linelist[j].x,linelist[j].y,linelist[j].z);
   		glVertex3f(linelist[j+1].x,linelist[j+1].y,linelist[j+1].z);
   		glEnd();
		}
	}

	// Points 
	if (ndot > 0) {
		oldsize = dot[0].size;
		glPointSize(options.pointscale*oldsize);
		glBegin(GL_POINTS);
   	for (i=0;i<ndot;i++) {
			if (oldsize != dot[i].size) {
				glEnd();
				oldsize = dot[i].size;
				glPointSize(options.pointscale*oldsize);
				glBegin(GL_POINTS);
			}
			glColor4f(dot[i].colour.r,dot[i].colour.g,dot[i].colour.b,transparency);
   	   glVertex3f(dot[i].p.x,dot[i].p.y,dot[i].p.z);
		}
		glEnd();
		glPointSize(options.pointscale);
	}

	// Lines 
	if (nline > 0) {
      oldsize = line[0].width;
      glLineWidth(options.linescale*oldsize);
      glBegin(GL_LINES);
   	for (i=0;i<nline;i++) {
   	   if (oldsize != line[i].width) {
   	      glEnd();
				oldsize = line[i].width;
   	      glLineWidth(options.linescale*oldsize);
   	      glBegin(GL_LINES);
   	   }
			for (j=0;j<2;j++) {
				glColor4f(line[i].colour[j].r,line[i].colour[j].g,line[i].colour[j].b,transparency);
				glVertex3f(line[i].p[j].x,line[i].p[j].y,line[i].p[j].z);
			}
   	}
      glEnd();
      glLineWidth(options.linescale);
   }  

   // Draw the bounding box
   if (options.showboundingbox) {
      glPushMatrix();
      glScalef(options.scalefactor.x,options.scalefactor.y,options.scalefactor.z);
      CreateBox(pmin,pmax,0);
      glPopMatrix();
   }

	// Finally turn lighting back on for the textured objects 
	// Draw the textured objects last - for transparency reasons
	if (options.rendermode != SHADE_FLAT)
		glEnable(GL_LIGHTING);

	// Textured balls
   for (i=0;i<nballt;i++) {
      if (!options.showtextures || options.rendermode == WIREFRAME) {
			glColor4f(ballt[i].colour.r,ballt[i].colour.g,ballt[i].colour.b,transparency);
         CreateASphere(ballt[i].p,ballt[i].r,options.sphereresolution,1,1);
         continue;
      }
      glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D,ballt[i].textureid);
		glColor4f(white.r,white.g,white.b,transparency);
      CreateASphere(ballt[i].p,ballt[i].r,options.sphereresolution,1,1);
      glDisable(GL_TEXTURE_2D);
   }
	
	// Textured faces
   for (i=0;i<nface4t;i++) {
      if (!options.showtextures || options.rendermode == WIREFRAME) {
         glBegin(GL_QUADS);
			glColor4f(face4t[i].colour.r,face4t[i].colour.g,face4t[i].colour.b,transparency);
			for (j=0;j<4;j++) 
      		glVertex3f(face4t[i].p[j].x,face4t[i].p[j].y,face4t[i].p[j].z);
         glEnd();
         continue;
      }
		if (face4t[i].trans == 't')
			glDepthMask(GL_FALSE);
      glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D,face4t[i].textureid);
      glBegin(GL_QUADS);
		glColor4f(white.r,white.g,white.b,transparency);
      normal = CalcNormal(face4t[i].p[0],face4t[i].p[1],face4t[i].p[2]);
      glNormal3f(normal.x,normal.y,normal.z);
     	glTexCoord2f(0.0,1.0/face4t[i].scale);
      glVertex3f(face4t[i].p[0].x,face4t[i].p[0].y,face4t[i].p[0].z);
     	glTexCoord2f(1.0/face4t[i].scale,1.0/face4t[i].scale);
      glVertex3f(face4t[i].p[1].x,face4t[i].p[1].y,face4t[i].p[1].z);
     	glTexCoord2f(1.0/face4t[i].scale,0.0);
      glVertex3f(face4t[i].p[2].x,face4t[i].p[2].y,face4t[i].p[2].z);
     	glTexCoord2f(0.0,0.0);
      glVertex3f(face4t[i].p[3].x,face4t[i].p[3].y,face4t[i].p[3].z);
      glEnd();
      glDisable(GL_TEXTURE_2D);
		glDepthMask(GL_TRUE);
   }

	glPopMatrix();
	glDepthMask(GL_TRUE);

	if (update)
		glEndList();
}

/*
   Set up the lighting environment
*/
void MakeLighting(void)
{
	int i;
	double delta;
	GLfloat p[4];
	GLfloat white[4] = {1.0,1.0,1.0,1.0};
	GLfloat black[4] = {0.0,0.0,0.0,1.0};
	int deflightlist[8] = {GL_LIGHT0,GL_LIGHT1,GL_LIGHT2,GL_LIGHT3,
                          GL_LIGHT4,GL_LIGHT5,GL_LIGHT6,GL_LIGHT7};

   /* 
		Turn off all the lights 
		Leave the ambient component up to the global setting
		Set the specular component appropriately
	*/
	for (i=0;i<8;i++) {
   	glDisable(deflightlist[i]);
		glLightfv(deflightlist[i],GL_AMBIENT,black);
		switch (options.rendermode) {
   	case WIREFRAME:
   	case SHADE_DIFFUSE:
      	glLightfv(deflightlist[i],GL_SPECULAR,black);
			break;
   	case SHADE_SPECULAR:
      	glLightfv(deflightlist[i],GL_SPECULAR,white);
			break;
		}
	}

	// Lighting options, specified in the lightform 
   glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,options.localviewer);
   glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,options.twosided);

	// Rendering mode specific settings 
	switch (options.rendermode) {
	case WIREFRAME:
		glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
		glLightModelfv(GL_LIGHT_MODEL_AMBIENT,white);
		glShadeModel(GL_SMOOTH);
		glDisable(GL_LIGHTING);
		break;
	case SHADE_FLAT:
      glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      glLightModelfv(GL_LIGHT_MODEL_AMBIENT,white);
      glShadeModel(GL_FLAT);
      glDisable(GL_LIGHTING);
		break;
	case SHADE_DIFFUSE:
	case SHADE_SPECULAR:
		glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      glLightModelfv(GL_LIGHT_MODEL_AMBIENT,globalambient);
		glShadeModel(GL_SMOOTH);
		glEnable(GL_LIGHTING);
		break;
	}

   /*
      Create a default lighting environment if no lights were
      specified in the scene. Two lights are placed at the two
      extremities of the model, they are positional lights.
   */
   if (nlight <= 0) {
		delta = VectorLength(pmin,pmax) / 20;
		for (i=0;i<defaultlights;i++) {
			if (deflightpos[i].x == 0)
				p[0] = pmin.x - delta;
			else
				p[0] = pmax.x + delta;
         if (deflightpos[i].y == 0)
            p[1] = pmin.y - delta;
         else
            p[1] = pmax.y + delta;
         if (deflightpos[i].z == 0)
            p[2] = pmin.z - delta;
         else
            p[2] = pmax.z + delta;
			p[3] = 1.0;
			glLightfv(deflightlist[i],GL_POSITION,p);
			glLightfv(deflightlist[i],GL_DIFFUSE,white);
			glEnable(deflightlist[i]);
		}
   } else {
		for (i=0;i<nlight;i++) {
      	glLightfv(deflightlist[i],GL_POSITION,lights[i].p);
      	glLightfv(deflightlist[i],GL_DIFFUSE,lights[i].c);
      	glEnable(deflightlist[i]);
		}
	}
}

/*
	Set up the material attributes
*/
void MakeMaterial(void)
{
	GLfloat white[4] = {1.0,1.0,1.0,1.0};
	GLfloat shin[1]  = {0.0};

	glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,emission);

	switch (options.rendermode) {
	case WIREFRAME:
		glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,white);
		glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,shin);
		break;
	case SHADE_DIFFUSE:
      glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,white);
      glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,shin);
		break;
	case SHADE_SPECULAR:
      glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,specularcolour);
      glMaterialfv(GL_FRONT_AND_BACK,GL_SHININESS,shininess);
		break;
	}
}

/*
   Deal with plain key strokes
	The x and y are the screen coordinate of the mouse at the time
*/
void HandleKeyboard(unsigned char key,int x, int y)
{
	XYZ origin = {0.0,0.0,0.0};

   switch (key) {
   case 27: 									/* ESC 			*/
   case 'Q':									/* Quit 			*/
   case 'q': 
		CleanExit();
		break;
	case 'a':									/* Start/stop autospin */
	case 'A':
      if (options.autospin.x != 0 || options.autospin.y != 0 || options.autospin.z != 0) {
         options.autospin = origin;
      } else {
			options.autospin.x = 1; 
			options.autospin.y = 0; 
			options.autospin.z = 0;
		}
		break;
	case 'd':
		options.debug = !options.debug;
		break;
	case 'D':									/* Delete everything */
		DeleteGeometry();
		break;
	case 'h':									/* Go home		*/
	case 'H':
		camera = options.camerahome;
		break;
	case '1':									/* Special camera views */
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
		CameraHome((key-'1')+3);
		break;
	case 'f':      							/* Focus */
	case 'F':
      CameraHome(FOCUS);
		break;
	case '[':                           /* Roll anti clockwise */
		RotateCamera(0.0,0.0,-1.0,KEYBOARDCONTROL);
		break;
	case ']':                           /* Roll clockwise */
		RotateCamera(0.0,0.0,1.0,KEYBOARDCONTROL);
		break;
	case 'i':									/* Translate camera up */
		TranslateCamera(0.0,1.0,KEYBOARDCONTROL);
		break;
   case 'k':                           /* Translate camera down */
      TranslateCamera(0.0,-1.0,KEYBOARDCONTROL);
      break;
   case 'j':                           /* Translate camera left */
      TranslateCamera(-1.0,0.0,KEYBOARDCONTROL);
      break;
   case 'l':                           /* Translate camera right */
      TranslateCamera(1.0,0.0,KEYBOARDCONTROL);
      break;
	case '-':									/* Move backward or increase speed */
	case '_':
		if (options.interaction == INSPECT) {
			FlyCamera(-50.0);
		} else {
			if (ABS(camera.speed) < EPSILON) {
				camera.speed = -1;
			} else if (ABS(camera.speed-1) < EPSILON) {
      	   camera.speed = 0;
			} else {
				camera.speed--;
			}
		}
		break;
   case '+':									/* Move forward or increase speed */
	case '=':
      if (options.interaction == INSPECT) {
         FlyCamera(50.0);
      } else {
      	if (ABS(camera.speed) < EPSILON) {
      	   camera.speed = 1;
      	} else if (ABS(camera.speed+1) < EPSILON) {
      	   camera.speed = 0;
      	} else {
      	  	camera.speed++;
			}
		}
      break;
	case 'c':    								/* Write the camera position to log */
		AppendCameraPosition();
		break;
	case 's':
		break;
	case 'S':
		break;
   }
}

/*
   Deal with special key strokes
	The exact operation depends on whether we're inspecting or flying
	In inspect mode they rotate about the focal point
	In fly mode they direct the camera view direction
*/
void HandleSpecialKeyboard(int key,int x,int y)
{
	XYZ origin = {0.0,0.0,0.0};

   switch (key) {
   case GLUT_KEY_LEFT:  
		RotateCamera(-1.0,0.0,0.0,KEYBOARDCONTROL); 
		if (options.autospin.x != 0 || options.autospin.y != 0 || options.autospin.z != 0) {
			options.autospin = origin;
			options.autospin.x = 1;
		}
		break;
   case GLUT_KEY_RIGHT: 
		RotateCamera(1.0,0.0,0.0,KEYBOARDCONTROL);  
      if (options.autospin.x != 0 || options.autospin.y != 0 || options.autospin.z != 0) {
			options.autospin = origin;
         options.autospin.x = -1; 
		}
		break;
   case GLUT_KEY_UP:    
		if (glutGetModifiers() == GLUT_ACTIVE_SHIFT) {
			break;
		}
		if (options.interaction == WALK) {
			TranslateCamera(0.0,1.0,KEYBOARDCONTROL);
			break;
		}
		RotateCamera(0.0,1.0,0.0,KEYBOARDCONTROL);  
      if (options.autospin.x != 0 || options.autospin.y != 0 || options.autospin.z != 0) {
			options.autospin = origin;
         options.autospin.y = 1;
		}
		break;
   case GLUT_KEY_DOWN:  
      if (glutGetModifiers() == GLUT_ACTIVE_SHIFT) {
         break;
      }
      if (options.interaction == WALK) {
         TranslateCamera(0.0,-1.0,KEYBOARDCONTROL);
         break;
      }
		RotateCamera(0.0,-1.0,0.0,KEYBOARDCONTROL); 
      if (options.autospin.x != 0 || options.autospin.y != 0 || options.autospin.z != 0) {
			options.autospin = origin;
         options.autospin.y = -1;
		}
		break;
	case GLUT_KEY_F1:
      options.showhelpinfo = !options.showhelpinfo;
		break;
   case GLUT_KEY_F2:
		options.showboundingbox = !options.showboundingbox;
      break;
   case GLUT_KEY_F4:
      break;
   case GLUT_KEY_F5:
      break;
   case GLUT_KEY_F6:
      break;
   case GLUT_KEY_F7:
      break;
   case GLUT_KEY_F8:
      break;
   case GLUT_KEY_F9:
		options.windowdump = 3;
      break;
   case GLUT_KEY_F10:
		options.windowdump = 1;
      break;
   case GLUT_KEY_F11:
		options.makeqtvr = TRUE;
      break;
   case GLUT_KEY_F12:
		options.recordimages = !options.recordimages;
      break;
	case GLUT_KEY_PAGE_UP:
		break;
   case GLUT_KEY_PAGE_DOWN:
      break;
	case GLUT_KEY_HOME:
		camera = options.camerahome;
		break;
	case GLUT_KEY_END:
		CleanExit();
		break;
	case GLUT_KEY_INSERT:
		break;
   }
}

/*
	Rotate (ix,iy) or roll (iz) the camera.
	Depends on the interaction mode.
	Correctly update all camera attributes
*/
void RotateCamera(double ix,double iy,double iz,int source)
{
	XYZ vp,vu,vd;
   XYZ right;
	XYZ newvp,newr;
	double delta,radius;
	double dx,dy,dz;

	vu = camera.vu;
	Normalise(&vu);
	vp = camera.vp;
	vd = camera.vd;
	Normalise(&vd);
   CROSSPROD(vd,vu,right);
   Normalise(&right);

	/* Determine the amount of change */
	delta = options.interactspeed * DTOR;
	if (source == MOUSECONTROL)
		delta /= 2;

	/* 
		Handle the roll 
		Just change the up vector
	*/
	if (ABS(iz) > EPSILON) {
		camera.vu.x += iz * right.x * delta;
      camera.vu.y += iz * right.y * delta;
      camera.vu.z += iz * right.z * delta;
		Normalise(&camera.vu);
		return;
	}

	/*
		Fly mode
		Change the view direction vector along up or right vector
	*/
	if (options.interaction == FLY) {
		camera.vd.x = vd.x + delta * ix * right.x + delta * iy * vu.x;
      camera.vd.y = vd.y + delta * ix * right.y + delta * iy * vu.y;
      camera.vd.z = vd.z + delta * ix * right.z + delta * iy * vu.z;
		Normalise(&camera.vd);

      /* Determine the new up vector */
		if (ABS(ix) > ABS(iy)) {
      	CROSSPROD(camera.vd,camera.vu,right);
      	Normalise(&camera.vu);
         CROSSPROD(right,camera.vd,camera.vu);
         Normalise(&camera.vu);
		} else if (ABS(ix) < ABS(iy)) {
         CROSSPROD(right,camera.vd,camera.vu);
         Normalise(&camera.vu);
		}

   	camera.focus.x = camera.vp.x + camera.focallength * camera.vd.x;
   	camera.focus.y = camera.vp.y + camera.focallength * camera.vd.y;
   	camera.focus.z = camera.vp.z + camera.focallength * camera.vd.z;

		return;
	}

   /*
      Walk mode
      ix changes the view direction
		iy changes the height along the up vector
   */
   if (options.interaction == WALK) {
      camera.vd.x = vd.x + delta * ix * right.x;
      camera.vd.y = vd.y + delta * ix * right.y;
      camera.vd.z = vd.z + delta * ix * right.z;
      Normalise(&camera.vd);

      camera.focus.x = camera.vp.x + camera.focallength * camera.vd.x;
      camera.focus.y = camera.vp.y + camera.focallength * camera.vd.y;
      camera.focus.z = camera.vp.z + camera.focallength * camera.vd.z;

      return;
   }

	/* 
		Inspect mode 
		Handle special case when the camera coincident with the rotate point
	*/
   dx = camera.vp.x - camera.pr.x;
   dy = camera.vp.y - camera.pr.y;
   dz = camera.vp.z - camera.pr.z;
   radius = sqrt(dx*dx + dy*dy + dz*dz);
	if (ABS(radius) > 0.0001) {

		/* Determine the new view point */
		delta *= radius;
		newvp.x = vp.x + delta * ix * right.x + delta * iy * vu.x - camera.pr.x; 
		newvp.y = vp.y + delta * ix * right.y + delta * iy * vu.y - camera.pr.y;
		newvp.z = vp.z + delta * ix * right.z + delta * iy * vu.z - camera.pr.z;
		Normalise(&newvp);
		camera.vp.x = camera.pr.x + radius * newvp.x;
		camera.vp.y = camera.pr.y + radius * newvp.y;
		camera.vp.z = camera.pr.z + radius * newvp.z;

   	/* Determine the new right vector */
   	newr.x = camera.vp.x + right.x - camera.pr.x;
   	newr.y = camera.vp.y + right.y - camera.pr.y;
   	newr.z = camera.vp.z + right.z - camera.pr.z;
   	Normalise(&newr);
   	newr.x = camera.pr.x + radius * newr.x - camera.vp.x;
   	newr.y = camera.pr.y + radius * newr.y - camera.vp.y;
   	newr.z = camera.pr.z + radius * newr.z - camera.vp.z;
		Normalise(&newr);

		camera.vd.x = camera.pr.x - camera.vp.x;
		camera.vd.y = camera.pr.y - camera.vp.y;
		camera.vd.z = camera.pr.z - camera.vp.z;
		Normalise(&camera.vd);

		/* Determine the new up vector */
		CROSSPROD(newr,camera.vd,camera.vu);
		Normalise(&camera.vu);

	} else {		/* Rotating about the camera position */

		// Calculate the new view direction 
		camera.vd.x += delta * ix * right.x + delta * iy * vu.x;
      camera.vd.y += delta * ix * right.y + delta * iy * vu.y;
      camera.vd.z += delta * ix * right.z + delta * iy * vu.z;
		Normalise(&camera.vd);

      // Determine the new up vector 
      CROSSPROD(right,camera.vd,camera.vu);
      Normalise(&camera.vu);

	}

   /* Update the focal point */
   Normalise(&camera.vd);
   camera.focus.x = camera.vp.x + camera.focallength * camera.vd.x;
   camera.focus.y = camera.vp.y + camera.focallength * camera.vd.y;
   camera.focus.z = camera.vp.z + camera.focallength * camera.vd.z;
}

/*
	Fly the camera forward or backward at a given speed
	In INSPECT mode we just get shifted in increments.
	In WALK and FLY mode it is continuous
*/
void FlyCamera(double speed)
{
	double delta;
	XYZ vp;
	
	/* Determine the amount to move */
	if (options.deltamove <= 0)
		delta = speed * VectorLength(pmin,pmax) / 2000;
	else
		delta = speed * options.deltamove;

	/* Move the camera */
   vp.x = camera.vp.x + delta * camera.vd.x;
   vp.y = camera.vp.y + delta * camera.vd.y;
   vp.z = camera.vp.z + delta * camera.vd.z;

	/* Update the permanent settings */
	camera.vp = vp;
   if (options.projectiontype == PERSPECTIVE) {
      camera.pr.x += delta * camera.vd.x;
      camera.pr.y += delta * camera.vd.y;
      camera.pr.z += delta * camera.vd.z;
   }

	/* Update the focal distance */
   camera.focus.x = camera.vp.x + camera.focallength * camera.vd.x;
   camera.focus.y = camera.vp.y + camera.focallength * camera.vd.y;
   camera.focus.z = camera.vp.z + camera.focallength * camera.vd.z;
}

/*
   Translate (pan) the camera view point
	In response to i,j,k,l, and arrow keys
*/
void TranslateCamera(double ix,double iy,int source)
{
   XYZ vp,vu,vd,pr;
   XYZ right;
   double delta;

   vu = camera.vu;
   Normalise(&vu);
   vp = camera.vp;
   vd = camera.vd;
	pr = camera.pr;
   Normalise(&vd);
   CROSSPROD(vd,vu,right);
   Normalise(&right);

	/* Determine the amount to translate by */
	if (options.deltamove <= 0)
		delta = options.interactspeed * VectorLength(pmin,pmax) / 500.0;
	else
		delta = options.deltamove;
	if (source == MOUSECONTROL)
		delta /= 2;
	
	/* Move the camera horizontally */
	vp.x += iy * vu.x * delta;
   vp.y += iy * vu.y * delta;
	vp.z += iy * vu.z * delta;
	if (options.projectiontype == PERSPECTIVE) {
   	pr.x += iy * vu.x * delta;
   	pr.y += iy * vu.y * delta;
   	pr.z += iy * vu.z * delta;
	}

	/* Move the camera vertically */
   vp.x += ix * right.x * delta;
   vp.y += ix * right.y * delta;
   vp.z += ix * right.z * delta;
	if (options.projectiontype == PERSPECTIVE) {
   	pr.x += ix * right.x * delta;
   	pr.y += ix * right.y * delta;
   	pr.z += ix * right.z * delta;
	}

	/* Update the camera */
	camera.vp = vp;
	camera.pr = pr;
   camera.focus.x = camera.vp.x + camera.focallength * camera.vd.x;
   camera.focus.y = camera.vp.y + camera.focallength * camera.vd.y;
   camera.focus.z = camera.vp.z + camera.focallength * camera.vd.z;
}

/*
   Handle mouse events
	Right button events are passed to the menu handler
	Turn off autopilot mode on left and middle mouse click
*/
void HandleMouse(int button,int state,int x,int y)
{
   int i,maxselect = 100,nhits = 0;
   GLuint selectlist[100];
   GLint viewport[4];

   if (state == GLUT_DOWN) {

		if (button == GLUT_LEFT_BUTTON) {

			if (options.selectmode) {
      		glSelectBuffer(maxselect,selectlist);
      		glRenderMode(GL_SELECT);
      		glInitNames();
      		glPushName(-1);

      		glMatrixMode(GL_PROJECTION);
      		glLoadIdentity();
      		glGetIntegerv(GL_VIEWPORT,viewport); /* Get the viewport bounds */
      		gluPickMatrix(x,viewport[3]-y,3.0,3.0,viewport);
				CreateProjection('c');
      		glMatrixMode(GL_MODELVIEW);
      		glDrawBuffer(GL_BACK_LEFT);
      		glLoadIdentity();
      		gluLookAt(camera.vp.x,camera.vp.y,camera.vp.z,
      		          camera.focus.x,camera.focus.y,camera.focus.z,
      		          camera.vu.x,camera.vu.y,camera.vu.z);
      		MakeLighting();
      		MakeMaterial();
      		MakeGeometry(FALSE);
      		DrawExtras();
	
      		if ((nhits = glRenderMode(GL_RENDER)) < 0) {
					fprintf(stderr,"Select buffer overflow\n");
				} else {
      		   for (i=0;i<nhits;i++) {
      		      fprintf(stderr,"Selected item %d, at depth %d -> %d\n",
							(int)selectlist[4*i+3],(int)selectlist[4*i+1],(int)selectlist[4*i+2]);
      		   }
      		}	
			}

         interfacestate.currentbutton = GLUT_LEFT_BUTTON;
			options.autopilot = FALSE;

      } else if (button == GLUT_MIDDLE_BUTTON) {

         interfacestate.currentbutton = GLUT_MIDDLE_BUTTON;
			options.autopilot = FALSE;

      }
   }

   interfacestate.mousexlast = x;
   interfacestate.mouseylast = y;
}

/*
   Handle mouse dragging.
	Left button
		Rotate vertically and horizontally in INSPECT and FLY modes
		Rotate horizaontally and translate vertically in WALK mode
	Middle button rolls in all modes 
	Interaction rate is a fraction of that for keyboard control
	Right button is reserverd for the menus.
*/
void HandleMouseMotion(int x,int y)
{
	double dx,dy;

	dx = x - interfacestate.mousexlast;
	dy = y - interfacestate.mouseylast;

	if (interfacestate.currentbutton == GLUT_LEFT_BUTTON) {
		if (options.interaction == WALK) {
      	if (dy > 0)
         	TranslateCamera(0.0,1.0,MOUSECONTROL);
      	else if (dy < 0)
         	TranslateCamera(0.0,-1.0,MOUSECONTROL);
			RotateCamera(-dx,0.0,0.0,MOUSECONTROL);
		} else {
  			RotateCamera(-dx,dy,0.0,MOUSECONTROL);
		}
	} else if (interfacestate.currentbutton == GLUT_MIDDLE_BUTTON) {
		RotateCamera(0.0,0.0,dx,MOUSECONTROL);
	}

   interfacestate.mousexlast = x;
   interfacestate.mouseylast = y;
}

/*
	Move the 3D mouse around on passive mouse movement events
	Only do it if we are in 3D curcor mode 
*/
void HandlePassiveMotion(int x,int y)
{
	static int xlast=-1,ylast=-1;
	double dx,dy;

   dx = x - xlast;
   dy = y - ylast;

   xlast = x;
   ylast = y;
}

/*
   How to handle visibility
*/
void HandleVisibility(int visible)
{
   if (visible == GLUT_VISIBLE)
      glutIdleFunc(HandleIdle);
   else
      glutIdleFunc(NULL);
}

/*
   What to do on an idle event
	Check all the forms (if they are visible) for events
*/
void HandleIdle(void)
{
   static double tstart = -1;
	double tstop;
	FL_OBJECT *retobj;

	// QuickTime VR 
	static int theta=0,phi=40;
   double range;
   XYZ right,up = {0,0,1};

	// Make qtvr frames 
   if (options.makeqtvr) {
		range = VectorLength(camera.vp,camera.pr);
		camera.vp.x = camera.pr.x + range * cos(theta*DTOR) * cos(phi*DTOR);
		camera.vp.y = camera.pr.y + range * sin(theta*DTOR) * cos(phi*DTOR);
		camera.vp.z = camera.pr.z + range * sin(phi*DTOR);
		camera.vd = VectorSub(camera.vp,camera.pr);
		Normalise(&camera.vd);
		right = CrossProduct(camera.vd,up);
		camera.vu = CrossProduct(right,camera.vd);
		Normalise(&camera.vu);
		theta += 10;
		if (theta % 360 == 0) {
			phi -= 10;
			theta = 0;
		}
		if (phi < -40) {
			options.makeqtvr = FALSE;
			theta = 0;
			phi = 60;
		}
		options.windowdump = 1;
		fprintf(stderr,"qtvr: %d %d\n",theta,phi);
      glutPostRedisplay();
		return;
   }

	// Events from the camera form 
	if (cameraformvisible) {
		retobj = fl_check_forms();
   	if (retobj == cameraform->Hide) {
      	fl_hide_form(cameraform->cameraform);
			cameraformvisible = FALSE;
		}
		if (retobj == cameraform->Eyesep) {
			camera.eyesep = fl_get_slider_value(cameraform->Eyesep);
		}
   	if (retobj == cameraform->Focallength) {
      	camera.focallength = fl_get_slider_value(cameraform->Focallength);
   		camera.focus.x = camera.vp.x + camera.focallength * camera.vd.x;
   		camera.focus.y = camera.vp.y + camera.focallength * camera.vd.y;
   		camera.focus.z = camera.vp.z + camera.focallength * camera.vd.z;
   	}
		if (retobj == cameraform->Aperture) {
			camera.aperture = fl_get_slider_value(cameraform->Aperture);
		}
	}

	// Events from the material form 
	if (materialformvisible) {
		retobj = fl_check_forms();
		if (retobj == materialform->Hide) {
			fl_hide_form(materialform->materialform);
			materialformvisible = FALSE;
			MakeGeometry(TRUE);
		}
		if (retobj == materialform->Shininess) {
			shininess[0] = fl_get_slider_value(materialform->Shininess);
		}
      if (retobj == materialform->Emission) {
         emission[0] = fl_get_slider_value(materialform->Emission);
         emission[1] = emission[0];
         emission[2] = emission[0];
      }
		if (retobj == materialform->Transparency) {
			transparency = fl_get_slider_value(materialform->Transparency);
		}
		if (retobj == materialform->Red) {
			specularcolour[0] = atof(fl_get_input(materialform->Red));
		}
      if (retobj == materialform->Green) {
         specularcolour[1] = atof(fl_get_input(materialform->Green));
      }
      if (retobj == materialform->Blue) {
         specularcolour[2] = atof(fl_get_input(materialform->Blue));
      }
	}

	// Events from the light form 
   if (lightformvisible) {
      retobj = fl_check_forms();
      if (retobj == lightform->Hide) {
         fl_hide_form(lightform->lightform);
         lightformvisible = FALSE;
         MakeGeometry(TRUE);
      }
      if (retobj == lightform->Ambient) {
         globalambient[0] = fl_get_slider_value(lightform->Ambient);
			globalambient[1] = globalambient[0];
			globalambient[2] = globalambient[0];
      }
		if (retobj == lightform->LocalViewer) {
			options.localviewer = fl_get_button(lightform->LocalViewer);
		}
      if (retobj == lightform->TwoSided) {
         options.twosided = fl_get_button(lightform->TwoSided);
      }
		if (retobj == lightform->DefaultLights) {
			defaultlights = fl_get_counter_value(lightform->DefaultLights);
		}
	}

	// Events from the appearance form 
	if (appearformvisible) {
      retobj = fl_check_forms();
      if (retobj == appearform->Hide) {
         fl_hide_form(appearform->appearform);
         appearformvisible = FALSE;
         MakeGeometry(TRUE);
      }
		if (retobj == appearform->PointScale) {
			options.pointscale = atof(fl_get_input(appearform->PointScale));
			MakeGeometry(TRUE);
		}
		if (retobj == appearform->LineScale) {
			options.linescale = atof(fl_get_input(appearform->LineScale));
			MakeGeometry(TRUE);
		}
		if (retobj == appearform->Xscale) {
         options.scalefactor.x = atof(fl_get_input(appearform->Xscale));
         MakeGeometry(TRUE);
      }
      if (retobj == appearform->Yscale) {
         options.scalefactor.y = atof(fl_get_input(appearform->Yscale));
         MakeGeometry(TRUE);
      }
      if (retobj == appearform->Zscale) {
         options.scalefactor.z = atof(fl_get_input(appearform->Zscale));
         MakeGeometry(TRUE);
      }
	}

	// Is it time for another update 
   if (tstart < 0) 
      tstart = GetRunTime();
   tstop = GetRunTime();
   if (tstop - tstart > 1.0/options.targetfps) {
      glutPostRedisplay();
      tstart = tstop;
   }
}

/*
	Save a geometry file
	Note this upgrades the primitives so the output won't be the
	same as an input geom file. For example, "p" becomes "P", this
	reflects the general upgrading that occurs when reading a 
	geom file.
*/
void SaveGeomFile(void)
{
	int i,j;
   const char *fn;
   char fname[256];
   FILE *fptr;

   /* Get the file name */
   fn = fl_show_file_selector("GEOM file","","","");
   if (fn == NULL || strlen(fn) <= 0)
      return;
   strcpy(fname,fn);

   /* Try to open the file */
   if ((fptr = fopen(fname,"w")) == NULL)
      return;

	for (i=0;i<nlight;i++) 
		fprintf(fptr,"i %g %g %g %g %g %g\n",
			lights[i].p[0],lights[i].p[1],lights[i].p[2],
			lights[i].c[0],lights[i].c[1],lights[i].c[2]);

	for (i=0;i<nball;i++)
		fprintf(fptr,"s %g %g %g %g %g %g %g\n",
			ball[i].p.x,ball[i].p.y,ball[i].p.z,
			ball[i].r,
			ball[i].colour.r,ball[i].colour.g,ball[i].colour.b);

   for (i=0;i<nballt;i++)
      fprintf(fptr,"st %g %g %g %g %g %g %g %s\n",
         ballt[i].p.x,ballt[i].p.y,ballt[i].p.z,
			ballt[i].r,
         ballt[i].colour.r,ballt[i].colour.g,ballt[i].colour.b,
			ballt[i].texturename);

	for (i=0;i<ndisk;i++) 
		fprintf(fptr,"d %g %g %g %g %g %g %g %g %g %g %g\n",
			disk[i].p.x,disk[i].p.y,disk[i].p.z,
         disk[i].n.x,disk[i].n.y,disk[i].n.z,
			disk[i].r1,disk[i].r2,
			disk[i].colour.r,disk[i].colour.g,disk[i].colour.b);

	for (i=0;i<ncone;i++)
		fprintf(fptr,"c %g %g %g %g %g %g %g %g %g %g %g\n",
			cone[i].p1.x,cone[i].p1.y,cone[i].p1.z,
			cone[i].p2.x,cone[i].p2.y,cone[i].p2.z,
			cone[i].r1,cone[i].r2,
			cone[i].colour.r,cone[i].colour.g,cone[i].colour.b);

	for (i=0;i<ndot;i++)
		fprintf(fptr,"P %g %g %g %g %g %g %g\n",
			dot[i].p.x,dot[i].p.y,dot[i].p.z,
			dot[i].colour.r,dot[i].colour.g,dot[i].colour.b,
			dot[i].size);

	for (i=0;i<nline;i++) {
		fprintf(fptr,"lc %g %g %g %g %g %g %g %g %g %g %g %g %g\n",
			line[i].p[0].x,line[i].p[0].y,line[i].p[0].z,
			line[i].p[1].x,line[i].p[1].y,line[i].p[1].z,
			line[i].colour[0].r,line[i].colour[0].g,line[i].colour[0].b,
         line[i].colour[1].r,line[i].colour[1].g,line[i].colour[1].b,
			line[i].width);
	}

	for (i=0;i<nface3;i++) {
		fprintf(fptr,"f3nc");
		for (j=0;j<3;j++)
			fprintf(fptr," %g %g %g",face3[i].p[j].x,face3[i].p[j].y,face3[i].p[j].z);
      for (j=0;j<3;j++)
         fprintf(fptr," %g %g %g",face3[i].n[j].x,face3[i].n[j].y,face3[i].n[j].z);
      for (j=0;j<3;j++)
         fprintf(fptr," %g %g %g",
				face3[i].colour[j].r,face3[i].colour[j].g,face3[i].colour[j].b);
		fprintf(fptr,"\n");
	}

	for (i=0;i<nface4;i++) {
      fprintf(fptr,"f4nc");
      for (j=0;j<4;j++)
         fprintf(fptr," %g %g %g",face4[i].p[j].x,face4[i].p[j].y,face4[i].p[j].z);
      for (j=0;j<4;j++)
         fprintf(fptr," %g %g %g",face4[i].n[j].x,face4[i].n[j].y,face4[i].n[j].z);
      for (j=0;j<4;j++)
         fprintf(fptr," %g %g %g",
            face4[i].colour[j].r,face4[i].colour[j].g,face4[i].colour[j].b);
      fprintf(fptr,"\n");
   }

	for (i=0;i<nface4t;i++) {
      fprintf(fptr,"f4t");
      for (j=0;j<4;j++)
         fprintf(fptr," %g %g %g",
				face4t[i].p[j].x,face4t[i].p[j].y,face4t[i].p[j].z);
      fprintf(fptr," %g %g %g",
         face4t[i].colour.r,face4t[i].colour.g,face4t[i].colour.b);
		fprintf(fptr," %s %g %c\n",
			face4t[i].texturename,face4t[i].scale,face4t[i].trans);
      fprintf(fptr,"\n");
   }

	for (i=0;i<nlabel;i++) {
		fprintf(fptr,"t %g %g %g %g %g %g %g %g %g %g %g %g %s\n",
			label[i].p.x,label[i].p.y,label[i].p.z,
			label[i].right.x,label[i].right.y,label[i].right.z,
			label[i].up.x,label[i].up.y,label[i].up.z,
			label[i].colour.r,label[i].colour.g,label[i].colour.b,
			label[i].s);
   }

	fclose(fptr);
}

/*
	Read a geometry file
	For documentation see
		http://www.swin.edu.au/astronomy/pbourke/3dformats/geom/
	In general primitives get upgraded to higher forms.
*/
int ReadGeometry(char *name)
{
	int i,d;
	int gotitalready = -1;
	int linecount = 0;
	FILE *fptr;
	char id[32],texturename[64],trans[32];
	double r,r2,scale,size,width;
	XYZ p[4],n[4];
	COLOUR c[4];
	const char *fn;
	XYZ origin = {0.0,0.0,0.0};

	if (strlen(name) < 1) {
      fn = fl_show_file_selector("Geometry file","","","");
      if (fn == NULL || strlen(fn) <= 0) 
     		return(TRUE); 
      if ((fptr = fopen(fn,"r")) == NULL) {
         fl_show_message("File Error","Unable to open geometry file",fn);
         return(FALSE);
      }
	} else {
   	if ((fptr = fopen(name,"r")) == NULL) {
      	fl_show_message("File Error","Unable to open geometry file",name);
      	return(FALSE);
		}
   }

	while (fscanf(fptr,"%s",id) == 1) {
		if (strcmp(id,"#") == 0) {											/* Comment */
			;
		} else if (strcmp(id,"t") == 0) {								/* Label */
			if (!ReadVector(fptr,&(p[0]))) {
            fprintf(stderr,"Failed reading label position (%d)",linecount);
            fclose(fptr);
            continue;
			}
         if (!ReadVector(fptr,&(p[1]))) {
            fprintf(stderr,"Failed reading label right vector (%d)",linecount);
            fclose(fptr);
            continue;
         }
         if (!ReadVector(fptr,&(p[2]))) {
            fprintf(stderr,"Failed reading label up vector (%d)",linecount);
            fclose(fptr);
            continue;
         }
         if (!ReadColour(fptr,&(c[0]))) {
            fprintf(stderr,"Failed reading label colour (%d)",linecount);
            fclose(fptr);
            continue;
         }
         if ((label = realloc(label,(nlabel+1)*sizeof(LABEL))) == NULL) {
            fprintf(stderr,"realloc for label failed\n");
            exit(-1);
         }
         label[nlabel].p      = p[0];
			label[nlabel].right  = p[1];
			label[nlabel].up     = p[2];
			label[nlabel].colour = c[0];
         i = 0;
         while ((d = fgetc(fptr)) != '\n' && d != '\r' && i < 63) {
				if (d == ' ' && i == 0)
					;
				else
            	label[nlabel].s[i++] = d;
			}
			label[nlabel].s[i] = '\0';
			nlabel++;

		} else if (strcmp(id,"m") == 0) {                        /* Marker */

         if (fscanf(fptr,"%d %lf",&d,&size) != 2 ||
				 !ReadVector(fptr,&(p[0])) ||
             !ReadColour(fptr,&(c[0]))) {
            fprintf(stderr,"Failed reading marker (%d)\n",linecount);
            fclose(fptr);
            continue;
         }
			if (d < 0 || d > 3) {
				fprintf(stderr,"Unexpected marker: %d (%d)\n",d,linecount);
				continue;
			} 
			AddMarker2Database(d,size,p[0],c[0]);

		} else if (strcmp(id,"s") == 0) {								/* Ball */

			if (!ReadVector(fptr,&(p[0])) || 
				  fscanf(fptr,"%lf",&r) != 1 ||
				 !ReadColour(fptr,&(c[0]))) {
            fprintf(stderr,"Failed reading ball (%d)",linecount);
            fclose(fptr);
            continue;
         }
			if ((ball = realloc(ball,(nball+1)*sizeof(BALL))) == NULL) {
            fprintf(stderr,"realloc for ball failed\n");
            exit(-1);
         }
			ball[nball].p = p[0];
			ball[nball].r = r;
			ball[nball].colour = c[0];
			nball++;

      } else if (strcmp(id,"st") == 0) {                        /* Ball */

         if (!ReadVector(fptr,&(p[0])) ||
              fscanf(fptr,"%lf",&r) != 1 ||
             !ReadColour(fptr,&(c[0]))) {
            fprintf(stderr,"Failed reading ball (%d)",linecount);
            fclose(fptr);
            continue;
         }
         if (fscanf(fptr,"%s",texturename) != 1) {
            fprintf(stderr,"Failed reading texture name for ballt (%d)\n",linecount);
            fclose(fptr);
            continue;
         }
         if ((ballt = realloc(ballt,(nballt+1)*sizeof(BALLT))) == NULL) {
            fprintf(stderr,"realloc for textured ball failed\n");
            exit(-1);
         }
         ballt[nballt].p = p[0];
         ballt[nballt].r = r;
         ballt[nballt].colour = c[0];
			strcpy(ballt[nballt].texturename,texturename);
			/* Check to see if the texture already exists */
			gotitalready = -1;
			for (i=0;i<nballt;i++) {
				if (strcmp(ballt[i].texturename,texturename) == 0) {
					gotitalready = i;
					break;
				}
			}
			if (gotitalready >= 0) {
				ballt[nballt].rgba = NULL;
				ballt[nballt].textureid = ballt[gotitalready].textureid;
            if (options.debug)
               fprintf(stderr,"   Resusing texture \"%s\"\n",texturename);
			} else {
				ballt[nballt].rgba = ReadTGATexture(texturename,&(ballt[nballt].width),&(ballt[nballt].height));
				glGenTextures(1,&(ballt[nballt].textureid));
				if (options.debug) 
					fprintf(stderr,"Texture: %s (%dx%d), id: %d\n",
						texturename,ballt[nballt].width,ballt[nballt].height,(int)ballt[nballt].textureid);
				glBindTexture(GL_TEXTURE_2D,ballt[nballt].textureid);
      		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
      		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
      		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
      		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
         	glTexImage2D(GL_TEXTURE_2D,0,4,
         	   ballt[nballt].width,ballt[nballt].height,
 	            0,GL_RGBA,GL_UNSIGNED_BYTE,ballt[nballt].rgba);
      		glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
			}
         nballt++;

      } else if (strcmp(id,"d") == 0) {                        /* Disk */

         if (!ReadVector(fptr,&(p[0])) ||
				 !ReadVector(fptr,&(n[0])) ||
              fscanf(fptr,"%lf %lf",&r,&r2) != 2 ||
             !ReadColour(fptr,&(c[0]))) {
            fprintf(stderr,"Failed reading disk (%d)",linecount);
            fclose(fptr);
            continue;
         }
         if ((disk = realloc(disk,(ndisk+1)*sizeof(DISK))) == NULL) {
            fprintf(stderr,"realloc for disk failed\n");
            exit(-1);
         }
         disk[ndisk].p = p[0];
			disk[ndisk].n = n[0];
         disk[ndisk].r1 = r;
			disk[ndisk].r2 = r2;
         disk[ndisk].colour = c[0];
         ndisk++;

      } else if (strcmp(id,"c") == 0) {                        /* Cone */

         if (!ReadVector(fptr,&(p[0])) ||
             !ReadVector(fptr,&(p[1])) ||
              fscanf(fptr,"%lf %lf",&r,&r2) != 2 ||
             !ReadColour(fptr,&(c[0]))) {
            fprintf(stderr,"Failed reading cone (%d)",linecount);
            fclose(fptr);
            continue;
         }
         if ((cone = realloc(cone,(ncone+1)*sizeof(CONE))) == NULL) {
            fprintf(stderr,"realloc for cone failed\n");
            exit(-1);
         }
         cone[ncone].p1 = p[0];
         cone[ncone].p2 = p[1];
         cone[ncone].r1 = r;
         cone[ncone].r2 = r2;
         cone[ncone].colour = c[0];
         ncone++;

      } else if (strcmp(id,"i") == 0) {                        /* Light */

         if (!ReadVector(fptr,&(p[0])) ||
             !ReadColour(fptr,&(c[0]))) {
            fprintf(stderr,"Failed reading light (%d)",linecount);
            fclose(fptr);
            continue;
         }
			if (nlight < MAXLIGHT) {
         	lights[nlight].p[0] = p[0].x;
            lights[nlight].p[1] = p[0].y;
            lights[nlight].p[2] = p[0].z;
            lights[nlight].p[3] = 1;			/* Positional lights */
         	lights[nlight].c[0] = c[0].r;
            lights[nlight].c[1] = c[0].g;
            lights[nlight].c[2] = c[0].b;
            lights[nlight].c[3] = 1;
         	nlight++;
			}

		} else if (strcmp(id,"p") == 0 || strcmp(id,"P") == 0) {		/* Dot */	

			size = 1;
         if (!ReadVector(fptr,&(p[0])) ||
             !ReadColour(fptr,&(c[0]))) {
            fprintf(stderr,"Failed reading dot (%d)",linecount);
            fclose(fptr);
            continue;
         }
			if (strcmp(id,"P") == 0) {
				if (fscanf(fptr,"%lf",&size) != 1) {
					fprintf(stderr,"Failed reading dot size (%d)",linecount);
					fclose(fptr);
            	continue;
				}
			}
			if ((dot = realloc(dot,(ndot+1)*sizeof(DOT))) == NULL) {
            fprintf(stderr,"realloc for dot failed\n");
            exit(-1);
         }
			dot[ndot].p = p[0];
			dot[ndot].colour = c[0];
			dot[ndot].size = ABS(size);
			ndot++;

		} else if (strcmp(id,"l") == 0 || strcmp(id,"L") == 0) {		/* Line	*/

			width = 1;
         if (!ReadVector(fptr,&(p[0])) || 
				 !ReadVector(fptr,&(p[1])) ||
             !ReadColour(fptr,&(c[0]))) {   
            fprintf(stderr,"Failed reading line (%d)\n",linecount);   
            fclose(fptr);  
            continue;
         }
			if (strcmp(id,"L") == 0) {
				if (fscanf(fptr,"%lf",&width) != 1) {
					fprintf(stderr,"Failed reading line width (%d)\n",linecount);
					continue;
				}
			}
			AddLine2Database(p[0],p[1],c[0],c[0],width);

      } else if (strcmp(id,"lc") == 0) {    /* Colour Line  */

         width = 1;
         if (!ReadVector(fptr,&(p[0])) ||
             !ReadVector(fptr,&(p[1])) ||
             !ReadColour(fptr,&(c[0])) ||
				 !ReadColour(fptr,&(c[1]))) {
            fprintf(stderr,"Failed reading colour line (%d)\n",linecount);
            fclose(fptr);
            continue;
         }
         AddLine2Database(p[0],p[1],c[0],c[1],width);

		} else if (strcmp(id,"f3") == 0) {					/* 3 face */

         if (!ReadVector(fptr,&(p[0])) ||
             !ReadVector(fptr,&(p[1])) ||
				 !ReadVector(fptr,&(p[2])) || 
             !ReadColour(fptr,&(c[0]))) {
            fprintf(stderr,"Failed reading face3 (%d)\n",linecount);
            fclose(fptr);
            continue;
         }
			AddFace2Database(p,3,c[0],1.0,origin);

      } else if (strcmp(id,"f3n") == 0) {              /* 3 face */

         if (!ReadVector(fptr,&(p[0])) ||
             !ReadVector(fptr,&(p[1])) ||
             !ReadVector(fptr,&(p[2])) ||
             !ReadVector(fptr,&(n[0])) ||
             !ReadVector(fptr,&(n[1])) ||
             !ReadVector(fptr,&(n[2])) ||
             !ReadColour(fptr,&(c[0]))) {
            fprintf(stderr,"Failed reading face3n (%d)\n",linecount);
            fclose(fptr);
            continue;
         }
         if ((face3 = realloc(face3,(nface3+1)*sizeof(FACE3))) == NULL) {
            fprintf(stderr,"realloc for face3n failed\n");
            exit(-1);
         }
			for (i=0;i<3;i++) {
         	face3[nface3].p[i] = p[i];
         	face3[nface3].n[i] = n[i];
         	face3[nface3].colour[i] = c[0];
			}
         nface3++;

      } else if (strcmp(id,"f3c") == 0) {              /* 3 face */

         if (!ReadVector(fptr,&(p[0])) ||
             !ReadVector(fptr,&(p[1])) ||
             !ReadVector(fptr,&(p[2])) ||
             !ReadColour(fptr,&(c[0])) ||
             !ReadColour(fptr,&(c[1])) ||
             !ReadColour(fptr,&(c[2]))) {
            fprintf(stderr,"Failed reading face3 (%d)\n",linecount);
            fclose(fptr);
            continue;
         }
         if ((face3 = realloc(face3,(nface3+1)*sizeof(FACE3)))==NULL) {
            fprintf(stderr,"realloc for face3 failed\n");
            exit(-1);
         }
         for (i=0;i<3;i++) {
            face3[nface3].p[i] = p[i];
            face3[nface3].n[i] = CalcNormal(p[0],p[1],p[2]);
            face3[nface3].colour[i] = c[i];
         }
         nface3++;

      } else if (strcmp(id,"f3nc") == 0) {              /* 3 face */

         if (!ReadVector(fptr,&(p[0])) ||
             !ReadVector(fptr,&(p[1])) ||
             !ReadVector(fptr,&(p[2])) ||
             !ReadVector(fptr,&(n[0])) ||
             !ReadVector(fptr,&(n[1])) ||
             !ReadVector(fptr,&(n[2])) ||
             !ReadColour(fptr,&(c[0])) ||
             !ReadColour(fptr,&(c[1])) ||
             !ReadColour(fptr,&(c[2]))) {
            fprintf(stderr,"Failed reading face3 (%d)\n",linecount);
            fclose(fptr);
            continue;
         }
         if ((face3 = realloc(face3,(nface3+1)*sizeof(FACE3)))==NULL) {
            fprintf(stderr,"realloc for face3 failed\n");
            exit(-1);
         }
			for (i=0;i<3;i++) {
         	face3[nface3].p[i] = p[i];
         	face3[nface3].n[i] = n[i];
         	face3[nface3].colour[i] = c[i];
			}
         nface3++;

		} else if (strcmp(id,"f4") == 0) {              /* 4 face */

         if (!ReadVector(fptr,&(p[0])) ||
             !ReadVector(fptr,&(p[1])) ||
             !ReadVector(fptr,&(p[2])) ||
				 !ReadVector(fptr,&(p[3])) ||
             !ReadColour(fptr,&(c[0]))) {
            fprintf(stderr,"Failed reading face4 (%d)\n",linecount);
            fclose(fptr);
            continue;
         }
			AddFace2Database(p,4,c[0],1.0,origin);

      } else if (strcmp(id,"f4n") == 0) {             /* 4 face, normal */

         if (!ReadVector(fptr,&(p[0])) ||
             !ReadVector(fptr,&(p[1])) ||
             !ReadVector(fptr,&(p[2])) ||
             !ReadVector(fptr,&(p[3])) ||
             !ReadVector(fptr,&(n[0])) ||
             !ReadVector(fptr,&(n[1])) ||
             !ReadVector(fptr,&(n[2])) ||
				 !ReadVector(fptr,&(n[3])) ||
             !ReadColour(fptr,&(c[0]))) {
            fprintf(stderr,"Failed reading face4 (%d)\n",linecount);
            fclose(fptr);
            continue;
         }
			if ((face4 = realloc(face4,(nface4+1)*sizeof(FACE4))) == NULL) {
            fprintf(stderr,"realloc for face4 failed\n");
            exit(-1);
         }
			for (i=0;i<4;i++) {
				face4[nface4].p[i] = p[i];
         	face4[nface4].n[i] = n[i];
         	face4[nface4].colour[i] = c[0];
			}
			nface4++;

      } else if (strcmp(id,"f4c") == 0) {            /* 4 face, col */

         if (!ReadVector(fptr,&(p[0])) ||
             !ReadVector(fptr,&(p[1])) ||
             !ReadVector(fptr,&(p[2])) ||
             !ReadVector(fptr,&(p[3])) ||
             !ReadColour(fptr,&(c[0])) ||
             !ReadColour(fptr,&(c[1])) ||
             !ReadColour(fptr,&(c[2])) ||
             !ReadColour(fptr,&(c[3]))) {
            fprintf(stderr,"Failed reading face4 (%d)\n",linecount);
            fclose(fptr);
            continue;
         }
         if ((face4 = realloc(face4,(nface4+1)*sizeof(FACE4)))==NULL) {
            fprintf(stderr,"realloc for face4 failed\n");
            exit(-1);
         }
         for (i=0;i<4;i++) {
            face4[nface4].p[i] = p[i];
            face4[nface4].n[i] = CalcNormal(p[(i-1+4)%4],p[i],p[(i+1)%4]);
            face4[nface4].colour[i] = c[i];
         }
         nface4++;

      } else if (strcmp(id,"f4nc") == 0) {            /* 4 face, normal, col */

         if (!ReadVector(fptr,&(p[0])) ||
             !ReadVector(fptr,&(p[1])) ||
             !ReadVector(fptr,&(p[2])) ||
             !ReadVector(fptr,&(p[3])) ||
             !ReadVector(fptr,&(n[0])) ||
             !ReadVector(fptr,&(n[1])) ||
             !ReadVector(fptr,&(n[2])) ||
             !ReadVector(fptr,&(n[3])) ||
				 !ReadColour(fptr,&(c[0])) ||
				 !ReadColour(fptr,&(c[1])) ||
				 !ReadColour(fptr,&(c[2])) ||
             !ReadColour(fptr,&(c[3]))) {
            fprintf(stderr,"Failed reading face4 (%d)\n",linecount);
            fclose(fptr);
            continue;
         }
			if ((face4 = realloc(face4,(nface4+1)*sizeof(FACE4)))==NULL) {
				fprintf(stderr,"realloc for face4 failed\n");
				exit(-1);
			}
			for (i=0;i<4;i++) {
         	face4[nface4].p[i] = p[i];
         	face4[nface4].n[i] = n[i];
         	face4[nface4].colour[i] = c[i];
			}
			nface4++;

      } else if (strcmp(id,"f4t") == 0) {             /* 4 face textured */

         if (!ReadVector(fptr,&(p[0])) ||
             !ReadVector(fptr,&(p[1])) ||
             !ReadVector(fptr,&(p[2])) ||
             !ReadVector(fptr,&(p[3])) ||
             !ReadColour(fptr,&(c[0]))) {
            fprintf(stderr,"Failed reading face4t (%d)\n",linecount);
            fclose(fptr);
            continue;
         }
			if (fscanf(fptr,"%s",texturename) != 1) {
            fprintf(stderr,"Failed reading texture name for face4t (%d)\n",linecount);
            fclose(fptr);
            continue;
			}
         if (fscanf(fptr,"%lf",&scale) != 1 || scale <= 0) {
            fprintf(stderr,"Failed reading texture scale for face4t (%d)\n",linecount);
            fclose(fptr);
            continue;
         }
         if (fscanf(fptr,"%s",trans) != 1) {
            fprintf(stderr,"Failed reading tex transparency for face4t (%d)\n",linecount);
            fclose(fptr);
            continue;
         }
         if ((face4t = realloc(face4t,(nface4t+1)*sizeof(FACE4T))) == NULL) {
            fprintf(stderr,"realloc for face4t failed\n");
            exit(-1);
         }
			for (i=0;i<4;i++)
         	face4t[nface4t].p[i] = p[i];
         face4t[nface4t].colour = c[0];
			strcpy(face4t[nface4t].texturename,texturename);
			face4t[nface4t].trans  = trans[0];
			face4t[nface4t].scale  = scale;
         gotitalready = -1;
         for (i=0;i<nface4t;i++) {
            if (strcmp(face4t[i].texturename,texturename) == 0) {
               gotitalready = i;
               break;
            }
         }
         if (gotitalready > 0) {
            face4t[nface4t].rgba = NULL;
            face4t[nface4t].textureid = face4t[gotitalready].textureid;
				if (options.debug)
					fprintf(stderr,"   Resusing texture \"%s\"\n",texturename);
         } else {
				face4t[nface4t].rgba = ReadTGATexture(texturename,&(face4t[nface4t].width),&(face4t[nface4t].height)); 
				glGenTextures(1,&(face4t[nface4t].textureid));
				if (options.debug)
					fprintf(stderr,"Texture: %s (%dx%d), id: %d\n",
						texturename,face4t[nface4t].width,face4t[nface4t].height,(int)face4t[nface4t].textureid);
         	glBindTexture(GL_TEXTURE_2D,face4t[nface4t].textureid);
      		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
      		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
				glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
      		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
         	glTexImage2D(GL_TEXTURE_2D,0,4,
         	   face4t[nface4t].width,face4t[nface4t].height,
         	   0,GL_RGBA,GL_UNSIGNED_BYTE,face4t[nface4t].rgba);
      		glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
			}
        	nface4t++;

		}

		// Skip everything to the end of the line, treat as a comment 
		while (strcmp(id,"t") != 0 && (d = fgetc(fptr)) != '\n' && d != '\r')
			;

		linecount++;
	}
	fclose(fptr);

	if (options.debug) {
		fprintf(stderr,"Read %d lines\n",linecount);
		fprintf(stderr,"Dot:     %d\n",ndot);
      fprintf(stderr,"Ball:    %d\n",nball);
		fprintf(stderr,"Ballt:   %d\n",nballt);
		fprintf(stderr,"Disk:    %d\n",ndisk);
		fprintf(stderr,"Cone:    %d\n",ncone);
		fprintf(stderr,"Lights:  %d\n",nlight);
      fprintf(stderr,"Line:    %d\n",nline);
		fprintf(stderr,"Face3:   %d\n",nface3);
      fprintf(stderr,"Face4:   %d\n",nface4);
		fprintf(stderr,"Face4t:  %d\n",nface4t);
		fprintf(stderr,"Label:   %d\n",nlabel);
	}

	if (linecount > 0)
		return(TRUE);
	else
		return(FALSE);
}

/*
	Read an OFF file
	An off file consists of 
	- number of vertices
	- the list of vertices, normals, colours
	- number of faces
	- the list of faces
*/
int ReadOFF(char *name)
{
   int i,j;
   FILE *fptr;
	int noffvert     = 0;
	XYZ *offvert     = NULL;
	XYZ *offnorm     = NULL;
	COLOUR *offcol   = NULL;
	int noffface     = 0;
	OFFFACE *offface = NULL;
   const char *fn;

	/* Get the file name */
	if (strlen(name) < 1) {
   	fn = fl_show_file_selector("OFF file","","","");
   	if (fn == NULL || strlen(fn) <= 0) 
      	return(TRUE);
   	if ((fptr = fopen(fn,"r")) == NULL) {
      	fl_show_message("File Error","Unable to open OFF file",fn);
      	return(FALSE);
   	}
	} else {
   	if ((fptr = fopen(name,"r")) == NULL) {
      	fl_show_message("File Error","Unable to open OFF file",name);
      	return(FALSE);
		}
   }

	/* Read the vertices */
	if (fscanf(fptr,"%d",&noffvert) != 1) {
		fprintf(stderr,"Failed to read the number of vertices\n");
		goto failed;
	}

	/* Allocate the storage for the vertices, colour, and normals */
	if ((offvert = malloc(noffvert*sizeof(XYZ))) == NULL) {
		fprintf(stderr,"Failed to allocate memory for the vertices\n");
		exit(-1);
	}
   if ((offnorm = malloc(noffvert*sizeof(XYZ))) == NULL) {
      fprintf(stderr,"Failed to allocate memory for the normals\n");
      exit(-1);
   }
   if ((offcol = malloc(noffvert*sizeof(COLOUR))) == NULL) {
      fprintf(stderr,"Failed to allocate memory for the colours\n");
      exit(-1);
   }

	/* Read the vertices, normals, colours */
	for (i=0;i<noffvert;i++) {
		if (fscanf(fptr,"%lf %lf %lf",
			&(offvert[i].x),&(offvert[i].y),&(offvert[i].z)) != 3) {
      	fprintf(stderr,"Failed to read the vertices\n");
      	exit(-1);
		}
      if (fscanf(fptr,"%lf %lf %lf",
         &(offnorm[i].x),&(offnorm[i].y),&(offnorm[i].z)) != 3) {
         fprintf(stderr,"Failed to read the normals\n");
         exit(-1);
      }
      if (fscanf(fptr,"%lf %lf %lf",
         &(offcol[i].r),&(offcol[i].g),&(offcol[i].b)) != 3) {
         fprintf(stderr,"Failed to read the colours\n");
         exit(-1);
      }
	}

	/* Read the number of faces */
   if (fscanf(fptr,"%d",&noffface) != 1) {
      fprintf(stderr,"Failed to read the number of faces\n");
      exit(-1);
   }

   /* Allocate the storage for the faces */
   if ((offface = malloc(noffface*sizeof(OFFFACE))) == NULL) {
      fprintf(stderr,"Failed to allocate memory for the faces\n");
      exit(-1);
   }

	/* Read the faces */
	for (i=0;i<noffface;i++) {
		if (fscanf(fptr,"%d",&(offface[i].nv)) != 1) {
      	fprintf(stderr,"Failed to read the number of vertices in a face\n");
      	exit(-1);
		}
		if (offface[i].nv < 1 || offface[i].nv > 4) {
			fprintf(stderr,"Illegal number of vertices in a face [1..4]\n");
			exit(-1);
		}
		for (j=0;j<offface[i].nv;j++) {
			if (fscanf(fptr,"%d",&(offface[i].p[j])) != 1) {
         	fprintf(stderr,"Failed to read a faces vertex index\n");
         	exit(-1);
			}
		}
	}

	/* Convert the off geometry */
	for (i=0;i<noffface;i++) {
		switch (offface[i].nv) {
		case 1:
         if ((dot = realloc(dot,(ndot+1)*sizeof(DOT))) == NULL) {
            fprintf(stderr,"realloc for dot failed\n");
            exit(-1);
         }
         dot[ndot].p = offvert[offface[i].p[0]];
         dot[ndot].colour = offcol[offface[i].p[0]];
			dot[ndot].size = 1;
         ndot++;
			break;
		case 2:
         if ((line = realloc(line,(nline+1)*sizeof(LINE))) == NULL) {
            fprintf(stderr,"realloc for line failed\n");
            exit(-1);
         }
         line[nline].p[0]   = offvert[offface[i].p[0]];
         line[nline].p[1]   = offvert[offface[i].p[1]];
         line[nline].colour[0] = offcol[offface[i].p[0]];
         line[nline].colour[1] = offcol[offface[i].p[0]];
			line[nline].width = 1;
         nline++;
			break;
		case 3:
         if ((face3 = realloc(face3,(nface3+1)*sizeof(FACE3)))==NULL) {
            fprintf(stderr,"realloc for face3 failed\n");
            exit(-1);
         }
			for (j=0;j<3;j++) {
         	face3[nface3].p[j]      = offvert[offface[i].p[j]];
         	face3[nface3].n[j]      = offnorm[offface[i].p[j]];
         	face3[nface3].colour[j] = offcol[offface[i].p[j]];
			}
         nface3++;
			break;
		case 4:
         if ((face4 = realloc(face4,(nface4+1)*sizeof(FACE4)))==NULL) {
            fprintf(stderr,"realloc for face4 failed\n");
            exit(-1);
         }
			for (j=0;j<4;j++) {
         	face4[nface4].p[j]      = offvert[offface[i].p[j]];
         	face4[nface4].n[j]      = offnorm[offface[i].p[j]];
         	face4[nface4].colour[j] = offcol[offface[i].p[j]];
			}
         nface4++;
			break;
		}
	}

	if (options.debug) {
		fprintf(stderr,"Read %d vertices\n",noffvert);
		fprintf(stderr,"Read %d faces\n",noffface);
	}

	return(TRUE);

	failed:
	if (offvert != NULL) free(offvert);
	if (offnorm != NULL) free(offnorm);
	if (offcol  != NULL) free(offcol);
	if (offface != NULL) free(offface);
	exit(-1);
}

/*
	Clean up the geometry
	face4's get decomposed to face3
	face3's get decomposed to lines
	lines to points
*/
void CleanGeometry(void)
{
	int i,j;
	double e;
	int badface3 = 0,badface4 = 0,badline = 0;
	int badnormal = 0;
	XYZ n;

	/* Deal with 4 vertex lines */
   for (i=0;i<nface4;i++) {
      for (j=0;j<4;j++) {
         e = VectorLength(face4[i].p[j],face4[i].p[(j+1)%4]);
         if (e < EPSILON) {
            if ((face3 = realloc(face3,(nface3+1)*sizeof(FACE3))) == NULL) {
               fprintf(stderr,"realloc for face3 failed\n");
               exit(-1);
            }
            face3[nface3].p[0]   = face4[i].p[j];
            face3[nface3].p[1]   = face4[i].p[(j+2)%4];
				face3[nface3].p[2]   = face4[i].p[(j+3)%4];
            face3[nface3].n[0]   = face4[i].n[j];
            face3[nface3].n[1]   = face4[i].n[(j+2)%4];
            face3[nface3].n[2]   = face4[i].n[(j+3)%4];
            face3[nface3].colour[0] = face4[i].colour[j];
            face3[nface3].colour[1] = face4[i].colour[(j+2)%4];
            face3[nface3].colour[2] = face4[i].colour[(j+3)%4];
            nface3++;
            if (nface4 > 1)
               face4[i] = face4[nface4-1];
            nface4--;
            i--;
            badface4++;
				break;
         }
      }
	}

	/* Fix up face3 */
	for (i=0;i<nface3;i++) {
		for (j=0;j<3;j++) {
			e = VectorLength(face3[i].p[j],face3[i].p[(j+1)%3]);
			if (e < EPSILON) {
        		if ((line = realloc(line,(nline+1)*sizeof(LINE))) == NULL) {
            	fprintf(stderr,"realloc for line failed\n");
            	exit(-1);
         	}
         	line[nline].width  = 1;
         	line[nline].p[0]   = face3[i].p[j];
         	line[nline].p[1]   = face3[i].p[(j+2)%3];
         	line[nline].colour[0] = face3[i].colour[j];
            line[nline].colour[1] = face3[i].colour[j];
				nline++;
				if (nface3 > 1)
					face3[i] = face3[nface3-1];
				nface3--;
				i--;
				badface3++;
				break;
			}
		} 
	}

	/* Degenerate lines are turned into dots */
	for (i=0;i<nline;i++) {
		e = VectorLength(line[i].p[0],line[i].p[1]);
		if (e < EPSILON) {
			if ((dot = realloc(dot,(ndot+1)*sizeof(DOT))) == NULL) {
            fprintf(stderr,"realloc for dot failed\n");
            exit(-1);
         }
         dot[ndot].size   = line[i].width;
         dot[ndot].p      = line[i].p[0];
         dot[ndot].colour = line[i].colour[0];
         ndot++;
			if (nline > 1)
         	line[i] = line[nline-1];
         nline--;
         i--;
         badline++;
		}
	}

	/* Check for any null normals */
	for (i=0;i<nface3;i++) {
		for (j=0;j<3;j++) {
			if (Modulus(face3[i].n[j]) < EPSILON) {
				n = CalcNormal(face3[i].p[j],face3[i].p[(j+1)%3],face3[i].p[(j+2)%3]);
				face3[i].n[j] = n;
				badnormal++;
			}
		}
	}
   for (i=0;i<nface4;i++) {
      for (j=0;j<4;j++) {
         if (Modulus(face4[i].n[j]) < EPSILON) {
            n = CalcNormal(face4[i].p[j],face4[i].p[(j+1)%4],face4[i].p[(j+3)%4]);
            face4[i].n[j] = n;
            badnormal++;
         }
      }
   }

	if (options.debug) {
		fprintf(stderr,"Simplified face4s = %d\n",badface4);
		fprintf(stderr,"Simplified face3s = %d\n",badface3);
		fprintf(stderr,"Simplified lines  = %d\n",badline);
		fprintf(stderr,"Bad normals       = %d\n",badnormal);
	}
}

/*
	Delete/free all the data structures
*/
void DeleteGeometry(void)
{
	int i;

   ndot     = 0; if (dot     != NULL) { free(dot);     dot     = NULL; }
   nline    = 0; if (line    != NULL) { free(line);    line    = NULL; }
   nface3   = 0; if (face3   != NULL) { free(face3);   face3   = NULL; }
   nface4   = 0; if (face4   != NULL) { free(face4);   face4   = NULL; }
	nball    = 0; if (ball    != NULL) { free(ball);    ball    = NULL; }
	ndisk    = 0; if (disk    != NULL) { free(disk);    disk    = NULL; }
	ncone    = 0; if (cone    != NULL) { free(cone);    cone    = NULL; }
	nlabel   = 0; if (label   != NULL) { free(label);   label   = NULL; }

	/* Remove texture facets */
	for (i=0;i<nface4t;i++) {
		glDeleteTextures(1,&(face4t[i].textureid));
		if (face4t[i].rgba != NULL) { 	
			free(face4t[i].rgba); 
			face4t[i].rgba = NULL; 
		}
	}
	nface4t = 0; 
	if (face4t != NULL) { 
		free(face4t); 
		face4t = NULL; 
	}

	/* Remove texture balls */
   for (i=0;i<nballt;i++) {
		glDeleteTextures(1,&(ballt[i].textureid));
      if (ballt[i].rgba != NULL) {
         free(ballt[i].rgba);
         ballt[i].rgba = NULL;
      }
   }
   nballt = 0;
   if (ballt != NULL) {
      free(ballt);
      ballt = NULL;
   }

	MakeGeometry(TRUE);
}

/*
	Set the camera to the "home" position (mode == 1)
	Focus the camera to the center of the geometry (mode = 2)
	Move to a "axis aligned" position (mode 3..8)
*/
void CameraHome(int mode)
{
	double size;

	// Stop any forward movement 
   camera.speed = 0;

	// Find the bounds of the geometry 
	CalcBounds();

	size = MAX((pmax.x-pmin.x),(pmax.y-pmin.y));
	size = MAX(size,(pmax.z-pmin.z));

	// Set the camera attributes 
	camera.pr = pmid;
	switch (mode) {
	case HOME:
		camera.focallength = 1.75*size;
		camera.aperture = 45;
   	camera.eyesep = size / 20.0;
		camera.vp = pmid;
		camera.vp.z += 1.75*size;
      camera.vu.x = 0; camera.vu.y = 1; camera.vu.z = 0;
		break;
	case FOCUS:
		break;
   case FRONT:
		camera.vp = pmid;
      camera.vp.z += 1.75 * size;
      camera.vu.x = 0; camera.vu.y = 1; camera.vu.z = 0;
		break;
   case BACK:
		camera.vp = pmid;
      camera.vp.z -= 1.75 * size;
      camera.vu.x = 0; camera.vu.y = 1; camera.vu.z = 0;
		break;
	case LEFT:
		camera.vp = pmid;
		camera.vp.x -= 1.75 * size;
		camera.vu.x = 0; camera.vu.y = 1; camera.vu.z = 0;
		break;
   case RIGHT:
		camera.vp = pmid;
      camera.vp.x += 1.75 * size;
		camera.vu.x = 0; camera.vu.y = 1; camera.vu.z = 0;
		break;
   case TOP:
		camera.vp = pmid;
      camera.vp.y += 1.75 * size;
      camera.vu.x = 0; camera.vu.y = 0; camera.vu.z = 1;
		break;
   case BOTTOM:
		camera.vp = pmid;
      camera.vp.y -= 1.75 * size;
      camera.vu.x = 0; camera.vu.y = 0; camera.vu.z = 1;
		break;
	case AXONOMETRIC:
		camera.vp = pmid;
      camera.vp.x -= size;
      camera.vp.y += size;
      camera.vp.z -= size;
      camera.vu.x = 0; camera.vu.y = 1; camera.vu.z = 0;
		break;
	}

	// Point the camera toward the mid point of the model 
   camera.vd.x = pmid.x - camera.vp.x;
   camera.vd.y = pmid.y - camera.vp.y;
   camera.vd.z = pmid.z - camera.vp.z;

	// Update the focal point 
   Normalise(&camera.vd);
   camera.focus.x = camera.vp.x + camera.focallength * camera.vd.x;
   camera.focus.y = camera.vp.y + camera.focallength * camera.vd.y;
   camera.focus.z = camera.vp.z + camera.focallength * camera.vd.z;

	// Update the slider bounds 
	fl_set_slider_bounds(cameraform->Eyesep,0.0,size/5);
	fl_set_slider_bounds(cameraform->Aperture,0.0,90.0);
   fl_set_slider_bounds(cameraform->Focallength,size/2,6*size);
	UpdateForm();

	if (options.debug) {
		fprintf(stderr,"Bounds = (%g,%g,%g) -> (%g,%g,%g)\n",
			pmin.x,pmin.y,pmin.z,pmax.x,pmax.y,pmax.z);
		fprintf(stderr,"Midpoint = (%g,%g,%g)\n",pmid.x,pmid.y,pmid.z);
	}
}

/*
	Set the auto camera position
	The first time, get the file name and read the flight path
	See AppendCameraPosition() for the order of the camera attributes.
		vp,vd,vu,focallength,aperture,eyesep
*/
int AutoPilot(int mode,char *animfile)
{
   const char *fn;
   char fname[256];
	static FILE *fptr = NULL;
	static int cycle = 0;
	XYZ vp,vd,vu;
	double focallength,eyesep,aperture;
	
	if (!options.autopilot)
		return(FALSE);

	/* Are we getting a new flight path ? */
	if (mode == 0) {

		/* If we already have one then close the file */
      if (fptr != NULL)
         fclose(fptr);

		/* Ask the user for the file name or was it passed to us? */
		if (strlen(animfile) < 1) {
   		fn = fl_show_file_selector("Flight path file","","","");
   		if (fn == NULL || strlen(fn) <= 0) {
      		fname[0] = '\0';
				fptr = NULL;
      		return(FALSE);
   		}
   		strcpy(fname,fn);
		} else {
			strcpy(fname,animfile);
		}

		/* Attempt to open the file */
		if ((fptr = fopen(fname,"r")) == NULL) {
			fptr = NULL;
			return(FALSE);
		}
	}

	/* Get the next camera position */
	tryagain:
	if (fscanf(fptr,"%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",
		&vp.x,&vp.y,&vp.z, 
		&vd.x,&vd.y,&vd.z, 
		&vu.x,&vu.y,&vu.z,
		&focallength,&aperture,&eyesep) != 12) {
		if (cycle > 0) {
			rewind(fptr);
			cycle = 0;
			goto tryagain;
		} else {
			fclose(fptr);
			fptr = NULL;
			return(FALSE);
		}
	}
	cycle++;
	Normalise(&vd);
	Normalise(&vu);
	camera.vp = vp;
	camera.vd = vd;
	camera.vu = vu;
	camera.focallength = focallength;
   camera.focus.x = camera.vp.x + camera.focallength * camera.vd.x;
   camera.focus.y = camera.vp.y + camera.focallength * camera.vd.y;
   camera.focus.z = camera.vp.z + camera.focallength * camera.vd.z;
	camera.aperture = aperture;
	camera.eyesep = eyesep;

	return(TRUE);
}

/*
	Update the form contents
*/
void UpdateForm(void)
{
	char s[128];

	/* Camera form */
   fl_set_slider_bounds(cameraform->Eyesep,0.0,camera.focallength/2);
   fl_set_slider_bounds(cameraform->Aperture,0.0,90.0);
   fl_set_slider_bounds(cameraform->Focallength,
      camera.focallength/2,4*camera.focallength);
	fl_set_slider_value(cameraform->Eyesep,camera.eyesep);
	fl_set_slider_value(cameraform->Aperture,camera.aperture);
	fl_set_slider_value(cameraform->Focallength,camera.focallength);

	/* Materials form */
   fl_set_slider_bounds(materialform->Shininess,1.0,200.0);
   fl_set_slider_bounds(materialform->Emission,0.0,1.0);
   fl_set_slider_bounds(materialform->Transparency,0.0,1.0);
	sprintf(s,"%g",specularcolour[0]);
	fl_set_input(materialform->Red,s);
	sprintf(s,"%g",specularcolour[1]);
   fl_set_input(materialform->Green,s);
	sprintf(s,"%g",specularcolour[2]);
   fl_set_input(materialform->Blue,s);
   fl_set_slider_value(materialform->Shininess,shininess[0]);
	fl_set_slider_value(materialform->Emission,emission[0]);
	fl_set_slider_value(materialform->Transparency,transparency);

	/* Light */
   fl_set_slider_bounds(lightform->Ambient,0.0,1.0);
   fl_set_button(lightform->LocalViewer,options.localviewer);
   fl_set_button(lightform->TwoSided,options.twosided);
   fl_set_counter_value(lightform->DefaultLights,defaultlights);
   fl_set_counter_bounds(lightform->DefaultLights,0.0,8.0);
   fl_set_counter_step(lightform->DefaultLights,1.0,1.0);
   fl_set_counter_precision(lightform->DefaultLights,0);
	fl_set_slider_value(lightform->Ambient,globalambient[0]);

	/* Appearance */
   sprintf(s,"%g",options.pointscale);
   fl_set_input(appearform->PointScale,s);
   sprintf(s,"%g",options.linescale);
   fl_set_input(appearform->LineScale,s);
   sprintf(s,"%g",options.scalefactor.x);
   fl_set_input(appearform->Xscale,s);
   sprintf(s,"%g",options.scalefactor.y);
   fl_set_input(appearform->Yscale,s);
   sprintf(s,"%g",options.scalefactor.z);
   fl_set_input(appearform->Zscale,s);
}

/*
	Create a pair of axes as the initial scene
	No error checking....sorry.
*/
void CreateAxes(void)
{
	COLOUR red = {1,0,0},green = {0,1,0}, blue = {0,0,1};

	nline = 3;
	line = malloc(nline*sizeof(LINE));
	line[0].p[0].x   = -1; line[0].p[0].y   = 0;  line[0].p[0].z   = 0;
   line[0].p[1].x   =  1; line[0].p[1].y   = 0;  line[0].p[1].z   = 0;
	line[0].width = 1;
	line[0].colour[0] = red;
   line[0].colour[1] = red;
   line[1].p[0].x   = 0;  line[1].p[0].y   = -1; line[1].p[0].z   = 0;
   line[1].p[1].x   = 0;  line[1].p[1].y   =  1; line[1].p[1].z   = 0;
	line[1].width = 1;
   line[1].colour[0] = green;
	line[1].colour[1] = green;
   line[2].p[0].x   = 0;  line[2].p[0].y   = 0;  line[2].p[0].z   = -1;
   line[2].p[1].x   = 0;  line[2].p[1].y   = 0;  line[2].p[1].z   =  1;
	line[2].width = 1;
   line[2].colour[0] = blue;
	line[2].colour[1] = blue;
}

/*
	Save attributes in a tagged type format
	Do nothing if the file can't be opened
	See ReadViewFile()
*/
void SaveViewFile(CAMERA c)
{
	const char *fn;
	char fname[256];
	FILE *fptr;
	
	/* Get the file name */
	fn = fl_show_file_selector("View file","","","");
   if (fn == NULL || strlen(fn) <= 0) 
      return;
	strcpy(fname,fn);

	/* Try to open the file */
   if ((fptr = fopen(fname,"w")) == NULL) 
		return;

	/* Write the attributes */
	fprintf(fptr,"vp %g %g %g\n",c.vp.x,c.vp.y,c.vp.z);
	fprintf(fptr,"vd %g %g %g\n",c.vd.x,c.vd.y,c.vd.z);
	fprintf(fptr,"vu %g %g %g\n",c.vu.x,c.vu.y,c.vu.z);
	fprintf(fptr,"pr %g %g %g\n",c.pr.x,c.pr.y,c.pr.z);
	fprintf(fptr,"focallength %g\n",c.focallength);
	fprintf(fptr,"aperture %g\n",c.aperture);
	fprintf(fptr,"eyeseparation %g\n",c.eyesep);
	if (options.showboundingbox)
		fprintf(fptr,"-bb\n");
	fprintf(fptr,"-sh %g\n",shininess[0]);
	if (options.deltamove > 0)
		fprintf(fptr,"-dm %g\n",options.deltamove);
	fprintf(fptr,"-bg %g %g %g\n",options.background.r,options.background.g,options.background.b);
	fclose(fptr);
}

/*
	Read the saved settings
	Do nothing if the file open failed
	Since tags are used, an attribute that doesn't exist in the
   file won't be changed.
*/
void ReadViewFile(char *name)
{
   const char *fn;
   char fname[256];
	char s[64];
   FILE *fptr;

	/* Get the file name if one wasn't provided */
	if (strlen(name) < 1) {
   	fn = fl_show_file_selector("View file","","","");
   	if (fn == NULL || strlen(fn) <= 0) {
			fprintf(stderr,"Failed to open view file \"%s\"\n",name);
      	return;
		}
   	strcpy(fname,fn);
	} else {
		strcpy(fname,name);
	}

	/* Try to open the file */
   if ((fptr = fopen(fname,"r")) == NULL) 
      return;

	/* Read the tags and associated data */
	while (fscanf(fptr,"%s",s) == 1) {
		if (strcmp(s,"vp") == 0)
			fscanf(fptr,"%lf %lf %lf",&camera.vp.x,&camera.vp.y,&camera.vp.z);
   	if (strcmp(s,"vd") == 0)
      	fscanf(fptr,"%lf %lf %lf",&camera.vd.x,&camera.vd.y,&camera.vd.z);
   	if (strcmp(s,"vu") == 0)
      	fscanf(fptr,"%lf %lf %lf",&camera.vu.x,&camera.vu.y,&camera.vu.z);
   	if (strcmp(s,"pr") == 0)
      	fscanf(fptr,"%lf %lf %lf",&camera.pr.x,&camera.pr.y,&camera.pr.z);
   	if (strcmp(s,"focallength") == 0)
      	fscanf(fptr,"%lf",&camera.focallength);
      if (strcmp(s,"aperture") == 0)
         fscanf(fptr,"%lf",&camera.aperture);
      if (strcmp(s,"eyeseparation") == 0)
         fscanf(fptr,"%lf",&camera.eyesep);
		if (strcmp(s,"-bg") == 0)
			fscanf(fptr,"%lf %lf %lf",&options.background.r,&options.background.g,&options.background.b);
		if (strcmp(s,"-dm") == 0)
			fscanf(fptr,"%lf",&options.deltamove);
		if (strcmp(s,"-sh") == 0)
			fscanf(fptr,"%f",&(shininess[0]));
      if (strcmp(s,"-bb") == 0)
         options.showboundingbox = TRUE;
		options.camerahome = camera;
	}
   fclose(fptr);

	if (options.debug) {
		fprintf(stderr,"vp = %g %g %g\n",camera.vp.x,camera.vp.y,camera.vp.z);
		fprintf(stderr,"vd = %g %g %g\n",camera.vd.x,camera.vd.y,camera.vd.z);
		fprintf(stderr,"vu = %g %g %g\n",camera.vu.x,camera.vu.y,camera.vu.z);
		fprintf(stderr,"pr = %g %g %g\n",camera.pr.x,camera.pr.y,camera.pr.z);
		fprintf(stderr,"focal length = %g\n",camera.focallength);
		fprintf(stderr,"aperture = %g\n",camera.aperture);
		fprintf(stderr,"eye separation = %g\n",camera.eyesep);
	}

	Normalise(&camera.vd);
   camera.focus.x = camera.vp.x + camera.focallength * camera.vd.x;
   camera.focus.y = camera.vp.y + camera.focallength * camera.vd.y;
   camera.focus.z = camera.vp.z + camera.focallength * camera.vd.z;
}

/*
	Append the camera attributes to "camera.path"
	Do nothing if the file can't be opened
*/
void AppendCameraPosition(void)
{
	FILE *fptr;

   if ((fptr = fopen("camera.path","a")) != NULL) {
      fprintf(fptr,"%g %g %g ",camera.vp.x,camera.vp.y,camera.vp.z);
      fprintf(fptr,"%g %g %g ",camera.vd.x,camera.vd.y,camera.vd.z);
      fprintf(fptr,"%g %g %g ",camera.vu.x,camera.vu.y,camera.vu.z);
		fprintf(fptr,"%g ",camera.focallength);
		fprintf(fptr,"%g ",camera.aperture);
		fprintf(fptr,"%g\n",camera.eyesep);
      fclose(fptr);
   }
}

/*
	Display extra stuff such as
	- help text
	- bounding box
*/
void DrawExtras(void)
{
	int i,y = 10;
	XYZ p;
	char s[64];
	double delta;
	COLOUR colour = {0.0,0.0,0.0};

   // Inverse colour for the text
   colour.r = 1 - options.background.r;
   colour.g = 1 - options.background.g;
   colour.b = 1 - options.background.b;
   glColor3f(colour.r,colour.g,colour.b);

	// Turn off OpenGL stuff
	glLoadName(-1);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);

   // Draw the light sources in debug mode 
   if (options.debug) {
      for (i=0;i<nlight;i++) {
         p.x = lights[i].p[0];
         p.y = lights[i].p[1];
         p.z = lights[i].p[2];
			Draw3DPlus(p,VectorLength(pmin,pmax)/200);
      }
		if (nlight <= 0) {
			delta = VectorLength(pmin,pmax) / 20;
			for (i=0;i<defaultlights;i++) {
				if (deflightpos[i].x == 0)
					p.x = pmin.x - delta;
				else
					p.x = pmax.x + delta;
            if (deflightpos[i].y == 0)
               p.y = pmin.y - delta;
            else
               p.y = pmax.y + delta;
            if (deflightpos[i].z == 0)
               p.z = pmin.z - delta;
            else
               p.z = pmax.z + delta;
				Draw3DPlus(p,VectorLength(pmin,pmax)/200);
			}
		}
   }

   // Are we showing text help information? 
   if (options.showhelpinfo) 
      DrawGLText(10,options.screenheight - 20,keyboardstring);

	if (options.debug) {
   	sprintf(s,"Frame rate: %.1f frames/second",interfacestate.framerate);
   	DrawGLText(10,y+=13,s);
		sprintf(s,"vp: %g %g %g",camera.vp.x,camera.vp.y,camera.vp.z);
		DrawGLText(10,y+=13,s);
      sprintf(s,"vd: %g %g %g",camera.vd.x,camera.vd.y,camera.vd.z);
      DrawGLText(10,y+=13,s);
      sprintf(s,"vu: %g %g %g",camera.vu.x,camera.vu.y,camera.vu.z);
      DrawGLText(10,y+=13,s);
      sprintf(s,"aperture: %g",camera.aperture);
      DrawGLText(10,y+=13,s);
      sprintf(s,"focal length: %g",camera.focallength);
      DrawGLText(10,y+=13,s);
		sprintf(s,"eye sep: %g",camera.eyesep);
      DrawGLText(10,y+=13,s);
	}

	glEnable(GL_DEPTH_TEST);
}

/*
	Add a line segment to the line database
*/
void AddLine2Database(XYZ p1,XYZ p2,COLOUR c1,COLOUR c2,double w)
{
   if ((line = realloc(line,(nline+1)*sizeof(LINE))) == NULL) {
      fprintf(stderr,"realloc for line failed\n");
      exit(-1);
   }
	w = ABS(w);
   line[nline].width = w;
   line[nline].p[0] = p1;
   line[nline].p[1] = p2;
   line[nline].colour[0] = c1;
   line[nline].colour[1] = c2;
   nline++;
}

/*
	Add a simple face to the database
	Scale and then translate it
*/
void AddFace2Database(XYZ *p,int n,COLOUR c,double scale,XYZ shift)
{
	int i;

	if (n == 3) {
      if ((face3 = realloc(face3,(nface3+1)*sizeof(FACE3))) == NULL) {
         fprintf(stderr,"realloc for face3 failed\n");
         exit(-1);
      }
      for (i=0;i<3;i++) {
         face3[nface3].p[i].x = shift.x + scale*p[i].x;
         face3[nface3].p[i].y = shift.y + scale*p[i].y;
         face3[nface3].p[i].z = shift.z + scale*p[i].z;
         face3[nface3].n[i] = CalcNormal(p[0],p[1],p[2]);
         face3[nface3].colour[i] = c;
      }
      nface3++;
	} else if (n == 4) {
      if ((face4 = realloc(face4,(nface4+1)*sizeof(FACE4))) == NULL) {
         fprintf(stderr,"realloc for face4 failed\n");
         exit(-1);
      }
      for (i=0;i<4;i++) {
         face4[nface4].p[i].x = shift.x + scale*p[i].x;
         face4[nface4].p[i].y = shift.y + scale*p[i].y;
         face4[nface4].p[i].z = shift.z + scale*p[i].z;
         face4[nface4].n[i] = CalcNormal(p[(i-1+4)%4],p[i],p[(i+1)%4]);
         face4[nface4].colour[i] = c;
      }
      nface4++;
	}
}

/*
	Add a marker to the database
*/
void AddMarker2Database(int type,double size,XYZ p,COLOUR c)
{
	XYZ q[4];
	XYZ tetra[4] = {{ 1, 1, 1},  {-1,-1, 1},   {1,-1,-1},  {-1, 1,-1}};
	XYZ box[8]   = {{-1,-1, 1},  {1,-1, 1},    {1,-1,-1},  {-1,-1,-1},
					    {-1, 1, 1},  {1, 1, 1},    {1, 1,-1},  {-1, 1,-1}};
	XYZ octa[6]  = {{-1,0,1},  {1,0,1},  {1,0,-1},  {-1,0,-1},  
				       {0,-1.41421356,0},  {0,1.41421356,0}};

   switch (type) {
   case 0:
		q[0] = tetra[0]; q[1] = tetra[1]; q[2] = tetra[2]; 
		AddFace2Database(q,3,c,size/2,p);
      q[0] = tetra[1]; q[1] = tetra[3]; q[2] = tetra[2]; 
      AddFace2Database(q,3,c,size/2,p);
      q[0] = tetra[0]; q[1] = tetra[2]; q[2] = tetra[3]; 
      AddFace2Database(q,3,c,size/2,p);
      q[0] = tetra[0]; q[1] = tetra[3]; q[2] = tetra[1]; 
      AddFace2Database(q,3,c,size/2,p);
      break;
   case 1:
		q[0] = p; q[1] = p;
		q[0].x -= size/2; q[1].x += size/2;
      AddLine2Database(q[0],q[1],c,c,1.0);
      q[0] = p; q[1] = p; 
      q[0].y -= size/2; q[1].y += size/2;
      AddLine2Database(q[0],q[1],c,c,1.0);
      q[0] = p; q[1] = p;  
      q[0].z -= size/2; q[1].z += size/2;
      AddLine2Database(q[0],q[1],c,c,1.0);
      break;
   case 2:
      q[0] = box[0]; q[1] = box[1]; q[2] = box[2]; q[3] = box[3];    /* Bottom */
      AddFace2Database(q,4,c,size/2,p);
      q[0] = box[4]; q[1] = box[7]; q[2] = box[6]; q[3] = box[5];    /* Top */
      AddFace2Database(q,4,c,size/2,p);
      q[0] = box[0]; q[1] = box[4]; q[2] = box[5]; q[3] = box[1];    /* Front */
      AddFace2Database(q,4,c,size/2,p);
      q[0] = box[3]; q[1] = box[2]; q[2] = box[6]; q[3] = box[7];    /* Back */
      AddFace2Database(q,4,c,size/2,p);
      q[0] = box[1]; q[1] = box[5]; q[2] = box[6]; q[3] = box[2];    /* Left */
      AddFace2Database(q,4,c,size/2,p);
      q[0] = box[0]; q[1] = box[3]; q[2] = box[7]; q[3] = box[4];    /* Right */
      AddFace2Database(q,4,c,size/2,p);
      break;
	case 3:
      q[0] = octa[0]; q[1] = octa[5]; q[2] = octa[1];    /* Top half */
      AddFace2Database(q,3,c,size/2,p);
      q[0] = octa[1]; q[1] = octa[5]; q[2] = octa[2];
      AddFace2Database(q,3,c,size/2,p);
      q[0] = octa[2]; q[1] = octa[5]; q[2] = octa[3];
      AddFace2Database(q,3,c,size/2,p);
      q[0] = octa[3]; q[1] = octa[5]; q[2] = octa[0];
      AddFace2Database(q,3,c,size/2,p);
      q[0] = octa[0]; q[1] = octa[4]; q[2] = octa[1];    /* Bottom half */
      AddFace2Database(q,3,c,size/2,p);
      q[0] = octa[1]; q[1] = octa[4]; q[2] = octa[2];
      AddFace2Database(q,3,c,size/2,p);
      q[0] = octa[2]; q[1] = octa[4]; q[2] = octa[3];
      AddFace2Database(q,3,c,size/2,p);
      q[0] = octa[3]; q[1] = octa[4]; q[2] = octa[0];
      AddFace2Database(q,3,c,size/2,p);
		break;
   }
}

/*
	Calculate the model bounds
	Update pmin,pmax,pmid
	Optionally update rangemin,rangemax
*/
void CalcBounds(void)
{
	int i;
	XYZ p;

   pmin.x = 1e32;  pmin.y = 1e32;  pmin.z = 1e32;
   pmax.x = -1e32; pmax.y = -1e32; pmax.z = -1e32;
	rangemin = 1e32;
	rangemax = -1e32;

   for (i=0;i<ndot;i++)
		UpdateBounds(dot[i].p);
   for (i=0;i<nball;i++) {
      p.x = ball[i].p.x + ball[i].r;
      p.y = ball[i].p.y + ball[i].r;
      p.z = ball[i].p.z + ball[i].r;
      UpdateBounds(p);
      p.x = ball[i].p.x - ball[i].r;
      p.y = ball[i].p.y - ball[i].r;
      p.z = ball[i].p.z - ball[i].r;
      UpdateBounds(p);
   }
   for (i=0;i<nballt;i++) {
      p.x = ballt[i].p.x + ballt[i].r;
      p.y = ballt[i].p.y + ballt[i].r;
      p.z = ballt[i].p.z + ballt[i].r;
      UpdateBounds(p);
      p.x = ballt[i].p.x - ballt[i].r;
      p.y = ballt[i].p.y - ballt[i].r;
      p.z = ballt[i].p.z - ballt[i].r;
      UpdateBounds(p);
   }
   for (i=0;i<ndisk;i++) {
      p.x = disk[i].p.x + disk[i].r2;
      p.y = disk[i].p.y + disk[i].r2;
      p.z = disk[i].p.z + disk[i].r2;
      UpdateBounds(p);
      p.x = disk[i].p.x - disk[i].r2;
      p.y = disk[i].p.y - disk[i].r2;
      p.z = disk[i].p.z - disk[i].r2;
      UpdateBounds(p);
   }
   for (i=0;i<ncone;i++) {
      p.x = cone[i].p1.x + cone[i].r1;
      p.y = cone[i].p1.y + cone[i].r1;
      p.z = cone[i].p1.z + cone[i].r1;
      UpdateBounds(p);
      p.x = cone[i].p1.x - cone[i].r1;
      p.y = cone[i].p1.y - cone[i].r1;
      p.z = cone[i].p1.z - cone[i].r1;
		UpdateBounds(p);

      p.x = cone[i].p2.x + cone[i].r2;
      p.y = cone[i].p2.y + cone[i].r2;
      p.z = cone[i].p2.z + cone[i].r2;
      UpdateBounds(p);
      p.x = cone[i].p2.x - cone[i].r2;
      p.y = cone[i].p2.y - cone[i].r2;
      p.z = cone[i].p2.z - cone[i].r2;
      UpdateBounds(p);
   }
   for (i=0;i<nline;i++) {
      UpdateBounds(line[i].p[0]);
      UpdateBounds(line[i].p[1]);
   }
   for (i=0;i<nface3;i++) {
      UpdateBounds(face3[i].p[0]);
      UpdateBounds(face3[i].p[1]);
      UpdateBounds(face3[i].p[2]);
   }
   for (i=0;i<nface4;i++) {
      UpdateBounds(face4[i].p[0]);
      UpdateBounds(face4[i].p[1]);
      UpdateBounds(face4[i].p[2]);
      UpdateBounds(face4[i].p[3]);
   }
   for (i=0;i<nface4t;i++) {
      UpdateBounds(face4t[i].p[0]);
      UpdateBounds(face4t[i].p[1]);
      UpdateBounds(face4t[i].p[2]);
      UpdateBounds(face4t[i].p[3]);
   }
   for (i=0;i<nlabel;i++)
      UpdateBounds(label[i].p);

   if (pmax.x <= pmin.x) { pmax.x = pmin.x; pmax.x += 0.01; pmin.x -= 0.01; }
   if (pmax.y <= pmin.y) { pmax.y = pmin.y; pmax.y += 0.01; pmin.y -= 0.01; }
   if (pmax.z <= pmin.z) { pmax.z = pmin.z; pmax.z += 0.01; pmin.z -= 0.01; }
	if (rangemin >= rangemax) { 
		rangemax = rangemin; 
		rangemin -= 0.01; 
		rangemax += 0.01; 
	}

   /* The center of the model */
   pmid.x = (pmin.x + pmax.x) / 2;
   pmid.y = (pmin.y + pmax.y) / 2;
   pmid.z = (pmin.z + pmax.z) / 2;

   if (options.debug)
      fprintf(stderr,"Range: %g to %g\n",rangemin,rangemax);
}

void UpdateBounds(XYZ p)
{
	double dist;

   pmin.x = MIN(pmin.x,p.x);
   pmin.y = MIN(pmin.y,p.y);
   pmin.z = MIN(pmin.z,p.z);
   pmax.x = MAX(pmax.x,p.x);
   pmax.y = MAX(pmax.y,p.y);
   pmax.z = MAX(pmax.z,p.z);
	
	dist = VectorLength(p,camera.vp);
	rangemin = MIN(rangemin,dist);
	rangemax = MAX(rangemax,dist);
}

void CleanExit(void)
{
	glutDestroyWindow(glutGetWindow()); 
   ClearAllBuffers();
	exit(0);
}

