#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#if defined(__linux__)
   #include <GL/glut.h>
#elif defined(__APPLE__)
   #include <GLUT/glut.h>
#else
   #warning "Not sure where to find glut.h"
#endif
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/file.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/times.h>

#include "paulslib.h"
#include "bitmaplib.h"
#include "opengllib.h"
#include "forms.h"

#define NOSTEREO     0
#define ACTIVESTEREO 1
#define DUALSTEREO   2

// Projection
#define PERSPECTIVE  0
#define ORTHOGRAPHIC 1
#define FISHEYE      2

// Output display (projector/lens configuration)
#define NODOME      0
#define TRUNCBOTTOM 1
#define TRUNCTOP    2
#define VSPHERICAL  3
#define HSPHERICAL  4
#define WARPMAP     5
#define MIRROR1     6

// Warp map types
#define POLAR 1
#define RECTANGULAR 2

#define TEXTURESIZE 512

// One triangle 
typedef struct {
   double u[3],v[3]; // Texture coordinates
   XYZ p[3];         // Vertices 
   XYZ wp[3];        // Vertices for warpmap
   double wi[3];     // Intensity of node, for warpmap
} DOMEFACE;

typedef struct {
   double x,y;         // Coordinate
   double u,v;         // UV coordinate
   double i;           // Intensity
} MESHNODE;

typedef struct {
   XYZ vp;           	/* View position            */
   XYZ vd;           	/* View direction vector    */
   XYZ vu;           	/* View up direction        */
	XYZ pr;					/* Point to rotate about 	 */
   double focallength;  /* Focal Length along vd	 */
	XYZ focus;           /* Focal point - Derived	 */
   double aperture;  	/* Camera VERTICAL aperture */
	double eyesep;			/* Eye separation				 */
	double speed;        /* Speed in vd direction    */
	double fishrotate;   /* Rotate fisheye camera    */
} CAMERA;

typedef struct {
	int debug;
	int stereo;
	int fullscreen;
   int screenwidth;
   int screenheight;
	int dometype;
	int showdomeintensity;
	int rendermode;
	int showtextures;
	int showboundingbox;
	int recordimages;
	int windowdump;
	int makeqtvr;
	int showhelpinfo;
	int projectiontype;
	int selectmode;
	int topbottom;
	double targetfps;
	int doantialias;
	int cursortype;
	int interaction;
	int autopilot;
	XYZ autospin;
	XYZ scalefactor;
	int combineeyes;
	double interactspeed;
	double deltamove;
	COLOUR background;
	int sphereresolution;
	double pointscale,linescale;
   int faceorder;
   int twosided;
   int localviewer;
	int texturetype;
	CAMERA camerahome;
} OPTIONS;

typedef struct {
	int currentbutton;
	int mousexlast,mouseylast;
	double framerate;
} INTERFACESTATE;

typedef struct {
	XYZ p;
	COLOUR colour;
	double size;
} DOT;
typedef struct {
	XYZ p[2];
	COLOUR colour[2];
	double width;
} LINE;
typedef struct {
   XYZ p[3];
	XYZ n[3];
   COLOUR colour[3];
} FACE3;
typedef struct {
   XYZ p[4];
   XYZ n[4];
   COLOUR colour[4];
} FACE4;
typedef struct {
   XYZ p[4];
	COLOUR colour;
   char texturename[64];
	int width,height;
	double scale;
	int trans;
	BITMAP4 *rgba;
	GLuint textureid;
} FACE4T;
typedef struct {
	XYZ p;
	double r;
	COLOUR colour;
} BALL;
typedef struct {
   XYZ p;
   double r;
   COLOUR colour;
   char texturename[64];
	int width,height;
	BITMAP4 *rgba;
	GLuint textureid;
} BALLT;
typedef struct {
   XYZ p;
	XYZ n;
   double r1,r2;
   COLOUR colour;
} DISK;
typedef struct {
   XYZ p1,p2;
   double r1,r2;
   COLOUR colour;
} CONE;
typedef struct {
	int nv;
	int p[4];
} OFFFACE;
typedef struct {
	XYZ p;
	XYZ right,up;
	COLOUR colour;
	char s[64];
} LABEL;
typedef struct {
	GLfloat p[4];
	GLfloat c[4];
} ALIGHT;

/* Different stereo modes */
#define NOSTEREO     0
#define ACTIVESTEREO 1
#define DUALSTEREO   2

/* Different rendering modes */
#define WIREFRAME      0
#define SHADE_FLAT     1
#define SHADE_DIFFUSE  2
#define SHADE_SPECULAR 3

#define SHUTTERGLASSES 0
#define ANAGLYPH_RB    1
#define ANAGLYPH_RG    2

#define INSPECT 0
#define FLY     1
#define WALK    2

#define NOCURSOR    0
#define ARROWCURSOR 1
#define CROSSCURSOR 2

#define HOME        1
#define FOCUS       2
#define FRONT       3
#define BACK        4
#define LEFT        5
#define RIGHT       6
#define TOP         7
#define BOTTOM      8
#define AXONOMETRIC 9

#define IMAGESPLIT 8

#define KEYBOARDCONTROL 0
#define MOUSECONTROL    1

/* Stereo.c */
void HandleDisplay(void);
void CreateMenus(void);
void CreateOpenGL(void);
void MakeGeometry(int);
void MakeLighting(void);
void MakeMaterial(void);
void HandleReshape(int,int);
void HandleKeyboard(unsigned char key,int x, int y);
void HandleSpecialKeyboard(int key,int x, int y);
void RotateCamera(double,double,double,int);
void FlyCamera(double);
void TranslateCamera(double,double,int);
void HandleMouse(int,int,int,int);
void HandleMouseMotion(int,int);
void HandlePassiveMotion(int,int);
void HandleVisibility(int vis);
void HandleIdle(void);
int  ReadGeometry(char *);
void CleanGeometry(void);
void SaveGeomFile(void);
int  ReadOFF(char *);
void DeleteGeometry(void);
void CameraHome(int);
int  AutoPilot(int,char *);
void UpdateForm(void);
void ReadViewFile(char *);
void SaveViewFile(CAMERA);
void AppendCameraPosition(void);
void CreateAxes(void);
void DrawExtras(void);
void CreateProjection(int);
void AddLine2Database(XYZ,XYZ,COLOUR,COLOUR,double);
void AddFace2Database(XYZ *,int,COLOUR,double,XYZ);
void AddMarker2Database(int,double,XYZ,COLOUR);
void CalcBounds(void);
void UpdateBounds(XYZ);
void CleanExit(void);

// menus.c
void HandleMainMenu(int);
void HandleAnimMenu(int);
void HandleWindowMenu(int);
void HandleRenderMenu(int);
void HandleSpeedMenu(int);
void HandleCameraMenu(int);
void HandleStereoMenu(int);
void HandleAppearMenu(int);
void HandleFileMenu(int);
void HandleViewMenu(int);
void HandleProjectionMenu(int);

// dome.c
void DomeTextures(void);
int  DomeCreate(void);
void DrawDome(int,int);
void DomeSplitFace(DOMEFACE *,int *);
void DomeFlatten(XYZ *,double);
int  ReadMesh(char *);
void MakeDefaultMesh(void);
void CalcWarpData(void);
void EstimateWarp(double,double,double *,double *,double *);
void SetVertex(DOMEFACE);
int MirrorPosition(XYZ,double,XYZ,XYZ,XYZ *);

// misc.c
void CreateASphere(XYZ,double,int,int,int);
void GiveUsage(char *);
int  ReadVector(FILE *,XYZ *);
int  ReadColour(FILE *,COLOUR *);


