#include "geomviewer.h"
#include "cameraform.h"
#include "materialform.h"
#include "lightform.h"
#include "appearform.h"

extern CAMERA camera;
extern OPTIONS options;

extern int cameraformvisible;
extern int materialformvisible;
extern int lightformvisible;
extern int appearformvisible;

extern FD_cameraform *cameraform;
extern FD_materialform *materialform;
extern FD_lightform *lightform;
extern FD_appearform *appearform;

/*
   Form the menus
*/
void CreateMenus(void)
{
   int mainmenu,appearmenu,rendermenu,speedmenu;
   int animmenu,filemenu,cameramenu,projectionmenu,viewmenu;
   int windowmenu;

   filemenu = glutCreateMenu(HandleFileMenu);
   glutAddMenuEntry("Read GEOM file \"-og\"",1);
   glutAddMenuEntry("Read OFF file \"-oo\"",2);
   glutAddMenuEntry("Read view file \"-vf\"",3);
   glutAddMenuEntry("-----------------------",100);
   glutAddMenuEntry("Save GEOM file",10);
   glutAddMenuEntry("Save view file",11);
   glutAddMenuEntry("Append camera keyframe \"W\"",12);
   glutAddMenuEntry("-----------------------",101);
   glutAddMenuEntry("Save image as TGA \"f10\"",20);

   windowmenu = glutCreateMenu(HandleWindowMenu);
   glutAddMenuEntry(" 160 x 120 (NTSC 1/16)",1);
   glutAddMenuEntry(" 192 x 144 (PAL 1/16)",2);
   glutAddMenuEntry(" 320 x 240 (NTSC 1/4)",3);
   glutAddMenuEntry(" 384 x 256 (Photo CD)",4);
   glutAddMenuEntry(" 384 x 288 (PAL 1/4)",5);
   glutAddMenuEntry(" 400 x 300",6);
   glutAddMenuEntry(" 400 x 400",7);
   glutAddMenuEntry(" 640 x 480 (Full NTSC)",8);
   glutAddMenuEntry(" 600 x 600",9);
   glutAddMenuEntry(" 720 x 576 (DVD PAL)",10);
   glutAddMenuEntry(" 768 x 576 (Full PAL)",11);
   glutAddMenuEntry(" 800 x 600",12);
   glutAddMenuEntry("1024 x 768",13);
   glutAddMenuEntry("1024 x 1024",14);

   rendermenu = glutCreateMenu(HandleRenderMenu);
   glutAddMenuEntry("Wireframe rendering \"-rw\"",1);
   glutAddMenuEntry("Hidden line, flat colour \"-rf\"",2);
   glutAddMenuEntry("Shading Diffuse \"-rd\"",3);
   glutAddMenuEntry("Shading Specular \"-rs\"",4);
   glutAddMenuEntry("-----------------------",100);
   glutAddMenuEntry("Toggle textures",5);
   glutAddMenuEntry("Toggle antialiasing",6);
   glutAddMenuEntry("Toggle between CCW and CW",8);
   glutAddMenuEntry("-----------------------",101);
   glutAddMenuEntry("Material attributes",20);
   glutAddMenuEntry("Lighting attributes",21);

   viewmenu = glutCreateMenu(HandleViewMenu);
   glutAddMenuEntry("Front \"1\"       ",1);
   glutAddMenuEntry("Back \"2\"        ",2);
   glutAddMenuEntry("Left \"3\"        ",3);
   glutAddMenuEntry("Right \"4\"       ",4);
   glutAddMenuEntry("Top \"5\"         ",5);
   glutAddMenuEntry("Bottom \"6\"      ",6);
   glutAddMenuEntry("Axonometric \"7\" ",7);

   projectionmenu = glutCreateMenu(HandleProjectionMenu);
   glutAddMenuEntry("Perspective",10);
   glutAddMenuEntry("Orthographic",11);
   glutAddMenuEntry("Eye Settings",12);
   glutAddMenuEntry("Fisheye: Truncated bottom",TRUNCBOTTOM);
   glutAddMenuEntry("Fisheye: Truncated top",TRUNCTOP);
   glutAddMenuEntry("Fisheye: Full dome (Vertical)",VSPHERICAL);
   glutAddMenuEntry("Fisheye: Full dome (Horizontal)",HSPHERICAL);
   glutAddMenuEntry("Fisheye: Warp map",WARPMAP);
   glutAddMenuEntry("Fisheye: Mirror, dome edge configuration",MIRROR1);

   cameramenu = glutCreateMenu(HandleCameraMenu);
   glutAddMenuEntry("Home camera position \"h\"",1);
   glutAddMenuEntry("Focus on model centroid \"f\"",2);
   glutAddSubMenu("Special views",viewmenu);
   glutAddSubMenu("Projection",projectionmenu);
   glutAddSubMenu("Window size",windowmenu);

   speedmenu = glutCreateMenu(HandleSpeedMenu);
   glutAddMenuEntry("Extremely slow",1);
   glutAddMenuEntry("Very slow",2);
   glutAddMenuEntry("Slow",3);
   glutAddMenuEntry("Medium",4);
   glutAddMenuEntry("Faster",5);
   glutAddMenuEntry("Fast",6);
   glutAddMenuEntry("Very fast",7);

   animmenu = glutCreateMenu(HandleAnimMenu);
   glutAddMenuEntry("Inspect mode",1);
   glutAddMenuEntry("Fly mode",2);
   glutAddMenuEntry("Walk mode",3);
   glutAddMenuEntry("--------------",100);
   glutAddMenuEntry("Toggle auto spin \"a\" \"-as\"",10);
   glutAddMenuEntry("Toggle auto pilot \"-ap fn\"",11);
   glutAddMenuEntry("--------------",101);
   glutAddSubMenu("Speed",speedmenu);

   appearmenu = glutCreateMenu(HandleAppearMenu);
   glutAddMenuEntry("Background colour \"-bg r g b\"",1);
   glutAddMenuEntry("Toggle bounding box \"f4\"",4);
   glutAddMenuEntry("Scale (x,y,z,point,line)",6);

   mainmenu = glutCreateMenu(HandleMainMenu);
   glutAddSubMenu("File",filemenu);
   glutAddSubMenu("Rendering",rendermenu);
   glutAddSubMenu("Camera",cameramenu);
   glutAddSubMenu("Appearance",appearmenu);
   glutAddSubMenu("Animation",animmenu);
   glutAddMenuEntry("Delete all \"D\"",1);
   glutAddMenuEntry("Show help info \"f1\"",2);
   glutAddMenuEntry("Quit \"q,esc\"",10);

   glutAttachMenu(GLUT_RIGHT_BUTTON);
}

/*
   Handle the File menu
*/
void HandleFileMenu(int whichone)
{
   switch (whichone) {
   case 1:         /* Open gem file */
      if (ReadGeometry("")) {
         CleanGeometry();
         CameraHome(HOME);
      }
      MakeGeometry(TRUE);
      break;
   case 2:         /* Open off file */
      if (ReadOFF("")) {
         CleanGeometry();
         CameraHome(HOME);
      }
      MakeGeometry(TRUE);
      break;
   case 3:        /* Open view file */
      ReadViewFile("");
      break;
   case 10:       /* Save as geom */
      SaveGeomFile();
      break;
   case 11:       /* Save view file */
      SaveViewFile(camera);
      break;
   case 12:       /* Append camera position */
      AppendCameraPosition();
      break;
   case 20: /* Save as image */
      if (options.stereo == ACTIVESTEREO)
         WindowDump("",options.screenwidth,options.screenheight,TRUE,12);
      else
         WindowDump("",options.screenwidth,options.screenheight,FALSE,12);
      break;
   }
}

/* 
   Handle the rendering menu
*/
void HandleRenderMenu(int whichone)
{  
   switch (whichone) {
   case 1: 
      options.rendermode = WIREFRAME;
      break;
   case 2:
      options.rendermode = SHADE_FLAT;
      break;
   case 3: 
      options.rendermode = SHADE_DIFFUSE;
      break;
   case 4: 
      options.rendermode = SHADE_SPECULAR;
      break;                        
   case 5:                          /* Toggle textures */
      options.showtextures = !options.showtextures;
      break;                        
   case 6:                          /* Antialiasing */
      options.doantialias = !options.doantialias;
      if (!options.doantialias) {
         glDisable(GL_LINE_SMOOTH);
         glDisable(GL_POINT_SMOOTH);
         glDisable(GL_POLYGON_SMOOTH);
      } else {
         glEnable(GL_LINE_SMOOTH);
         glEnable(GL_POINT_SMOOTH);
         glDisable(GL_POLYGON_SMOOTH);
         glHint(GL_POINT_SMOOTH_HINT,GL_NICEST);
         glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);
         //glHint(GL_POLYGON_SMOOTH_HINT,GL_NICEST);
      }
      break;                  
   case 8:
      if (options.faceorder == GL_CCW)
         options.faceorder = GL_CW;
      else
         options.faceorder = GL_CCW;
      glFrontFace(options.faceorder);
      break;                   
   case 20:                    /* Material setting dialog */
      UpdateForm();
      fl_show_form(materialform->materialform,FL_PLACE_CENTER,TRUE,NULL);
      materialformvisible = TRUE;
      break;
   case 21:
      UpdateForm();
      fl_show_form(lightform->lightform,FL_PLACE_CENTER,TRUE,NULL);
      lightformvisible = TRUE;
      break;
   }
   MakeGeometry(TRUE);
}

/* 
   Handle the rotate speed menu
*/
void HandleSpeedMenu(int whichone)
{  
   switch (whichone) {
   case 1: options.interactspeed = 0.01; break;
   case 2: options.interactspeed = 0.25; break;
   case 3: options.interactspeed = 0.5;  break;
   case 4: options.interactspeed = 1.0;  break;
   case 5: options.interactspeed = 2.0;  break;
   case 6: options.interactspeed = 5.0;  break;
   case 7: options.interactspeed = 20.0; break;
   }
}

/* 
   Handle the view menu
*/
void HandleViewMenu(int whichone)
{  
   switch (whichone) {        
   case 1:                    /* Front  */
   case 2:                    /* Back   */
   case 3:                    /* Left   */
   case 4:                    /* Right  */
   case 5:                    /* Top    */
   case 6:                    /* Bottom */
      CameraHome(whichone+2);
      break;                  
   case 7:                    /* Axonometric */
      CameraHome(AXONOMETRIC);
      break;
   }
}

/* 
   Handle the camera menu
*/
void HandleCameraMenu(int whichone)
{  
   switch (whichone) {        
   case 1:                    /* Move the camera to the home position */
      CameraHome(HOME);
      break;                  
   case 2:                    /* Focus the camera to the center of geometry */
      CameraHome(FOCUS);
      break;
   }
}

/* 
   Handle the window menu
*/
void HandleWindowMenu(int whichone)
{  
   switch (whichone) {
   case  1: options.screenwidth =  160; options.screenheight =  120; break;
   case  2: options.screenwidth =  192; options.screenheight =  144; break;
   case  3: options.screenwidth =  320; options.screenheight =  240; break;
   case  4: options.screenwidth =  384; options.screenheight =  256; break;
   case  5: options.screenwidth =  384; options.screenheight =  288; break;
   case  6: options.screenwidth =  400; options.screenheight =  300; break;
   case  7: options.screenwidth =  400; options.screenheight =  400; break;
   case  8: options.screenwidth =  640; options.screenheight =  480; break;
   case  9: options.screenwidth =  600; options.screenheight =  600; break;
   case 10: options.screenwidth =  720; options.screenheight =  576; break;
   case 11: options.screenwidth =  768; options.screenheight =  576; break;
   case 12: options.screenwidth =  800; options.screenheight =  600; break;
   case 13: options.screenwidth = 1024; options.screenheight =  768; break;
   case 14: options.screenwidth = 1024; options.screenheight = 1024; break;
   }
   glutReshapeWindow(options.screenwidth,options.screenheight);
}

/*
   Handle the projection menu
*/
void HandleProjectionMenu(int whichone)
{
   switch (whichone) {
	case TRUNCBOTTOM:
	case TRUNCTOP:
	case VSPHERICAL:
	case HSPHERICAL:
	case WARPMAP:
	case MIRROR1:
		options.projectiontype = FISHEYE;
		options.dometype = whichone;
   	CalcWarpData();
   	DrawDome(FALSE,TRUE);
		break;
   case 10:
      options.projectiontype = PERSPECTIVE;
      break;
   case 11:
      options.projectiontype = ORTHOGRAPHIC;
      break;
   case 12:
      UpdateForm();
      fl_show_form(cameraform->cameraform,FL_PLACE_CENTER,TRUE,NULL);
      cameraformvisible = TRUE;
      MakeGeometry(TRUE);
      break;
   }
}

/*
   Handle the fly menu
*/
void HandleAnimMenu(int whichone)
{
	XYZ origin = {0,0,0};

   switch (whichone) {
   case 1:                    /* Inspection mode */
      options.autospin = origin;
      options.autopilot = FALSE;
      options.interaction = INSPECT;
      break;
   case 2:                    /* Fly mode */
      options.autospin = origin;
      options.autopilot = FALSE;
      options.interaction = FLY;
      break;
   case 3:                    /* Walk mode */
      options.autospin = origin;
      options.autopilot = FALSE;
      options.interaction = WALK;
      break;
   case 10:                   /* Auto Spin */
      if (options.autospin.x != 0 || options.autospin.y != 0 || options.autospin.z != 0)
         options.autospin = origin;
      else
         options.autospin.x = 1;
      options.autopilot = FALSE;
      break;
   case 11:                   /* Autopilot */
      options.autospin = origin;
      options.autopilot = !options.autopilot;
      options.autopilot = AutoPilot(0,"");
      break;
   }
}

/*
   Handle the appearance menu
*/
void HandleAppearMenu(int whichone)
{
   int i,r,g,b;

   switch (whichone) {
   case 1:                    /* Background */
      i = fl_show_colormap(2);
      fl_getmcolor(i,&r,&g,&b);
      options.background.r = r / 255.0;
      options.background.g = g / 255.0;
      options.background.b = b / 255.0;
      glClearColor(options.background.r,options.background.g,options.background.b,0.0);
      break;
   case 3:
      break;
   case 4:
      options.showboundingbox = !options.showboundingbox;
      break;
   case 6:
      UpdateForm();
      fl_show_form(appearform->appearform,FL_PLACE_CENTER,TRUE,NULL);
      appearformvisible = TRUE;
      break;
   }
   MakeGeometry(TRUE);
}

/* 
   Handle the main menu
*/
void HandleMainMenu(int whichone)
{  
   switch (whichone) {        
   case 1:                    /* Delete all */
      DeleteGeometry();
      break;                  
   case 2:                    /* Show help information */
      options.showhelpinfo = !options.showhelpinfo;
      break;                  
   case 10:                   /* Quit */
      CleanExit();
      break;
   }
}

