#include "geomviewer.h"
#include "usagestring.h"

/*
   Create a simple sphere
	twidth and theight allow us to deal with power of 2 (or not) textures
*/       
void CreateASphere(XYZ c,double r,int n,int twidth,int theight)
{        
   int i,j,nn;
   double theta1,theta2,theta3;
   XYZ e,p;

   if (r < 0)
      r = -r;
   if (n < 0)
      n = -n;
   if (n < 4 || r <= 0) {
      glBegin(GL_POINTS);
      glVertex3f(c.x,c.y,c.z);
      glEnd();
      return;
   }     

   /* Make n even */
   n /= 2;
   n *= 2;
   nn = n/2;

   for (j=0;j<nn;j++) {
      theta1 = j * TWOPI / n - PID2; 
      theta2 = (j + 1) * TWOPI / n - PID2;

      glBegin(GL_QUAD_STRIP);

      for (i=0;i<=n;i++) {
         theta3 = i * TWOPI / n;

         e.x = cos(theta2) * cos(theta3);
         e.y = sin(theta2);
         e.z = cos(theta2) * sin(theta3);
         p.x = c.x + r * e.x;
         p.y = c.y + r * e.y;
         p.z = c.z + r * e.z;

         glNormal3f(e.x,e.y,e.z);
         glTexCoord2f(twidth*i/(double)n,2*theight*(j+1)/(double)n);
         glVertex3f(p.x,p.y,p.z);

         e.x = cos(theta1) * cos(theta3);
         e.y = sin(theta1);
         e.z = cos(theta1) * sin(theta3);
         p.x = c.x + r * e.x;
         p.y = c.y + r * e.y;
         p.z = c.z + r * e.z;
         
         glNormal3f(e.x,e.y,e.z);
         glTexCoord2f(twidth*i/(double)n,2*theight*j/(double)n);
         glVertex3f(p.x,p.y,p.z);
      }
      glEnd();
   }
}

/*
   Display the program usage information
*/
void GiveUsage(char *cmd)
{
   fprintf(stderr,"Usage: %s [command line options]\n",cmd);
   fprintf(stderr,"%s\n",helpstring);
   exit(-1);
}

int ReadVector(FILE *fptr,XYZ *p)
{
   if (fscanf(fptr,"%lf %lf %lf",&(p->x),&(p->y),&(p->z)) != 3)
      return(FALSE);
   return(TRUE);
}

int ReadColour(FILE *fptr,COLOUR *c)
{
   if (fscanf(fptr,"%lf %lf %lf",&(c->r),&(c->g),&(c->b)) != 3)
      return(FALSE);
   ClipColour(c);
   return(TRUE);
}


