#include "stdio.h"
#include "stdlib.h"
#include "math.h"

/*
	Simple generator of points on the Lorenz attractor
	dx/dt = a (y - x)
	dy/dt = x (b - z) - y
	dz/dt = x y - c z
*/

// Generate this many points
#define N 20000

int main(int argc,char **argv)
{
	int i=0;
	double x0,y0,z0,x1,y1,z1;

	// Standard parameter set
	double h = 0.003;
	double a = 10.0;
	double b = 28.0;
	double c = 8.0 / 3.0;

	// Initial conditions
  	x0 = 0.1;
  	y0 = 0;
  	z0 = 0;

	// Iterate, ignore the first 100 points
	for (i=0;i<N+100;i++) {
  		x1 = x0 + h * a * (y0 - x0);
  		y1 = y0 + h * (x0 * (b - z0) - y0);
  		z1 = z0 + h * (x0 * y0 - c * z0);

		// Write out the points in GEOM format
		if (i > 100)
			printf("p %g %g %g 1 0.5 0.5\n",x0,y0,z0);

      x0 = x1;
      y0 = y1;
      z0 = z1;
	}

	exit(0);
}

