#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#if defined(__linux__)
   #include <GL/glut.h>
#elif defined(__APPLE__)
   #include <GLUT/glut.h>
#else
   #warning "Not sure where to find glut.h"
#endif
#include <sys/time.h>
#include <sys/resource.h>
#include "paulslib.h"
#include "bitmaplib.h"
#include "opengllib.h"
#include "forms.h"

typedef struct {
   XYZ vp;           	/* View position           */
   XYZ vd;           	/* View direction vector   */
   XYZ vu;           	/* View up direction       */
	XYZ pr;					/* Point to rotate about 	*/
   double focallength;  /* Focal Length along vd	*/
   double aperture;  	/* Camera aperture         */
	double eyesep;			/* Eye separation				*/
	int screenwidth,screenheight;
} CAMERA;

typedef struct {
	int stereo;
	int combineeyes;
	int fullscreen;
	int geommode;
	int rendermode;
	int showhelpinfo;
	int colourmap;
	COLOUR background;
	COLOUR constcolour;
	int showaxes;
	int showaxesgrid;
	int showboundingbox;
	int showconstruction;
	int resolution;
	int debug;
	int currentbutton;
	XYZ scalefactors;
	int dobigdump;
	int windowdump;
	int recordimages;
	int ncontour;
	double targetfps;
	int datatype;
	int autopilot;
	int autospin;
	double dtheta;
} OPTIONS;

#define NOSTEREO     0
#define ACTIVESTEREO 1
#define DUALSTEREO   2

// Geometry modes
#define POINTS      1
#define SURFACE     2
#define SPIKES      3
#define BOXES       4
#define XCONTOUR    5
#define YCONTOUR    6
#define ZCONTOUR    7
#define XYCONTOUR   8

// Rendering modes
#define WIREFRAME   1
#define FLAT        2
#define SMOOTH      3
#define LIGHT       4 

#define HOME        1
#define FOCUS       2
#define FRONT       3
#define BACK        4
#define LEFT        5
#define RIGHT       6
#define TOP         7
#define BOTTOM      8
#define AXONOMETRIC 9

#define NODATA     -1
#define GRIDDED    0
#define TRIANGULAR 1
#define POINTED    2
#define GRIDDED2   3

void HandleDisplay(void);
void CreateEnvironment(void);
void MakeGeometry(void);
void DrawTheAxes(void);
void DrawBoundingBox(void);
void MakeLighting(void);
void HandleReshape(int,int);
void HandleKeyboard(unsigned char key,int x, int y);
void HandleSpecialKeyboard(int key,int x, int y);
void RotateCamera(double,double,double,double);
void TranslateCamera(int,int);
void HandleMouse(int,int,int,int);
void HandleMouseMotion(int,int);
void HandlePassiveMotion(int,int);
void HandleVisibility(int vis);
void HandleIdle(void);
void CameraHome(int);
void DrawExtras(void);
int  AutoPilot(int);
void UpdateForm(void);

// Misc.c
void ComputeBounds(void);
void GiveUsage(char *);
void DeleteData(void);
XYZ  GridNormal(int,int);
void DrawLights(void);
void MakeMaterial(void);
COLOUR SetColour(XYZ);
XYZ SetPoint(XYZ);
int XYZCompare(const void *,const void *);
void DrawABox(XYZ,XYZ);

// IO.c
void ReadViewFile(char *,CAMERA *);
void SaveViewFile(CAMERA);
int  ReadGrid(char *);
int  ReadRandom(char *,int);

// Menus.c
void CreateMenus(void);
void HandleMainMenu(int);
void HandleRenderMenu(int);
void HandleGeomMenu(int);
void HandleViewMenu(int);
void HandleColourMenu(int);
void HandleCameraMenu(int);
void HandleAppearMenu(int);
void HandleViewMenu(int);
void HandleFileMenu(int);
void HandleResolMenu(int);

