#include "glgraph.h"

extern OPTIONS options;
extern CAMERA camera;

#include "stereoform.h"
#include "scaleform.h"
#include "contourform.h"

extern FD_stereoform *stereoform;
extern int stereoformvisible;
extern FD_scaleform *scaleform;
extern FD_contourform *contourform;

/*
   Form the menu
*/
void CreateMenus(void)
{
   int mainmenu,rendermenu,appearmenu,geommenu;
   int filemenu,resolmenu,colourmenu,cameramenu,viewmenu;

   filemenu = glutCreateMenu(HandleFileMenu);
   glutAddMenuEntry("Read grid file with axes (-og)",1);
	glutAddMenuEntry("Read grid file without axes (-og2)",4);
   glutAddMenuEntry("Read random points and triangulate (-or)",2);
   glutAddMenuEntry("Read random points (-op)",3);
	glutAddMenuEntry("-----------------",100);
   glutAddMenuEntry("Read view file (-vf)",10);
   glutAddMenuEntry("Save view file",11);
   glutAddMenuEntry("-----------------",101);
   glutAddMenuEntry("Save image (w)",20);

   geommenu = glutCreateMenu(HandleGeomMenu);
   glutAddMenuEntry("Points (-rp)",1);
	glutAddMenuEntry("Surface (-rs)",2);
	glutAddMenuEntry("Spikes (-rk)",3);
	glutAddMenuEntry("Boxes (-rb)",4);
   glutAddMenuEntry("X Contours (-rx)",5);
   glutAddMenuEntry("Y Contours (-ry)",6);
   glutAddMenuEntry("Z Contours (-rz)",7);
   glutAddMenuEntry("XY Contours (-rxy)",8);

	rendermenu = glutCreateMenu(HandleRenderMenu);
   glutAddMenuEntry("Wireframe (-sw)",1);
   glutAddMenuEntry("Flat shaded (-sf)",2);
   glutAddMenuEntry("Smooth shaded (-ss)",3);
   glutAddMenuEntry("Lights (-sl)",4);

   resolmenu = glutCreateMenu(HandleResolMenu);
   glutAddMenuEntry("1:1",1);
   glutAddMenuEntry("2:2",2);
   glutAddMenuEntry("3:3",3);
   glutAddMenuEntry("4:4",4);
   glutAddMenuEntry("5:5",5);

   colourmenu = glutCreateMenu(HandleColourMenu);
   glutAddMenuEntry("blue-cyan-green-magenta-red",1);
   glutAddMenuEntry("blue-red",2);
   glutAddMenuEntry("grey scale",3);
   glutAddMenuEntry("red-yellow-green-cyan-blue-magenta-red",4);
   glutAddMenuEntry("green-yellow",5);
   glutAddMenuEntry("green-magenta",6);
   glutAddMenuEntry("blue-green-red-green-blue",7);
   glutAddMenuEntry("white-black-white",8);
   glutAddMenuEntry("red-blue-cyan-magenta",9);
   glutAddMenuEntry("blue-cyan-green-yellow-red-white",10);
   glutAddMenuEntry("Mars landscape",12);
   glutAddMenuEntry("Earth landscape",13);
   glutAddMenuEntry("Yellow-red",14);
   glutAddMenuEntry("blue-cyan-green-yellow-brown-white",15);
   glutAddMenuEntry("blue-green-red (ChromaDepth on black)",16);
   glutAddMenuEntry("yellow-magenta-cyan (CromaDepth on white)",17);
   glutAddMenuEntry("constant",100);

   appearmenu = glutCreateMenu(HandleAppearMenu);
   glutAddSubMenu("Colour ramp",colourmenu);
   glutAddMenuEntry("Height scale (-z n) \"z\"",1);
   glutAddMenuEntry("Background colour",2);
   glutAddSubMenu("Resolution (-res n)",resolmenu);
   glutAddMenuEntry("Contour levels \"c\"",3);
	glutAddMenuEntry("Toggle help information \"f1\"",10);
	glutAddMenuEntry("Toggle combine eyes \"f2\"",11);
	glutAddMenuEntry("Toggle image recording \"f3\"",12);
	glutAddMenuEntry("Toggle construction details \"f4\"",13);
	glutAddMenuEntry("Toggle Bounding Box \"f5\"",14);
   glutAddMenuEntry("Toggle Axes \"f6\"",15);
   glutAddMenuEntry("Toggle Axes Grid \"f7\"",16);

   viewmenu = glutCreateMenu(HandleViewMenu);
   glutAddMenuEntry("Front \"1\"       ",1);
   glutAddMenuEntry("Back \"2\"        ",2);
   glutAddMenuEntry("Left \"3\"        ",3);
   glutAddMenuEntry("Right \"4\"       ",4);
   glutAddMenuEntry("Top \"5\"         ",5);
   glutAddMenuEntry("Bottom \"6\"      ",6);
   glutAddMenuEntry("Axonometric \"7\" ",7);

   cameramenu = glutCreateMenu(HandleCameraMenu);
   glutAddMenuEntry("Toggle auto pilot",1);
   glutAddMenuEntry("Toggle auto spin",2);
   glutAddMenuEntry("Home \"h\"",3);
   glutAddSubMenu("Special views",viewmenu);
   glutAddMenuEntry("Stereo settings",4);

   mainmenu = glutCreateMenu(HandleMainMenu);
   glutAddSubMenu("File",filemenu);
   glutAddSubMenu("Geometry",geommenu);
   glutAddSubMenu("Rendering",rendermenu);
   glutAddSubMenu("Camera settings",cameramenu);
   glutAddSubMenu("Appearance",appearmenu);
   glutAddMenuEntry("Delete all \"D\"",9);
   glutAddMenuEntry("Quit",100);

   glutAttachMenu(GLUT_RIGHT_BUTTON);
}

/*
   Handle the colourmap menu
*/
void HandleColourMenu(int whichone)
{
   int i,r,g,b;

   options.colourmap = whichone;
   if (options.colourmap == 100) {
      i = fl_show_colormap(2);
      fl_getmcolor(i,&r,&g,&b);
      options.constcolour.r = r / 255.0;
      options.constcolour.g = g / 255.0;
      options.constcolour.b = b / 255.0;
   }
   MakeGeometry();
}

/*
   Handle the Open menu
*/
void HandleFileMenu(int whichone)
{
   switch (whichone) {
   case 1:                                         /* Open grid file */
		options.datatype = GRIDDED;
      ReadGrid("");
      CameraHome(HOME);
      MakeGeometry();
      break;
   case 4:                                         /* Open grid file */
		options.datatype = GRIDDED2;
      ReadGrid("");
      CameraHome(HOME);
      MakeGeometry();
      break;
   case 2:                          /* Open random point file and triangulate */
   case 3:                          /* Don't triangulate */
      if (whichone == 3) {
         ReadRandom("",FALSE);
         options.datatype = POINTED;
      } else {
         ReadRandom("",TRUE);
         options.datatype = TRIANGULAR;
      }
      CameraHome(HOME);
      MakeGeometry();
      break;
   case 10:                          /* Read view file */
      ReadViewFile("",&camera);
      break;
   case 11:                          /* Save view file */
      SaveViewFile(camera);
      break;
	case 20:
		options.windowdump = TRUE;
		break;
   }
}

/*
   Handle the geometry menu
*/
void HandleGeomMenu(int whichone)
{
   options.geommode = whichone;
	if (options.geommode != SURFACE && options.geommode != BOXES)
		options.rendermode = WIREFRAME;
   MakeGeometry();
}

/*
   Handle the rendering menu
*/
void HandleRenderMenu(int whichone)
{
   options.rendermode = whichone;
   MakeGeometry();
}

/*
   Handle the resolution menu
*/
void HandleResolMenu(int whichone)
{
   options.resolution = whichone;
   MakeGeometry();
}

/*
   Deal with the appearance menu
*/
void HandleAppearMenu(int whichone)
{
   int i,r,g,b;

   switch (whichone) {
   case 1:                                                  // z scale 
      HandleKeyboard('z',0,0);
      break;
   case 2:                                                  // Background 
      i = fl_show_colormap(2);
      fl_getmcolor(i,&r,&g,&b);
      options.background.r = r/255.0;
      options.background.g = g/255.0;
      options.background.b = b/255.0;
      glClearColor(options.background.r,options.background.g,options.background.b,0.0);
      break;
   case 3:                                                  // Contour levels 
      HandleKeyboard('c',0,0);
      break;
   case 10:                                                  // Help info
		HandleSpecialKeyboard(GLUT_KEY_F1,0,0);
      break;
   case 11:                                                  // combine eyes
      HandleSpecialKeyboard(GLUT_KEY_F2,0,0);
      break;
   case 12:                                                  // image recording
      HandleSpecialKeyboard(GLUT_KEY_F3,0,0);
      break;
   case 13:                                                  // construction
      HandleSpecialKeyboard(GLUT_KEY_F4,0,0);
      break;
   case 14:                                                  // bounding box
      HandleSpecialKeyboard(GLUT_KEY_F5,0,0);
      break;
   case 15:                                                  // show axes
      HandleSpecialKeyboard(GLUT_KEY_F6,0,0);
      break;
   case 16:                                                  // show grid
      HandleSpecialKeyboard(GLUT_KEY_F7,0,0);
      break;
   }
   MakeGeometry();
}

/*
   Handle the view menu
*/
void HandleViewMenu(int whichone)
{
   switch (whichone) {
   case 1:                    /* Front  */
   case 2:                    /* Back   */
   case 3:                    /* Left   */
   case 4:                    /* Right  */
   case 5:                    /* Top    */
   case 6:                    /* Bottom */
      CameraHome(whichone+2);
      break;
   case 7:                    /* Axonometric */
      CameraHome(AXONOMETRIC);
      break;
   }
}

/*
   Handle the camera menu
*/
void HandleCameraMenu(int whichone)
{
   switch (whichone) {
   case 1:                                                  /* Autopilot */
      options.autospin = FALSE;
      options.autopilot = !options.autopilot;
      options.autopilot = AutoPilot(0);
      break;
   case 2:                                                  /* Auto Spin */
      options.autospin = !options.autospin;
      options.autopilot = FALSE;
      break;
   case 3:                                                  /* Home view */
      CameraHome(HOME);
      break;
   case 4:                                                  /* Stereo Options */
      UpdateForm();
      fl_show_form(stereoform->stereoform,FL_PLACE_CENTER,TRUE,NULL);
      stereoformvisible = TRUE;
      break;
   }
}

/*
   Handle the main menu
*/
void HandleMainMenu(int whichone)
{
   switch (whichone) {
   case 9:                                                  /* Delete all */
      HandleKeyboard('D',0,0);
      break;
   case 100:                                                /* Quit */
      exit(0);
      break;
   }
}

